/*
 * The contents of this file are subject to the MonetDB Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://monetdb.cwi.nl/Legal/MonetDBLicense-1.1.html
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is the MonetDB Database System.
 *
 * The Initial Developer of the Original Code is CWI.
 * Portions created by CWI are Copyright (C) 1997-July 2008 CWI.
 * Copyright August 2008-2010 MonetDB B.V.
 * All Rights Reserved.
 */

#include "sql_config.h"
#include "sql.h"
#include "sql_result.h"
#include "sql_gencode.h"
#include <sql_storage.h>
#include <sql_scenario.h>
#include <store_sequence.h>
#include <sql_optimizer.h>
#include <sql_datetime.h>
#include <rel_optimizer.h>
#include <rel_select.h>
#include <rel_exp.h>
#include <rel_dump.h>
#include <rel_bin.h>
#include <math.h>
#include <bbp.h>
#include <cluster.h>
#include <replication.h>
#include <opt_dictionary.h>
#ifdef HAVE_MONETDB5_RDF
	#include <rdf/rdf.h>
#endif

backend *
backend_reset(backend *b)
{
	b->out = b->client->fdout;
	b->language = 0;

	b->currSchema = NULL;
	b->currTable = NULL;
	b->currColumn = NULL;
	b->currKey = NULL;
	b->currIndex = NULL;
	b->vtop = 0;
	b->q = NULL;
	return b;
}

backend *
backend_create(mvc *m, Client c)
{
	backend *b = NEW(backend);

	b->console = isAdministrator(c);
	b->mvc = m;
	b->client = c;
	return backend_reset(b);
}

void
backend_destroy(backend *b)
{
	_DELETE(b);
}



/*
static stmt *
sql_symbols2stmt(mvc *c, symbol *s) 
{
	dnode *d;
	list *l = create_stmt_list();

	for (d = s->data.lval->h; d; d = d->next) {
		symbol *sym = d->data.sym;
		stmt *st = sql_symbol2stmt(c, sym);

		if (!st) {
			list_destroy(l);
			return NULL;
		}
		list_append(l, st);
	}
	return stmt_list(l);
}
*/

stmt *
sql_symbol2stmt(mvc *c, symbol *sym)
{
	stmt *s = NULL;
	int status;
	sql_rel *r;

/*
	if (sym->token == SQL_MULSTMT) 
		return sql_symbols2stmt(c, sym);
*/

	status = c->session->status;
	r = rel_semantic(c, sym);
	if (!r) {
		if (c->errstr[0]) {
			return NULL;
		} else {
			c->session->status = status;
			/*printf("# falling back to direct mode \n");*/
			s = output_semantic(c, sym);
		}
	} else {
		r = rel_optimizer(c, r);
		if (c->emode == m_plan) {
			rel_print(c, r, 0);
		} else {
			s = output_rel_bin(c, r);
		}
		rel_destroy(r);
		if (s) {
			stmt *opt;

			/* only needed for delta tables */
			opt = rel2bin(c, s); 	
			stmt_destroy(s);
			s = bin_optimizer(c, opt);
			stmt_destroy(opt);
			return s;
		}
	}

	if (s){
		stmt *opt;

		if (mvc_debug_on(c,128)) 
			stmt2dot(s, 1, "/tmp/g1.dot");
		opt = rel2bin(c, s);
		stmt_destroy(s);
		s = opt;
		if (mvc_debug_on(c,128)) 
			stmt2dot(s, 2, "/tmp/g2.dot");
		opt = bin_optimizer(c, s);
		stmt_destroy(s);
		s = opt;
		if (mvc_debug_on(c,128)) 
			stmt2dot(s, 3, "/tmp/g3.dot");
	}
	return s;
}



int
sqlcleanup(mvc *c, int err)
{
	sql_destroy_params(c);
	sql_destroy_args(c);

	/* some statements dynamically disable caching */
	c->sym = NULL;
	if (c->sa) 
		c->sa = sa_reset(c->sa);
	if (err > 0)
		c->session->status = -err;
	if (err < 0)
		c->session->status = err;
	scanner_query_processed(&(c->scanner));
	return err;
}



str 
getContext(Client cntxt, MalBlkPtr mb, mvc **c, backend **b )
{
	(void)mb;
	if (cntxt == NULL ) 
		throw(SQL, "mvc","No client record");
	if (cntxt->state[MAL_SCENARIO_PARSER] == NULL || cntxt->state[MAL_SCENARIO_OPTIMIZE] == NULL)
		throw(SQL, "mvc","SQL module not initialized");
	*c = ((mvc *) cntxt ->state[MAL_SCENARIO_OPTIMIZE]); 
	if (*c == 0)
		throw(SQL, "mvc","SQL module not initialized, mvc struct missing");
	if (b)
		*b = ((backend*) cntxt->state[MAL_SCENARIO_PARSER]);
	if (b && *b == 0)
		throw(SQL, "mvc","SQL module not initialized, backend struct missing");
	return MAL_SUCCEED;
}

str
SQLmvc(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *sql = NULL;
	str msg = getContext(cntxt, mb, &sql, NULL);
	ptr *res = (ptr*) getArgReference(stk, pci, 0);

	if (msg)
		return msg;
	*res = (ptr)sql;
	return MAL_SUCCEED;
}

str
SQLtransaction(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *sql = NULL;
	str msg = getContext(cntxt, mb, &sql, NULL);
	int type = *(int*) getArgReference(stk, pci, 1);
	int chain = *(int*) getArgReference(stk, pci, 2);
	str name = *(str*) getArgReference(stk, pci, 3);
	char buf[BUFSIZ];
	int ret = 0;

	if (msg)
		return msg;
	if (name && strcmp(name, str_nil) == 0)
		name = NULL;

	switch (type) {
	case DDL_RELEASE:
		if (sql->session->auto_commit == 1) 
			throw(SQL, "sql.trans", "RELEASE SAVEPOINT: not allowed in auto commit mode");
		ret = mvc_release(sql, name);
		if (ret < 0) {
			snprintf(buf, BUFSIZ, "RELEASE SAVEPOINT: (%s) failed", name);
			throw(SQL, "sql.trans", buf);
		}
		break;
	case DDL_COMMIT:
		if (sql->session->auto_commit == 1) {
			if (name)
				throw(SQL,"sql.trans", "SAVEPOINT: not allowed in auto commit mode");
			else
				throw(SQL,"sql.trans", "COMMIT: not allowed in auto commit mode");
		}
		ret = mvc_commit(sql, chain, name);
		if (ret < 0 && !name) 
			throw(SQL, "sql.trans", "COMMIT: failed");
		if (ret < 0 && name) {
			snprintf(buf, BUFSIZ, "SAVEPOINT: (%s) failed", name);
			throw(SQL, "sql.trans", buf);
		}
		break;
	case DDL_ROLLBACK:
		if (sql->session->auto_commit == 1) 
			throw(SQL, "sql.trans", "ROLLBACK: not allowed in auto commit mode");
		ret = mvc_rollback(sql, chain, name);
		if (ret < 0 && name) {
			snprintf(buf, BUFSIZ, "ROLLBACK TO SAVEPOINT: (%s) failed", name);
			throw(SQL, "sql.trans", buf);
		}
		break;
	case DDL_TRANS:
		if (sql->session->auto_commit == 0) 
			throw(SQL, "sql.trans", "START TRANSACTION: cannot start a transaction within a transaction");
		if (sql->session->active)
			mvc_rollback(sql, 0, NULL);
		sql->session->auto_commit = 0;
		sql->session->ac_on_commit = 1;
		sql->session->level = chain;
		(void) mvc_trans(sql);
		break;
	default:
		throw(SQL, "sql.trans", "transaction unknown type");
	}
	return MAL_SUCCEED;
}

static str
create_table_or_view( mvc *sql, char *sname, sql_table *t, int temp)
{
	sql_schema *s = mvc_bind_schema(sql, sname);
	sql_table *nt = NULL;
	node *n;

	if (STORE_READONLY(active_store_type)) 
		return sql_error(sql, 06, "schema statements cannot be executed on a readonly database.");

	if (!s) 
		return sql_message("CREATE %s: schema '%s' doesn't exist", (t->query)?"TABLE":"VIEW", sname);

	if (mvc_bind_table(sql, s, t->base.name)) {
		char *cd = (temp == SQL_DECLARED_TABLE)?"DECLARE":"CREATE";
		return sql_message("%s TABLE: name '%s' already in use", cd, t->base.name);
	} else if (temp != SQL_DECLARED_TABLE &&!schema_privs(sql->role_id, s)){
		return sql_message("CREATE TABLE: insufficient privileges for user '%s' in schema '%s'", stack_get_string(sql, "current_user"), s->base.name);
	}

	/* first check default values */
	for (n = t->columns.set->h; n; n = n->next) {
		sql_column *c = n->data;

		if (c->def) {
			char buf[BUFSIZ];
			sql_rel *r = NULL;
			sql_exp *e;

			sql->sa = sa_create();
			snprintf(buf, BUFSIZ, "select %s;", c->def);
			r = rel_parse(sql, buf, m_deps); 
			if (!r || !is_project(r->op) || !r->exps || list_length(r->exps) != 1 || (e = rel_check_type(sql, &c->type, exp_dup(r->exps->h->data), type_equal)) == NULL) 
				throw(SQL, "sql.catalog", sql->errstr);
			exp_destroy(e);
			sa_destroy(sql->sa);
			sql->sa = NULL;
			rel_destroy(r);
		}
	}

	nt = sql_trans_create_table(sql->session->tr, s, t->base.name, t->query, t->type, t->system, temp, t->commit_action, t->sz);

	for (n = t->columns.set->h; n; n = n->next) {
		sql_column *c = n->data;
		mvc_copy_column(sql, nt, c);
	}
	if (t->idxs.set) {
		for (n = t->idxs.set->h; n; n = n->next) {
			sql_idx *i = n->data;
			mvc_copy_idx(sql, nt, i);
		}
	}
	if (t->keys.set) {
		for (n = t->keys.set->h; n; n = n->next) {
			sql_key *k = n->data;

			mvc_copy_key(sql, nt, k);
		}
	}
	/* TODO copy triggers */
	/* also create dependencies */
	if (t->query) {
		sql_rel *r = NULL;
		sql->sa = sa_create();
		r = rel_parse(sql, t->query, m_deps); 
		sa_destroy(sql->sa);
		sql->sa = NULL;
		rel_destroy(r);
	}
	return MAL_SUCCEED;
}

static str
alter_table( mvc *sql, char *sname, sql_table *t)
{
	sql_schema *s = mvc_bind_schema(sql, sname);
	sql_table *nt = NULL;
	node *n;

	if (!s) 
		return sql_message("ALTER TABLE: no such schema '%s'", sname);

	if ((nt = mvc_bind_table(sql, s, t->base.name)) == NULL) {
                return sql_message("ALTER TABLE: no such table '%s'", t->base.name);

	} else if (!schema_privs(sql->role_id, s)){
		return sql_message("ALTER TABLE: insufficient privileges for user '%s' in schema '%s'", stack_get_string(sql, "current_user"), s->base.name);
	}

	if (t->readonly != nt->readonly) 
		mvc_readonly( sql, nt, t->readonly);

	/* check for changes */
	if (t->columns.dset) for (n = t->columns.dset->h; n; n = n->next) {
		/* propagate alter table .. drop column */
		sql_column *c = n->data;
 		sql_column *nc = mvc_bind_column(sql, nt, c->base.name);
		mvc_drop_column(sql, nt, nc, c->drop_action);
	}
	/* check for changes on current cols */
	for (n = t->columns.set->h; n != t->columns.nelm; n = n->next) {

		/* null or default value changes */
		sql_column *c = n->data;
 		sql_column *nc = mvc_bind_column(sql, nt, c->base.name);

		if (c->null != nc->null) {
			mvc_null(sql, nc, c->null);
			/* for non empty check for nulls */
			if (c->null == 0){
				BAT *b = store_funcs.bind_col(sql->session->tr, nc, 0);

				/* TODO also check updates and inserts */
				if (BATcount(b) && b->T->nonil != TRUE) {
					BUN bun = BUNfnd(BATmirror(b), ATOMnilptr(b->ttype));
					if (bun != BUN_NONE)
						return sql_message ("ALTER TABLE: NOT NULL constraint violated for column %s.%s", c->t->base.name, c->base.name );
				}
				BBPunfix(b->batCacheid);
			}
		}
		if (c->def != nc->def)
			mvc_default(sql, nc, c->def);
	}
	for (; n; n = n->next) {
		/* propagate alter table .. add column */
		sql_column *c = n->data;
		mvc_copy_column(sql, nt, c);
	}
	if (t->idxs.set) {
		/* alter drop index */
		if (t->idxs.dset) for (n = t->idxs.dset->h; n; n = n->next) {
			sql_idx *i = n->data;
 			sql_idx *ni = mvc_bind_idx(sql, s, i->base.name);
			mvc_drop_idx(sql, s, ni);
		}
		/* alter add index */
		for (n = t->idxs.nelm; n; n = n->next) {
			sql_idx *i = n->data;
			mvc_copy_idx(sql, nt, i);
		}
	}
	if (t->keys.set) {
		/* alter drop key */
		if (t->keys.dset) for (n = t->keys.dset->h; n; n = n->next) {
			sql_key *k = n->data;
 			sql_key *nk = mvc_bind_key(sql, s, k->base.name);
			mvc_drop_key(sql, s, nk, k->drop_action);
		}
		/* alter add key */
		for (n = t->keys.nelm; n; n = n->next) {
			sql_key *k = n->data;
			mvc_copy_key(sql, nt, k);
		}
	}

	/* TODO copy triggers */
	/* also create dependencies */
	return MAL_SUCCEED;
}



static char *
drop_table(mvc *sql, char *sname, char *tname, int drop_action)
{
	sql_schema *s = NULL;
	sql_table *t = NULL;
	node *n; 

	if (sname && !(s=mvc_bind_schema(sql, sname))) 
		return sql_message("DROP TABLE: no such schema '%s'", sname);
	if (!s)
		s = cur_schema(sql);
	t = mvc_bind_table(sql, s, tname);
	if (!t && !sname) {
		s = tmp_schema(sql);
		t = mvc_bind_table(sql, s, tname);
	}
	if (!t) { 
		return sql_message("DROP TABLE: no such table '%s'", tname);
	} else if (!isTable(t)) {
		return sql_message("DROP TABLE: cannot drop VIEW '%s'", tname);
	} else if (t->system) {
		return sql_message("DROP TABLE: cannot drop system table '%s'", tname);
	} else if (!schema_privs(sql->role_id, s)) {
		return sql_message("DROP TABLE: access denied for %s to schema ;'%s'", stack_get_string(sql, "current_user"), s->base.name);
	}
	if ( !drop_action && t->keys.set) {
		for (n = t->keys.set->h; n; n = n->next) {
			sql_key *k = n->data;

			if (k->type == ukey || k->type == pkey) {
				sql_ukey *uk = (sql_ukey *) k;

				if (uk->keys && list_length(uk->keys)) {
					node *l = uk->keys->h;

					for (; l; l = l->next) {
						k = l->data;
						/* make sure it is not a self referencing key */
						if (k->t != t) 
							return sql_message("DROP TABLE: FOREIGN KEY %s.%s depends on %s", k->t->base.name, k->base.name, tname);
					}
				}
			}
		}
	}

	if (!drop_action && mvc_check_dependency(sql, t->base.id, TABLE_DEPENDENCY, NULL))
		return sql_message("DROP TABLE: unable to drop table %s (there are database objects which depend on it)\n", t->base.name);
	
	mvc_drop_table(sql, s, t, drop_action);
	return MAL_SUCCEED;
}

static char *
drop_view(mvc *sql, char *sname, char *tname, int drop_action)
{
	sql_table *t = NULL;
	sql_schema *ss = NULL;

	if (sname != NULL && (ss = mvc_bind_schema(sql, sname)) == NULL)
		return sql_message("DROP VIEW: no such schema '%s'", sname);

	if (ss == NULL)
		ss = cur_schema(sql);

	t = mvc_bind_table(sql, ss, tname);

	if (!schema_privs(sql->role_id, ss)) {  
		return sql_message("DROP VIEW: access denied for %s to schema '%s'", stack_get_string(sql, "current_user"), ss->base.name);
	} else if (!t) {
		return sql_message("DROP VIEW: unknown view '%s'", tname);
	} else if (!isView(t)) {
		return sql_message("DROP VIEW: unable to drop view '%s'", tname);
	} else if (t->system) {
		return sql_message("DROP VIEW: cannot drop system view '%s'", tname);
	} else if (! drop_action && mvc_check_dependency(sql, t->base.id, VIEW_DEPENDENCY, NULL)) {
			return sql_message("unable to drop view %s (there are database objects which depend on it)\n", t->base.name);
	} else {
		mvc_drop_table(sql, ss, t, drop_action);
		return MAL_SUCCEED;
	}
}

static str
drop_key( mvc *sql, char *sname, char *kname, int drop_action )
{
	sql_key *key;
	sql_schema *ss = NULL;

	if (sname != NULL && (ss = mvc_bind_schema(sql, sname)) == NULL)
		return sql_message("DROP VIEW: no such schema '%s'", sname);

	if (ss == NULL)
		ss = cur_schema(sql);

	if ((key = mvc_bind_key(sql, ss, kname )) == NULL) 
		return sql_message("ALTER TABLE: no such constraint '%s'", kname);
	if (!drop_action && mvc_check_dependency(sql, key->base.id, KEY_DEPENDENCY, NULL)) 
		return sql_message("ALTER TABLE: cannot drop constraint '%s': there are database objects which depend on it", key->base.name);
	mvc_drop_key(sql, ss, key, drop_action);
	return MAL_SUCCEED;
}

static str
create_seq( mvc *sql, char *sname, sql_sequence *seq)
{
	sql_schema *s = NULL;

	if (sname && !(s = mvc_bind_schema(sql, sname)))
		return sql_message("CREATE SEQUENCE: no such schema '%s'", sname);
	if (s == NULL)
		s = cur_schema(sql);
	if (find_sql_sequence(s, seq->base.name)) {
		return sql_message("CREATE SEQUENCE: name '%s' already in use", seq->base.name);
	} else if (!schema_privs(sql->role_id, s)) {
		return sql_message( "CREATE SEQUENCE: insufficient privileges for '%s' in schema '%s'", stack_get_string(sql, "current_user"), s->base.name);
	}
	sql_trans_create_sequence(sql->session->tr, s, seq->base.name, seq->start, seq->minvalue, seq->maxvalue, seq->increment, seq->cacheinc, seq->cycle, seq->bedropped);
	return NULL;
}

static str
alter_seq( mvc *sql, char *sname, sql_sequence *seq, lng *val)
{
	sql_schema *s = NULL;
	sql_sequence *nseq = NULL;

	if (sname && !(s = mvc_bind_schema(sql, sname)))
		return sql_message("ALTER SEQUENCE: no such schema '%s'", sname);
	if (s == NULL)
		s = cur_schema(sql);
	if (!(nseq = find_sql_sequence(s, seq->base.name))) {
		return sql_message( "ALTER SEQUENCE: no such sequence '%s'", seq->base.name);
	} else if (!schema_privs(sql->role_id, s)) {
		return sql_message( "ALTER SEQUENCE: insufficient privileges for '%s' in schema '%s'", stack_get_string(sql, "current_user"), s->base.name);
	}

	/* first alter the known values */
	sql_trans_alter_sequence(sql->session->tr, nseq, seq->minvalue, seq->maxvalue, seq->increment, seq->cacheinc, seq->cycle);
	if (val)
		sql_trans_sequence_restart(sql->session->tr, nseq, *val);
	return MAL_SUCCEED;
}




static str
drop_seq(mvc *sql, char *sname, char *name)
{
	sql_schema *s = NULL;
	sql_sequence *seq = NULL;

	if (sname && !(s = mvc_bind_schema(sql, sname)))
		return sql_message("DROP SEQUENCE: no such schema '%s'", sname);
	if (!s)
		s = cur_schema(sql);
	if (!(seq = find_sql_sequence(s, name))) {
		return sql_message("DROP SEQUENCE: no such sequence '%s'", name);
	} else if (!schema_privs(sql->role_id, s)) {
		return sql_message("DROP SEQUENCE: insufficient privileges for '%s' in schema '%s'", stack_get_string(sql, "current_user"),  s->base.name);
	}
	if (mvc_check_dependency(sql, seq->base.id, BEDROPPED_DEPENDENCY, NULL))
		return sql_message("DROP SEQUENCE: unable to drop sequence %s (there are database objects which depend on it)\n", seq->base.name);

	sql_trans_drop_sequence(sql->session->tr, s, seq, 0);
	return NULL;
}


str
SQLcatalog(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *sql = NULL;
	str msg = getContext(cntxt, mb, &sql, NULL);
	int type = *(int*) getArgReference(stk, pci, 1);
	str sname = *(str*) getArgReference(stk, pci, 2);

	if (msg)
		return msg;

	if (STORE_READONLY(active_store_type)) 
		return sql_message("schema statements cannot be executed on a readonly database.");

	switch (type) {
	case DDL_CREATE_SEQ:
	{
		sql_sequence *s = *(sql_sequence**) getArgReference(stk, pci, 3);
		msg = create_seq(sql, sname, s);
	} 	break;
	case DDL_ALTER_SEQ:
	{
		lng *val = NULL;
		sql_sequence *s = *(sql_sequence**) getArgReference(stk, pci, 3);
		if (getArgType(mb, pci, 4) == TYPE_lng) 
			val = getArgReference(stk, pci, 4);
		msg = alter_seq(sql, sname, s, val);
	} 	break;
	case DDL_DROP_SEQ: {
		str name = *(str*) getArgReference(stk, pci, 3);

		msg = drop_seq( sql, sname, name);
	} 	break;
	case DDL_CREATE_SCHEMA: {
		str name = *(str*) getArgReference(stk, pci, 3);
		int auth_id = sql->role_id;

		if (name && strcmp(name, str_nil) == 0)
			name = NULL;

		if (name && (auth_id = sql_find_auth(sql, name)) < 0) {
			msg = sql_message("CREATE SCHEMA: no such authorization '%s'", name);
		}
		if (sql->user_id != USER_MONETDB && sql->role_id != ROLE_SYSADMIN) {
			msg = sql_message("CREATE SCHEMA: insufficient privileges for user '%s'", stack_get_string(sql, "current_user"));
		}
		if (mvc_bind_schema(sql, sname)) {
			msg = sql_message("CREATE SCHEMA: name '%s' already in use", sname);
		} else {
			(void)mvc_create_schema(sql, sname, auth_id, sql->user_id);
		}
	}	break;
	case DDL_DROP_SCHEMA: {
		int action = *(int*) getArgReference(stk, pci, 4);
		sql_schema *s = mvc_bind_schema(sql, sname);

		if (!s) {
			msg = sql_message("DROP SCHEMA: name %s does not exist", sname);
		} else if (!schema_privs(sql->role_id, s)) {
			msg = sql_message("DROP SCHEMA: access denied for %s to schema ;'%s'", stack_get_string(sql, "current_user"), s->base.name);
		} else if (s == cur_schema(sql)) {
			msg = sql_message("DROP SCHEMA: cannot drop current schema");
		} else if (strcmp(sname, "sys") == 0 || strcmp(sname, "tmp") == 0) {
			msg = sql_message("DROP SCHEMA: access denied for '%s'", sname);
		} else if (sql_schema_has_user(sql, s)) {
			msg = sql_message("DROP SCHEMA: unable to drop schema '%s' (there are database objects which depend on it", sname);
		} else {
			mvc_drop_schema(sql, s, action);
		}
	} 	break;
	case DDL_CREATE_TABLE: 
	case DDL_CREATE_VIEW: 
	{
		sql_table *t = *(sql_table**) getArgReference(stk, pci, 3);
		int temp = *(int*) getArgReference(stk, pci, 4);

		msg = create_table_or_view(sql, sname, t, temp);
	} 	break;
	case DDL_DROP_TABLE: {
		int action = *(int*) getArgReference(stk, pci, 4);
		str name = *(str*) getArgReference(stk, pci, 3);

		msg = drop_table( sql, sname, name, action);
	} 	break;
	case DDL_DROP_VIEW: {
		int action = *(int*) getArgReference(stk, pci, 4);
		str name = *(str*) getArgReference(stk, pci, 3);

		msg = drop_view( sql, sname, name, action);
	} 	break;
	case DDL_ALTER_TABLE: 
	{
		sql_table *t = *(sql_table**) getArgReference(stk, pci, 3);
		msg = alter_table(sql, sname, t);
	} 	break;
	case DDL_DROP_CONSTRAINT: {
		int action = *(int*) getArgReference(stk, pci, 4);
		str name = *(str*) getArgReference(stk, pci, 3);

		msg = drop_key( sql, sname, name, action);
	} 	break;
	default:
		throw(SQL, "sql.catalog", "catalog unknown type");
	}
	if (msg)
		throw(SQL, "sql.catalog", msg);
	return MAL_SUCCEED;
}

/* setVariable(int *ret, str *name, any value) */
str
setVariable(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	char buf[BUFSIZ];
	mvc *m = NULL;
	int mtype = getArgType(mb, pci, 2);
	str varname   = *(str *) getArgReference(stk, pci, 1);
	str msg = getContext(cntxt,mb, &m, NULL);
	ValRecord *src;

	(void) cntxt;
	if (msg)
		return msg;

	if (mtype < 0 || mtype >= 255)
		throw(SQL, "sql.setVariable", "failed");
	if ( strcmp("optimizer",  varname)== 0) {
		msg = setOptimizers(*(str *) getArgReference(stk,pci,2));
		if ( msg != NULL && strcmp(msg,"default_pipe") == 0 )
			return MAL_SUCCEED;
		msg = SQLvalidatePipeline();
		if ( msg ){
			setOptimizers("default_pipe");
			return msg;
		}
	}
	src = &stk->stk[getArg(pci, 2)];
	if (stack_find_var(m, varname)) {
		stack_set_var(m, varname, src);
	} else {
		snprintf(buf, BUFSIZ, "variable '%s' unknown", varname);
		throw(SQL, "sql.setVariable", buf);
	}
	if ((msg = sql_update_var(m, varname)) != NULL) {
		snprintf(buf, BUFSIZ, "%s", msg);
		_DELETE(msg);
		throw(SQL, "sql.setVariable", buf);
	}
	return MAL_SUCCEED;
}

/* getVariable(int *ret, str *name) */
str
getVariable(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	int mtype = getArgType(mb, pci, 0);
	str varname   = *(str *) getArgReference(stk, pci, 1);
	str msg = getContext(cntxt,mb, &m, NULL);
	ValRecord *dst, *src;

	(void) cntxt;
	if (msg)
		return msg;

	if (mtype < 0 || mtype >= 255)
		throw(SQL, "sql.getVariable", "failed");
	src = stack_get_var(m, varname);
	if (!src) {
		char buf[BUFSIZ];
		snprintf(buf, BUFSIZ, "variable '%s' unknown", varname);
		throw(SQL, "sql.getVariable", buf);
	}
	dst = &stk->stk[getArg(pci, 0)];
	VALcopy(dst,src);
	return MAL_SUCCEED;
}

str
sql_variables(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	int i;
	mvc *m = NULL;
	BAT *vars, *r; 
	str msg = getContext(cntxt,mb, &m, NULL);
	int *res = (int *) getArgReference(stk, pci, 0);

	(void) cntxt;
	if (msg)
		return msg;

	vars = BATnew(TYPE_void,TYPE_str, m->topvars);
	BATseqbase(vars, 0);
	for (i=0; i<m->topvars && m->vars[i].s; i++) 
		BUNappend(vars, m->vars[i].name, FALSE);
	r = BATnew(TYPE_str,TYPE_bat,1);
	BUNins(r, "name", &vars->batCacheid, FALSE);
	BBPunfix(vars->batCacheid);
	*res = r->batCacheid;
	BBPkeepref(r->batCacheid);
	return MAL_SUCCEED;
}

/* str mvc_logfile(int *d, str *filename); */
str
mvc_logfile(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	int *res = (int *) getArgReference(stk, pci, 0);
	str filename = *(str *)getArgReference(stk, pci, 1);

	(void) cntxt;
	if (msg)
		return msg;
	if (m->scanner.log) {
		close_stream(m->scanner.log);
		m->scanner.log = NULL;
	}

	if (filename != str_nil) 
		m->scanner.log = open_wastream(filename);
	*res = 0;
	return MAL_SUCCEED;
}

/* str mvc_next_value(lng *res, str *sname, str *seqname); */
str
mvc_next_value(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_schema *s;
	lng *res  = (lng *) getArgReference(stk, pci, 0);
	str *sname  = (str *) getArgReference(stk, pci, 1);
	str *seqname  = (str *) getArgReference(stk, pci, 2);

	(void) cntxt;
	if (msg)
		return msg;
 	s = mvc_bind_schema(m, *sname);
	if (s) {
		sql_sequence *seq = find_sql_sequence(s, *seqname);

		if (seq && seq_next_value(seq, res)) {
			m->last_id = *res;
			stack_set_number(m, "last_id", m->last_id);
			return MAL_SUCCEED;
		}
	}
	throw(SQL, "sql.next_value", "error");
}

/* str mvc_bat_next_value(bat *res, int *sid, str *seqname); */
str
mvc_bat_next_value(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	BAT *b, *r;
	BUN p,q;
	sql_schema *s = NULL;
	sql_sequence *seq = NULL;
	seqbulk *sb = NULL;
	BATiter bi;
	bat *res  = (bat *) getArgReference(stk, pci, 0);
	int *sid  = (int *) getArgReference(stk, pci, 1);
	str *seqname  = (str *) getArgReference(stk, pci, 2);

	(void) cntxt;
	if (msg)
		return msg;

	if( (b = BATdescriptor(*sid)) == NULL )
		throw(SQL, "sql.next_value", "Cannot access descriptor");

	r = BATnew(b->htype, TYPE_lng, BATcount(b));
	if (!r) {
		BBPunfix(b->batCacheid);
		throw(SQL, "sql.next_value", "Cannot create bat");
	}
	BATseqbase(r, b->hseqbase);

	if (!BATcount(b)) {
		BBPunfix(b->batCacheid);
		BBPkeepref(r->batCacheid);
		*res = r->batCacheid;
		return MAL_SUCCEED;
	}

	bi = bat_iterator(b);
	BATloop(b,p,q){
		str sname = BUNtail(bi,BUNfirst(b));
		lng l;

		if (!s || strcmp(s->base.name, sname) != 0) {
			if (sb)
				seqbulk_destroy(sb);
			s = mvc_bind_schema(m, sname);
			seq = NULL;
			if (!s || 
			    (seq = find_sql_sequence(s, *seqname)) == NULL ||
			    !(sb = seqbulk_create(seq, BATcount(b))) ) {
				BBPunfix(b->batCacheid);
				BBPunfix(r->batCacheid);
				throw(SQL, "sql.next_value", "error");
			}
		}
		if (!seqbulk_next_value(sb, &l)) {
			BBPunfix(b->batCacheid);
			BBPunfix(r->batCacheid);
			seqbulk_destroy(sb);
			throw(SQL, "sql.next_value", "error");
		}
		BUNins(r, BUNhead(bi,p), &l, FALSE);
	}
	if (sb) 
		seqbulk_destroy(sb);
	BBPunfix(b->batCacheid);
	BBPkeepref(r->batCacheid);
	*res = r->batCacheid;
	return MAL_SUCCEED;
}

/* str mvc_get_value(lng *res, str *sname, str *seqname); */
str
mvc_get_value(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_schema *s;
	lng *res  = (lng *) getArgReference(stk, pci, 0);
	str *sname  = (str *) getArgReference(stk, pci, 1);
	str *seqname  = (str *) getArgReference(stk, pci, 2);

	(void) cntxt;
	if (msg)
		return msg;
 	s = mvc_bind_schema(m, *sname);
	if (s) {
		sql_sequence *seq = find_sql_sequence(s, *seqname);

		if (seq && seq_get_value(seq, res))
			return MAL_SUCCEED;
	}
	throw(SQL, "sql.get_value", "error");
}

str
mvc_getVersion(lng *version, int *clientid)
{
	mvc *m = NULL;
	Client cntxt = mal_clients+*clientid;
	str msg = getContext(cntxt, NULL, &m, NULL);

	if (msg)
		return msg;
	*version = -1;
	if (m->session->tr)
		*version = m->session->tr->stime;
	return MAL_SUCCEED;
}

/* str mvc_restart_seq(lng *res, str *sname, str *seqname, lng *start); */
str
mvc_restart_seq(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_schema *s;
	lng *res  = (lng *) getArgReference(stk, pci, 0);
	str *sname  = (str *) getArgReference(stk, pci, 1);
	str *seqname  = (str *) getArgReference(stk, pci, 2);
	lng *start  = (lng *) getArgReference(stk, pci, 3);

	(void) cntxt;
	if (msg)
		return msg;
 	s = mvc_bind_schema(m, *sname);
	if (s) {
		sql_sequence *seq = find_sql_sequence(s, *seqname);

		if (seq) {
			*res = sql_trans_sequence_restart(m->session->tr, seq, *start);
			return MAL_SUCCEED;
		}
	}
	throw(SQL, "sql.restart", "error");
}



BAT *
mvc_bind(mvc *m, char *sname, char *tname, char *cname, int access)
{
	sql_trans *tr = m->session->tr;
	BAT *b = NULL;
	sql_schema *s = NULL;
	sql_table *t = NULL;
	sql_column *c = NULL;

	s = mvc_bind_schema(m, sname);
	if ( s == NULL) 
		return NULL;
	t = mvc_bind_table(m, s, tname);
	if ( t == NULL) 
		return NULL;
	c = mvc_bind_column(m, t, cname);
	if ( c == NULL) 
		return NULL;

	b = store_funcs.bind_col(tr, c, access);
	return b;
}

BAT *
mvc_bind_dbat(mvc *m, char *sname, char *tname, int access)
{
	sql_trans *tr = m->session->tr;
	BAT *b = NULL;
	sql_schema *s = NULL;
	sql_table *t = NULL;
	s = mvc_bind_schema(m, sname);
	if ( s == NULL)
		return NULL;
	t = mvc_bind_table(m, s, tname);
	if ( t == NULL)
		return NULL;

	b = store_funcs.bind_del(tr, t, access);
	return b;
}

BAT *
mvc_bind_idxbat(mvc *m, char *sname, char *tname, char *iname, int access)
{
	sql_trans *tr = m->session->tr;
	BAT *b = NULL;
	sql_schema *s = mvc_bind_schema(m, sname);
	sql_idx *i = mvc_bind_idx(m, s, iname);

	(void) tname;
	b = store_funcs.bind_idx(tr, i, access);
	return b;
}

/* str mvc_bind_wrap(int *bid, str *sname, str *tname, str *cname, int *access); */
str
mvc_bind_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	BAT *b = NULL, *bn;
	int *bid = (int *)getArgReference(stk, pci, 0);
	mvc *m = *(mvc**)getArgReference(stk, pci, 1);
	str *sname = (str *)getArgReference(stk, pci, 2);
	str *tname = (str *)getArgReference(stk, pci, 3);
	str *cname = (str *)getArgReference(stk, pci, 4);
	int *access = (int *)getArgReference(stk, pci, 5);
	oid lval,hval=0;

	(void) mb;
	(void) cntxt;
	b = mvc_bind(m, *sname, *tname, *cname, *access);
	if (b) {
		if ( pci->argc == 8){
			/* partitioned access */
			lval = *(oid *)getArgReference(stk, pci, 6);
			hval = *(oid *)getArgReference(stk, pci, 7);
			bn =  BATslice(b, lval,hval);
			BATseqbase(bn, lval);
			BBPkeepref( *bid = bn->batCacheid);
			BBPreleaseref(b->batCacheid);
			return MAL_SUCCEED;
		}
		BBPkeepref( *bid = b->batCacheid);
		return MAL_SUCCEED;
	}
	throw(SQL, "sql.bind", "unable to find %s.%s(%s)", *sname, *tname, *cname);
}


/* str mvc_bind_idxbat_wrap(int *bid, str *sname, str *tname, str *iname, int *access); */
str
mvc_bind_idxbat_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	BAT *b = NULL,*bn;
	int *bid = (int *)getArgReference(stk, pci, 0);
	mvc *m = *(mvc**)getArgReference(stk, pci, 1);
	str *sname = (str *)getArgReference(stk, pci, 2);
	str *tname = (str *)getArgReference(stk, pci, 3);
	str *iname = (str *)getArgReference(stk, pci, 4);
	int *access = (int *)getArgReference(stk, pci, 5);
	oid lval,hval=0;

	(void) mb;
	(void) cntxt;
	b = mvc_bind_idxbat(m, *sname, *tname, *iname, *access);
	if (b) {
		if ( pci->argc == 8){
			/* partitioned access */
			lval = *(oid *)getArgReference(stk, pci, 6);
			hval = *(oid *)getArgReference(stk, pci, 7);
			bn =  BATslice(b, lval,hval);
			BATseqbase(bn, lval);
			BBPkeepref( *bid = bn->batCacheid);
			BBPreleaseref(b->batCacheid);
			return MAL_SUCCEED;
		}
		BBPkeepref( *bid = b->batCacheid);
		return MAL_SUCCEED;
	}
	throw(SQL, "sql.idxbind", "unable to find index %s for %s.%s",
			*iname, *sname, *tname);
}

str 
mvc_diff_idxbat_wrap(bat *bid, int *clientid, str *s, str *t, str *i, int *access, lng *oldversion, lng *version)
{
	mvc *m = NULL;
	Client cntxt = mal_clients+*clientid;
	str msg = getContext(cntxt, NULL, &m, NULL);
	BAT *b;

	if (msg)
		return msg;
	(void)oldversion;
	(void)version;
	b = mvc_bind(m, *s, *t, *i, *access);
	if (b) {
		BBPkeepref( *bid = b->batCacheid);
		return MAL_SUCCEED;
	}
	throw(SQL, "sql.idxbind", "unable to find index %s for %s.%s", *i, *s, *t);
}

/*mvc_append_wrap(int *bid, str *sname, str *tname, str *cname, ptr d) */
str
mvc_append_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	ptr *res = (ptr*)getArgReference(stk, pci, 0);
	mvc *m = *(mvc**)getArgReference(stk, pci, 1);
	str sname = *(str *)getArgReference(stk, pci, 2);
	str tname = *(str *)getArgReference(stk, pci, 3);
	str cname = *(str *)getArgReference(stk, pci, 4);
	ptr ins = (ptr)getArgReference(stk, pci, 5);
	int tpe = getArgType(mb, pci, 5);
	sql_schema *s;
	sql_table *t;
	sql_column *c;

	(void) mb;
	(void) cntxt;
	*res = (ptr)m;
	if (tpe > TYPE_any)
		tpe = TYPE_bat;
	if (tpe == TYPE_bat && (ins = BATdescriptor(*(int*)ins)) == NULL) 
		throw(SQL, "sql.append","Cannot access descriptor");
	if (ATOMextern(tpe)) 
		ins = *(ptr*)ins;
	s = mvc_bind_schema(m, sname);
	if ( s == NULL)
		throw(MAL,"sql.append","Schema missing");
	t = mvc_bind_table(m, s, tname);
	if ( t == NULL)
		throw(MAL,"sql.append","Table missing");
	if (tpe == TYPE_bat)
		BATaccessBegin((BAT*)ins, USE_ALL, MMAP_SEQUENTIAL);	
	c = mvc_bind_column(m, t, cname);
	if (c) {
		store_funcs.append_col(m->session->tr, c, ins, tpe);
	} else {
		sql_idx *i = mvc_bind_idx(m, s, cname);
		if (i)
			store_funcs.append_idx(m->session->tr, i, ins, tpe);
	}
	if (tpe == TYPE_bat) {
		BATaccessEnd((BAT*)ins, USE_ALL, MMAP_SEQUENTIAL);	
		BBPunfix(((BAT*)ins)->batCacheid);
	}
	return MAL_SUCCEED;
}

/*mvc_update_wrap(int *bid, str *sname, str *tname, str *cname, ptr d) */
str
mvc_update_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	ptr *res = (ptr*)getArgReference(stk, pci, 0);
	mvc *m = *(mvc**)getArgReference(stk, pci, 1);
	str sname = *(str *)getArgReference(stk, pci, 2);
	str tname = *(str *)getArgReference(stk, pci, 3);
	str cname = *(str *)getArgReference(stk, pci, 4);
	ptr ins = (ptr)getArgReference(stk, pci, 5);
	int tpe = getArgType(mb, pci, 5);
	sql_schema *s;
	sql_table *t;
	sql_column *c;

	(void) cntxt;
	*res = (ptr)m;
	if (tpe > TYPE_any)
		tpe = TYPE_bat;
	if( tpe == TYPE_bat && (ins = BATdescriptor(*(int*)ins)) == NULL) 
		throw(SQL, "sql.update","Cannot access descriptor");
	if (ATOMextern(tpe)) 
		ins = *(ptr*)ins;
	s = mvc_bind_schema(m, sname);
	if ( s == NULL)
		throw(MAL,"sql.update","Schema missing");
	t = mvc_bind_table(m, s, tname);
	if ( t == NULL)
		throw(MAL,"sql.update","Table missing");
	if (tpe == TYPE_bat)
		BATaccessBegin((BAT*)ins, USE_ALL, MMAP_SEQUENTIAL);	
	c = mvc_bind_column(m, t, cname);
	if (c) {
		store_funcs.update_col(m->session->tr, c, ins, tpe, 0);
	} else {
 		sql_idx *i = mvc_bind_idx(m, s, cname);
		if (i)
			store_funcs.update_idx(m->session->tr, i, ins, tpe);
	}
	if (tpe == TYPE_bat) {
		BATaccessEnd((BAT*)ins, USE_ALL, MMAP_SEQUENTIAL);	
		BBPunfix(((BAT*)ins)->batCacheid);
	}
	return MAL_SUCCEED;
}




str
mvc_bind_single_wrap(int *ret, int *bid, str *name)
{
	BAT *b, *bn;
	int r;
	BUN p;
	BATiter bni;

	if( (bn= BATdescriptor(*bid)) == NULL)
		throw(SQL, "sql.bind","Cannot access descriptor");

	p = BUNfnd(bn, *name);
	bni = bat_iterator(bn);
	r = *(int *) BUNtail(bni, p);
	if( (b= BATdescriptor(r)) == NULL)
		throw(SQL, "sql.bind","Cannot access descriptor");
	BBPkeepref(*ret= b->batCacheid);
	BBPunfix( bn->batCacheid);
	return MAL_SUCCEED;
}

/* str mvc_bind_dbat_wrap(int *ret, str *sname, str *tname, int *access); */
str
mvc_bind_dbat_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	BAT *b;
	int *ret = (int *)getArgReference(stk, pci, 0);
	mvc *m = *(mvc**)getArgReference(stk, pci, 1);
	str *sname = (str *)getArgReference(stk, pci, 2);
	str *tname = (str *)getArgReference(stk, pci, 3);
	int *access = (int*)getArgReference(stk, pci, 4);

	(void) mb;
	(void) cntxt;
	b = mvc_bind_dbat(m, *sname, *tname, *access);
	if (b) {
		BBPkeepref( *ret = b->batCacheid);
		return MAL_SUCCEED;
	}
	throw(SQL, "mvc_bind_dbat_wrap", "error");
}

str 
mvc_diff_dbat_wrap(bat *bid, int *clientid, str *s, str *t, int *access, lng *oldversion, lng *version )
{
	mvc *m = NULL;
	Client cntxt = mal_clients+*clientid;
	str msg = getContext(cntxt, NULL, &m, NULL);
	BAT *b;

	if (msg)
		return msg;
	(void)oldversion;
	(void)version;
	b = mvc_bind_dbat(m, *s, *t, *access);
	if (b) {
		BBPkeepref( *bid = b->batCacheid);
		return MAL_SUCCEED;
	}
	throw(SQL, "mvc_bind_dbat_wrap", "error");
}

/* str mvc_clear_table_wrap(wrd *res, str *sname, str *tname); */
str
mvc_clear_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	sql_schema *s;
	sql_table *t;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	wrd *res = (wrd *)getArgReference(stk, pci, 0);
	str *sname = (str *)getArgReference(stk, pci, 1);
	str *tname = (str *)getArgReference(stk, pci, 2);

	(void) cntxt;
	if (msg)
		return msg;
	s = mvc_bind_schema(m, *sname);
	if ( s == NULL)
		throw(MAL,"sql.clear_table","Schema missing");
	t = mvc_bind_table(m, s, *tname);
	if ( t == NULL)
		throw(MAL,"sql.clear_table","Table missing");
	*res = mvc_clear_table(m, t);
	return MAL_SUCCEED;
}

/*mvc_delete_wrap(int *d, str *sname, str *tname, ptr d) */
str
mvc_delete_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	ptr *res = (ptr*)getArgReference(stk, pci, 0);
	mvc *m = *(mvc**)getArgReference(stk, pci, 1);
	str sname = *(str *)getArgReference(stk, pci, 2);
	str tname = *(str *)getArgReference(stk, pci, 3);
	ptr ins = (ptr)getArgReference(stk, pci, 4);
	int tpe = getArgType(mb, pci, 4);
	BAT *b = NULL;

	sql_schema *s;
	sql_table *t;

	(void) mb;
	(void) cntxt;
	*res = (ptr)m;
	if (tpe > TYPE_any)
		tpe = TYPE_bat;
	if (tpe == TYPE_bat && (b = BATdescriptor(*(int*)ins)) == NULL) 
		throw(SQL, "sql.delete","Cannot access descriptor");
	if (tpe != TYPE_bat || (b->ttype != TYPE_oid && b->ttype != TYPE_void))
		throw(SQL, "sql.delete","Cannot access descriptor");
	s = mvc_bind_schema(m, sname);
	if ( s == NULL)
		throw(MAL,"sql.delete","Schema missing");
	t = mvc_bind_table(m, s, tname);
	if ( t == NULL)
		throw(MAL,"sql.delete","Table missing");
	store_funcs.delete_tab(m->session->tr, t, b, tpe);
	if (tpe == TYPE_bat)
		BBPunfix(((BAT*)ins)->batCacheid);
	return MAL_SUCCEED;
}

static int
mvc_result_row(mvc *m, int nr_cols, int qtype)
{
	m->results = res_table_create(m->session->tr, m->result_id++, nr_cols, qtype, m->results, NULL);
	return m->results->id;
}

/* str mvc_result_row_wrap(int *res_id, int *nr_cols, int *qtype, int *o); */
str
mvc_result_row_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	int *res_id = (int *)getArgReference(stk, pci, 0);
	int *nr_cols = (int *)getArgReference(stk, pci, 1);
	int *qtype = (int *)getArgReference(stk, pci, 2);
	int *o = (int *)getArgReference(stk, pci, 3);

	(void) cntxt;
	if (msg)
		return msg;
	(void)o; /* dummy order */
	*res_id = mvc_result_row(m, *nr_cols, *qtype); 
	if (*res_id < 0)
		throw(SQL, "sql.resultSet", "failed");
	return MAL_SUCCEED;
}

/* str mvc_result_file_wrap(int *res_id, int *nr_cols, unsigned char* *T, unsigned char* *R, unsigned char* *S, unsigned char* *N, bat *order_bid); */
str 
mvc_result_file_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	str res = MAL_SUCCEED;
	BAT *order = NULL;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	res_table *t = NULL;
	unsigned char *tsep = NULL, *rsep = NULL, *ssep = NULL, *ns = NULL;
	ssize_t len;
	int *res_id = (int *)getArgReference(stk, pci, 0);
	int *nr_cols = (int *)getArgReference(stk, pci, 1);
	unsigned char **T = (unsigned char **)getArgReference(stk, pci, 2);
	unsigned char **R = (unsigned char **)getArgReference(stk, pci, 3);
	unsigned char **S = (unsigned char **)getArgReference(stk, pci, 4);
	unsigned char **N = (unsigned char **)getArgReference(stk, pci, 5);
	int mtype = getArgType(mb, pci, 6);

	(void) cntxt;
	if (msg)
		return msg;
	if (isaBatType(mtype)) {
		bat *order_bid = (bat *)getArgReference(stk, pci, 6);
        	if ((order = BATdescriptor(*order_bid)) == NULL ) {
               		throw(SQL, "sql.resultSet", "Cannot access descriptor");
        	}
	}
	m->results = t = res_table_create(m->session->tr, m->result_id++, *nr_cols, Q_TABLE, m->results, order);
	len = strlen((char*)(*T));
	GDKstrFromStr(tsep=GDKmalloc(len+1), *T, len); len = 0;
	len = strlen((char*)(*R));
	GDKstrFromStr(rsep=GDKmalloc(len+1), *R, len); len = 0;
	len = strlen((char*)(*S));
	GDKstrFromStr(ssep=GDKmalloc(len+1), *S, len); len = 0;
	len = strlen((char*)(*N));
	GDKstrFromStr(ns=GDKmalloc(len+1), *N, len); len = 0;
	t->tsep = (char*)tsep;
	t->rsep = (char*)rsep;
	t->ssep = (char*)ssep;
	t->ns = (char*)ns;
	*res_id = t->id;
	if (*res_id < 0)
		res = createException(SQL, "sql.resultSet", "failed");
	if (order)
		BBPunfix(order->batCacheid);
	return res;
}

/* str mvc_result_table_wrap(int *res_id, int *nr_cols, int *qtype, bat *order_bid); */
str
mvc_result_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	str res = MAL_SUCCEED;
	BAT *order;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	int *res_id = (int *)getArgReference(stk, pci, 0);
	int *nr_cols = (int *)getArgReference(stk, pci, 1);
	int *qtype = (int *)getArgReference(stk, pci, 2);
	bat *order_bid = (bat *)getArgReference(stk, pci, 3);

	(void) cntxt;
	if (msg)
		return msg;
        if ((order = BATdescriptor(*order_bid)) == NULL ) {
                throw(SQL, "sql.resultSet", "Cannot access descriptor");
        }
	*res_id = mvc_result_table(m, *nr_cols, *qtype, order);
	if (*res_id < 0)
		res= createException(SQL, "sql.resultSet", "failed");
	BBPunfix(order->batCacheid);
	return res;
}

/* str mvc_result_column_wrap(int *ret, int *rs, str *tn, str *name, str *type, int *digits, int *scale, bat *bid); */
str
mvc_result_column_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	str res = MAL_SUCCEED;
	BAT *b;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	int *ret = (int *)getArgReference(stk, pci, 0);
	str *tn = (str *)getArgReference(stk, pci, 2);
	str *name = (str *)getArgReference(stk, pci, 3);
	str *type = (str *)getArgReference(stk, pci, 4);
	int *digits = (int *)getArgReference(stk, pci, 5);
	int *scale = (int *)getArgReference(stk, pci, 6);
	bat *bid = (bat *)getArgReference(stk, pci, 7);

	(void) cntxt;
	if (msg)
		return msg;
	if ((b = BATdescriptor(*bid)) == NULL)
		throw(SQL, "sql.rsColumn", "cannot access BAT descriptor");
	if (mvc_result_column(m, *tn, *name, *type, *digits, *scale, b)) 
		res = createException(SQL, "sql.rsColumn", "mvc_result_column failed");
	*ret = 0;
	BBPunfix(b->batCacheid);
	return res;
}

str
/*mvc_result_value_wrap(int *ret, int *rs, str *tn, str *name, str *type, int *digits, int *scale, ptr p, int mtype)*/
mvc_result_value_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	int *ret  = (int *) getArgReference(stk, pci, 0);
	str *tn   = (str *) getArgReference(stk, pci, 2);
	str *cn   = (str *) getArgReference(stk, pci, 3);
	str *type = (str *) getArgReference(stk, pci, 4);
	int *digits = (int *) getArgReference(stk, pci, 5);
	int *scale = (int *) getArgReference(stk, pci, 6);
	ptr p =  (ptr) getArgReference(stk, pci, 7);
	int mtype = getArgType(mb, pci, 7);
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);

	(void) cntxt;
	if (msg)
		return msg;
	if (ATOMextern(mtype)) 
		p = *(ptr*)p;
	if (mvc_result_value(m, *tn, *cn, *type, *digits, *scale, p, mtype))
		throw(SQL, "sql.rsColumn", "failed");
	*ret = 0;
	return MAL_SUCCEED;
}

/* str mvc_declared_table_wrap(int *res_id, str *name); */
str
mvc_declared_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_schema *s = NULL;
	int *res_id = (int *) getArgReference(stk, pci, 0);
	str *name = (str *) getArgReference(stk, pci, 1);

	(void) cntxt;
	if (msg)
		return msg;
	s = mvc_bind_schema(m, dt_schema);
	if ( s == NULL)
		throw(MAL,"sql.declared_table","Schema missing");
	(void)mvc_create_table(m, s, *name, tt_table, TRUE, SQL_DECLARED_TABLE, CA_DROP, 0);
	*res_id = 0;
	return MAL_SUCCEED;
}

/* str mvc_declared_table_column_wrap(int *ret, int *rs, str *tname, str *name, str *type, int *digits, int *scale); */
str 
mvc_declared_table_column_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_schema *s = NULL;
	sql_table *t = NULL;
	sql_subtype tpe;
	int *ret  = (int *) getArgReference(stk, pci, 0);
	int *rs = (int *) getArgReference(stk, pci, 1);
	str *tname = (str *) getArgReference(stk, pci, 2);
	str *name = (str *) getArgReference(stk, pci, 3);
	str *type = (str *) getArgReference(stk, pci, 4);
	int *digits = (int *) getArgReference(stk, pci, 5);
	int *scale = (int *) getArgReference(stk, pci, 6);

	(void) cntxt;
	if (msg)
		return msg;
	if (*rs != 0)
                throw(SQL, "sql.dtColumn", "Cannot access declared table");
	if (!sql_find_subtype(&tpe, *type, *digits, *scale))
                throw(SQL, "sql.dtColumn", "Cannot find column type");
	s = mvc_bind_schema(m, dt_schema);
	if ( s == NULL)
		throw(MAL,"sql.declared_table_column","Schema missing");
	t = mvc_bind_table(m, s, *tname);
	if ( t == NULL)
		throw(MAL,"sql.declared_table_column","Table missing");
	(void)mvc_create_column(m, t, *name, &tpe); 
	*ret = 0;
	return MAL_SUCCEED;
}

str
mvc_drop_declared_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci )
{
	mvc *m = NULL;
	str *name = (str *) getArgReference(stk, pci, 1);
	str msg = getContext(cntxt, mb, &m, NULL);
	sql_schema *s = NULL;
	sql_table *t = NULL;

	if (msg)
		return msg;
	s = mvc_bind_schema(m, dt_schema);
	if ( s == NULL)
		throw(MAL,"sql.drop","Schema missing");
	t = mvc_bind_table(m, s, *name);
	if ( t == NULL)
		throw(MAL,"sql.drop","Table missing");
	(void)mvc_drop_table(m, s, t, 0);
	return MAL_SUCCEED;
}

str
mvc_drop_declared_tables_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci )
{
	mvc *m = NULL;
	int i = *(int *) getArgReference(stk, pci, 1);
	str msg = getContext(cntxt, mb, &m, NULL);
	sql_schema *s = NULL;
	sql_table *t = NULL;

	if (msg)
		return msg;
	s = mvc_bind_schema(m, dt_schema);
	if ( s == NULL)
		throw(MAL,"sql.drop","Schema missing");
	while(i && s->tables.set->t) {
		t = s->tables.set->t->data;
		(void)mvc_drop_table(m, s, t, 0);
		i--;
	}
	return MAL_SUCCEED;
}

/* str mvc_affected_rows_wrap(int *r, wrd *nr, str *w); */
str
mvc_affected_rows_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	backend *b = NULL;
	mvc *c = NULL;
	str msg = getContext(cntxt,mb, &c, &b);
	int *r  = (int *) getArgReference(stk, pci, 0);
#ifndef NDEBUG
	int mtype = getArgType(mb, pci, 1);
#endif
	wrd nr;
	str *w = (str *) getArgReference(stk, pci, 2);

	(void) cntxt;
	if (msg)
		return msg;
	assert (mtype == TYPE_wrd);
	nr = *(wrd *) getArgReference(stk, pci, 1);
	if (mvc_export_affrows(c, b->out, nr, *w)) {
		throw(SQL, "sql.affectedRows", "failed");
	}
	(void)r;
	return NULL;
}

/* str mvc_export_head_wrap(int *ret, stream **s, int *res_id); */
str
mvc_export_head_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	backend *b = NULL;
	mvc *c = NULL;
	str msg = getContext(cntxt,mb, &c, &b);
	int *ret  = (int *) getArgReference(stk, pci, 0);
	stream **s  = (stream **) getArgReference(stk, pci, 1);
	int *res_id = (int *) getArgReference(stk, pci, 2);

	(void) cntxt;
	if (msg)
		return msg;
	if (mvc_export_head(c, *s, *res_id, FALSE)) {
		throw(SQL, "sql.exportHead", "failed");
	}
	(void)ret;
	return NULL;
}
/* str mvc_export_result_wrap(int *ret, stream **s, int *res_id); */
str
mvc_export_result_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	backend *b = NULL;
	mvc *c = NULL;
	str msg = getContext(cntxt,mb, &c, &b);
	int *ret  = (int *) getArgReference(stk, pci, 0);
	stream **s  = (stream **) getArgReference(stk, pci, 1);
	int *res_id = (int *) getArgReference(stk, pci, 2);

	(void) cntxt;
	if (msg)
		return msg;
	if (mvc_export_result(c, *s, *res_id)) {
		throw(SQL, "sql.exportResult", "failed");
	}
	(void)ret;
	return NULL;
}
/* str mvc_export_chunk_wrap(int *ret, stream **s, int *res_id, str *w); */
str
mvc_export_chunk_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	backend *b = NULL;
	mvc *c = NULL;
	str msg = getContext(cntxt,mb, &c, &b);
	int *ret  = (int *) getArgReference(stk, pci, 0);
	stream **s  = (stream **) getArgReference(stk, pci, 1);
	int *res_id = (int *) getArgReference(stk, pci, 2);
	BUN	offset = 0;
	BUN	nr = 0;

	if ( pci->argc == 5){
		offset = *(BUN *) getArgReference(stk,pci,3);
		nr = *(BUN *) getArgReference(stk,pci,4);
	} 

	(void) cntxt;
	if (msg)
		return msg;
	if (mvc_export_chunk(c, *s, *res_id, offset, nr)) {
		throw(SQL, "sql.exportChunk", "failed");
	}
	(void)ret;
	return NULL;
}

/* str mvc_export_operation_wrap(int *ret, str *w); */
str
mvc_export_operation_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	backend *b = NULL;
	mvc *c = NULL;
	str msg = getContext(cntxt,mb, &c, &b);
	int *ret  = (int *) getArgReference(stk, pci, 0);
	str *w = (str *) getArgReference(stk, pci, 1);

	(void) cntxt;
	if (msg)
		return msg;
	if (mvc_export_operation(c, b->out, *w)) {
		throw(SQL, "sql.exportOperation", "failed");
	}
	(void)ret;
	return NULL;
}


str
/*mvc_export_value_wrap(int *ret, int *qtype, str tn, str name, str type, int *digits, int *scale, int *eclass, ptr p, int mtype)*/
mvc_export_value_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	int *qtype= (int *) getArgReference(stk, pci, 1);
	str *tn   = (str *) getArgReference(stk, pci, 2);
	str *cn   = (str *) getArgReference(stk, pci, 3);
	str *type = (str *) getArgReference(stk, pci, 4);
	int *digits = (int *) getArgReference(stk, pci, 5);
	int *scale = (int *) getArgReference(stk, pci, 6);
	int *eclass = (int *) getArgReference(stk, pci, 7);
	ptr p =  (ptr) getArgReference(stk, pci, 8);
	int mtype = getArgType(mb, pci, 8);
	str *w = (str *) getArgReference(stk, pci, 9);
	backend *b = NULL;
	mvc *c = NULL;
	str msg = getContext(cntxt,mb, &c, &b);

	(void) cntxt;
	if (msg)
		return msg;
	if (ATOMextern(mtype)) 
		p = *(ptr*)p;
	if (b->out == NULL ||
		mvc_export_value(c, b->out, *qtype, *tn, *cn, *type, *digits, *scale, *eclass, p, mtype, *w, "NULL") != SQL_OK)
		throw(SQL, "sql.exportValue", "failed");
	return MAL_SUCCEED;
}

/* str mvc_import_table_wrap(int *res, bstream **s, str *sname, str *tname, unsigned char* *T, unsigned char* *R, unsigned char* *S, unsigned char* *N, lng *sz, lng *offset); */
str
mvc_import_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	BAT *b;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	unsigned char *tsep = NULL, *rsep = NULL, *ssep = NULL, *ns = NULL;
	ssize_t len = 0;
	int *res = (int *) getArgReference(stk, pci, 0);
	bstream **s = (bstream **) getArgReference(stk, pci, 1);
	str *sname = (str *) getArgReference(stk, pci, 2);
	str *tname = (str *) getArgReference(stk, pci, 3);
	unsigned char **T = (unsigned char **) getArgReference(stk, pci, 4);
	unsigned char **R = (unsigned char **) getArgReference(stk, pci, 5);
	unsigned char **S = (unsigned char **) getArgReference(stk, pci, 6);
	unsigned char **N = (unsigned char **) getArgReference(stk, pci, 7);
	lng *sz = (lng *) getArgReference(stk, pci, 8);
	lng *offset = (lng *) getArgReference(stk, pci, 9);

	if (msg)
		return msg;
	len = strlen((char*)(*T));
	GDKstrFromStr(tsep=GDKmalloc(len+1), *T, len); len = 0;
	len = strlen((char*)(*R));
	GDKstrFromStr(rsep=GDKmalloc(len+1), *R, len); len = 0;
	if (*S && strcmp(str_nil,*(char**)S)) {
		len = strlen((char*)(*S));
		GDKstrFromStr(ssep=GDKmalloc(len+1), *S, len); 
		len = 0;
	}
	len = strlen((char*)(*N));
	GDKstrFromStr(ns=GDKmalloc(len+1), *N, len); len = 0;
	b = mvc_import_table(cntxt, m, *s, *sname, *tname, (char*)tsep, (char*)rsep, (char*)ssep, (char*)ns, *sz, *offset);
	GDKfree(tsep);
	GDKfree(rsep);
	if(ssep)
		GDKfree(ssep);
	GDKfree(ns);
	if (!b)
		throw(SQL, "importTable", "%sfailed to import table", m->errstr);
	*res = b->batCacheid;
	BBPincref( *res, TRUE);
	BBPunfix(*res);
	return MAL_SUCCEED;
}

/* str mvc_bin_import_table_wrap(int *res, str *sname, str *tname, str *fname..); */
str
mvc_bin_import_table_wrap(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	BAT *b;
	mvc *m = NULL;
	str msg = getContext(cntxt, mb, &m, NULL);
	BUN cnt = 0;
	int i, *res = (int *) getArgReference(stk, pci, 0);
	str sname = *(str *) getArgReference(stk, pci, 1);
	str tname = *(str *) getArgReference(stk, pci, 2);
	sql_schema *s = mvc_bind_schema(m, sname);
	sql_table *t;
	node *n;

	if (msg)
		return msg;

	if ( s == NULL)
		throw(MAL,"sql.drop","Schema missing");
	t = mvc_bind_table(m, s, tname);
	if (!t) 
		throw(SQL,"sql", "table %s not found", tname);
	if (list_length(t->columns.set) != (pci->argc-3))
		throw(SQL,"sql", "Not enough columns in found");
	b = BATnew(TYPE_str, TYPE_bat, pci->argc-3);
	for (i = 3, n = t->columns.set->h; i<pci->argc && n; i++, n = n->next) {
		sql_column *col = n->data;
		BAT *c;

		c = BATattach(col->type.type->localtype, *(str*)getArgReference(stk, pci, i));
		if (c == NULL) {
			BBPunfix(b->batCacheid);
			throw(SQL, "sql", "failed to attach file %s",
			      *(str*)getArgReference(stk, pci, i));
		}

		BATsetaccess(c, BAT_READ);
		BATpropcheck(c, BATPROPS_ALL);
		BATpropcheck(BATmirror(c), BATPROPS_ALL);
		if (i!=3 && cnt != BATcount(c)) {
			BBPunfix(b->batCacheid);
			throw(SQL, "sql", "table %s not found", tname);
		}
		cnt = BATcount(c);
		BUNins(b, (ptr)col->base.name, (ptr)&c->batCacheid, FALSE);
		BBPunfix( c->batCacheid );
	}
	*res = b->batCacheid;
	BBPincref( *res, TRUE);
	BBPunfix(*res);
	return MAL_SUCCEED;
}


str
zero_or_one(ptr ret, int *bid) 
{
	BAT *b;
	BUN c, _s;
	ptr p;

	if ((b = BATdescriptor(*bid)) == NULL) {
		throw(SQL, "zero_or_one", "Cannot access descriptor");
	}
	c = BATcount(b);
	if (c == 0) {
		p = ATOMnilptr(b->ttype);
	} else if (c == 1) {
		BATiter bi = bat_iterator(b);
		p = BUNtail(bi,BUNfirst(b));
	} else {
		char buf[BUFSIZ];

		p = NULL;
                snprintf(buf, BUFSIZ, "cardinality violation (" BUNFMT ">1)", c);
		throw(SQL, "zero_or_one", buf);
	}
	_s = ATOMsize(ATOMtype(b->ttype));
	if (ATOMextern(b->ttype)) {
		_s = ATOMlen(ATOMtype(b->ttype), p);
		memcpy(*(ptr*) ret=GDKmalloc(_s), p, _s);
	} else if (b->ttype == TYPE_bat) {
		bat bid = *(bat*)p;
		*(BAT**) ret = BATdescriptor(bid);
	} else if (_s == 4) {
		*(int*) ret = *(int*)p;
	} else if (_s == 1) {
		*(bte*) ret = *(bte*)p;
	} else if (_s == 2) {
		*(sht*) ret = *(sht*)p;
	} else if (_s == 8) {
		*(lng*) ret = *(lng*)p;
	} else {
		memcpy(ret, p, _s);
	}
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}

str not_unique(bit *ret, int* bid) 
{
	BAT *b;

	if ((b = BATdescriptor(*bid)) == NULL) {
		throw(SQL, "not_unique", "Cannot access descriptor");
	}

	*ret = FALSE;
	if (BATtkey(b) || BATtdense(b) || BATcount(b) <= 1) {
		BBPunfix(b->batCacheid);
		return MAL_SUCCEED;
	} else if (b->tsorted&1) {
		BUN p,q;
		oid c = *(oid*)Tloc(b, BUNfirst(b)); 

		for(p=BUNfirst(b)+1, q=BUNlast(b);p<q;p++) {
			oid v = *(oid*)Tloc(b,p);
			if (v<=c) {
				*ret = TRUE;
				break;
			}
			c = v;
		}
	} else {
		BBPunfix(b->batCacheid);
		throw(SQL, "not_unique", "input should be sorted");
	}
	BBPunfix(b->batCacheid);
	return MAL_SUCCEED;
}

/* later we could optimize this to start from current BUN 
   And only search the from the first if second is not found.
 */
static inline int 
HASHfndTwice(BAT *b, ptr v)
{
	BATiter bi = bat_iterator(b);
	BUN i = BUN_NONE;
	int first = 1;
	
	HASHloop( bi, b->H->hash, i, v) {
		if (!first)
			return 1;
		first = 0;
	}
	return 0;
}

str 
not_unique_oids(bat *ret, bat* bid) 
{
	BAT *b, *bn = NULL;

	if ((b = BATdescriptor(*bid)) == NULL) {
		throw(SQL, "not_uniques", "Cannot access descriptor");
	}
	if (b->ttype != TYPE_oid && b->ttype != TYPE_wrd) {
		throw(SQL, "not_uniques", "Wrong types");
	}

	assert(b->htype == TYPE_oid);
	if (BATtkey(b) || BATtdense(b) || BATcount(b) <= 1) {
		bn = BATnew(TYPE_oid, TYPE_oid, 0);
	} else if (b->tsorted&1) { /* ugh handle both wrd and oid types */
		oid c = *(oid*)Tloc(b, BUNfirst(b)), *rf, *rh, *rt;
		oid *h = (oid*)Hloc(b,0), *vp, *ve;
		int first = 1;

		bn = BATnew(TYPE_oid, TYPE_oid, BATcount(b));
		vp = (oid*)Tloc(b, BUNfirst(b));
		ve = vp + BATcount(b);
		rf = rh = (oid*)Hloc(bn, BUNfirst(bn));
		rt = (oid*)Tloc(bn, BUNfirst(bn));
		*rh++ = *h++; 
		*rt++ = *vp; 
		for(vp++; vp < ve; vp++, h++) {
			oid v = *vp;
			if (v == c) {
				first = 0;
				*rh++ = *h;
				*rt++ = v;
			} else if (!first) {
				first = 1;
				*rh++ = *h;
				*rt++ = v;
			} else {
				*rh = *h;
				*rt = v;
			}
			c = v;
		}
		if (first)
			rh--;
		BATsetcount(bn, (BUN)(rh - rf)); 
	} else {
		oid *rf, *rh, *rt;
		oid *h = (oid*)Hloc(b,0), *vp, *ve;
		BAT *bm = BATmirror(b);

		if (BATprepareHash(bm))
			throw(SQL, "not_uniques", "hash creation failed");
		bn = BATnew(TYPE_oid, TYPE_oid, BATcount(b));
		vp = (oid*)Tloc(b, BUNfirst(b));
		ve = vp + BATcount(b);
		rf = rh = (oid*)Hloc(bn, BUNfirst(bn));
		rt = (oid*)Tloc(bn, BUNfirst(bn));
		for(; vp < ve; vp++, h++) {
			/* try to find value twice */
			if (HASHfndTwice(bm, vp)) {
				*rh++ = *h;
				*rt++ = *vp;
			}
		}
		BATsetcount(bn, (BUN)(rh - rf)); 
	}
	BBPunfix(b->batCacheid);
	BBPkeepref(*ret = bn->batCacheid);
	return MAL_SUCCEED;
}

lng scales[20] = {
	LL_CONSTANT(1),
	LL_CONSTANT(10),
	LL_CONSTANT(100),
	LL_CONSTANT(1000),
	LL_CONSTANT(10000),
	LL_CONSTANT(100000),
	LL_CONSTANT(1000000),
	LL_CONSTANT(10000000),
	LL_CONSTANT(100000000),
	LL_CONSTANT(1000000000),
	LL_CONSTANT(10000000000),
	LL_CONSTANT(100000000000),
	LL_CONSTANT(1000000000000),
	LL_CONSTANT(10000000000000),
	LL_CONSTANT(100000000000000),
	LL_CONSTANT(1000000000000000),
	LL_CONSTANT(10000000000000000),
	LL_CONSTANT(100000000000000000),
	LL_CONSTANT(1000000000000000000)
};





str 
bte_dec_round_wrap( bte *res, bte *v, bte *r )
{
	bte val = *v;
	bte add = *r;

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = bte_nil;
		return NULL;
	}

	add >>=1;
	if (val < 0) 
		add = -add;
	val += add; 
	val /= *r;
	*res = val;
	return NULL;
}

str 
bte_round_wrap( bte *res, bte *v, int *d, int *s, bte *r )
{
	/* shortcut nil */
	if (*v == bte_nil) {
		*res = bte_nil;
	} else if (-*r > *d) {
		*res = 0;
	} else if (*r > 0 && *r < *s) {
		int dff = *s - *r;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_bte_min < lres && lres <= (lng) GDK_bte_max);
		*res = (bte) lres;
	} else if (*r <= 0 && -*r + *s > 0) {
		int dff = -*r + *s;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_bte_min < lres && lres <= (lng) GDK_bte_max);
		*res = (bte) lres;
	} else {
		*res = *v;
	}
	return MAL_SUCCEED;
}

str
str_2dec_bte( bte *res, str *val, int *d, int *sc )
{
	char *s = strip_extra_zeros(*val);
	char *dot = strchr(s, '.');
	int digits = _strlen(s) - 1;
	int scale = digits - (int) (dot-s);
	lng value = 0;

	if (!dot) {
		if (GDK_STRNIL(*val)) {
			*res = bte_nil;
			return MAL_SUCCEED;
		} else {
			throw(SQL, "bte", 
				"\"%s\" is no decimal value (doesn't contain a '.')", *val);
		}
	}
		
	value = decimal_from_str(s);
	if (*s == '+' || *s == '-')
		digits --;
	if (scale < *sc){
		/* the current scale is too small, increase it by adding 0's */
		int d = *sc - scale;	/* CANNOT be 0! */

		value *= scales[d];
		scale += d;
		digits += d;
	} else if (scale > *sc){
		/* the current scale is too big, decrease it by correctly rounding */
		int d = scale - *sc;	/* CANNOT be 0 */
		lng rnd = scales[d]>>1;

		value += rnd;
		value /= scales[d];
		scale -= d;
		digits -= d;
	}
	if (digits > *d) {
		throw(SQL,"bte", 
		     "decimal (%s) doesn't have format (%d.%d)", *val, *d, *sc);
	}
	*res = (bte) value;
	return MAL_SUCCEED;
}

str
str_2num_bte( bte *res, str *v, int *len )
{
	int zero = 0;
	return str_2dec_bte( res, v, len, &zero );
}

str
batstr_2dec_bte( int *res, int *bid, int *d, int *sc )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		bte r;
		msg = str_2dec_bte( &r, &v, d, sc );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str
batstr_2num_bte( int *res, int *bid, int *len )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2num_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		bte r;
		msg = str_2num_bte( &r, &v, len );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str 
sht_dec_round_wrap( sht *res, sht *v, sht *r )
{
	sht val = *v;
	sht add = *r;

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = sht_nil;
		return NULL;
	}

	add >>=1;
	if (val < 0) 
		add = -add;
	val += add; 
	val /= *r;
	*res = val;
	return NULL;
}

str 
sht_round_wrap( sht *res, sht *v, int *d, int *s, bte *r )
{
	/* shortcut nil */
	if (*v == sht_nil) {
		*res = sht_nil;
	} else if (-*r > *d) {
		*res = 0;
	} else if (*r > 0 && *r < *s) {
		int dff = *s - *r;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_sht_min < lres && lres <= (lng) GDK_sht_max);
		*res = (sht) lres;
	} else if (*r <= 0 && -*r + *s > 0) {
		int dff = -*r + *s;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_sht_min < lres && lres <= (lng) GDK_sht_max);
		*res = (sht) lres;
	} else {
		*res = *v;
	}
	return MAL_SUCCEED;
}

str
str_2dec_sht( sht *res, str *val, int *d, int *sc )
{
	char *s = strip_extra_zeros(*val);
	char *dot = strchr(s, '.');
	int digits = _strlen(s) - 1;
	int scale = digits - (int) (dot-s);
	lng value = 0;

	if (!dot) {
		if (GDK_STRNIL(*val)) {
			*res = sht_nil;
			return MAL_SUCCEED;
		} else {
			throw(SQL, "sht", 
				"\"%s\" is no decimal value (doesn't contain a '.')", *val);
		}
	}
		
	value = decimal_from_str(s);
	if (*s == '+' || *s == '-')
		digits --;
	if (scale < *sc){
		/* the current scale is too small, increase it by adding 0's */
		int d = *sc - scale;	/* CANNOT be 0! */

		value *= scales[d];
		scale += d;
		digits += d;
	} else if (scale > *sc){
		/* the current scale is too big, decrease it by correctly rounding */
		int d = scale - *sc;	/* CANNOT be 0 */
		lng rnd = scales[d]>>1;

		value += rnd;
		value /= scales[d];
		scale -= d;
		digits -= d;
	}
	if (digits > *d) {
		throw(SQL,"sht", 
		     "decimal (%s) doesn't have format (%d.%d)", *val, *d, *sc);
	}
	*res = (sht) value;
	return MAL_SUCCEED;
}

str
str_2num_sht( sht *res, str *v, int *len )
{
	int zero = 0;
	return str_2dec_sht( res, v, len, &zero );
}

str
batstr_2dec_sht( int *res, int *bid, int *d, int *sc )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		sht r;
		msg = str_2dec_sht( &r, &v, d, sc );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str
batstr_2num_sht( int *res, int *bid, int *len )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2num_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		sht r;
		msg = str_2num_sht( &r, &v, len );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str 
int_dec_round_wrap( int *res, int *v, int *r )
{
	int val = *v;
	int add = *r;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = int_nil;
		return NULL;
	}

	add >>=1;
	if (val < 0) 
		add = -add;
	val += add; 
	val /= *r;
	*res = val;
	return NULL;
}

str 
int_round_wrap( int *res, int *v, int *d, int *s, bte *r )
{
	/* shortcut nil */
	if (*v == int_nil) {
		*res = int_nil;
	} else if (-*r > *d) {
		*res = 0;
	} else if (*r > 0 && *r < *s) {
		int dff = *s - *r;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_int_min < lres && lres <= (lng) GDK_int_max);
		*res = (int) lres;
	} else if (*r <= 0 && -*r + *s > 0) {
		int dff = -*r + *s;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_int_min < lres && lres <= (lng) GDK_int_max);
		*res = (int) lres;
	} else {
		*res = *v;
	}
	return MAL_SUCCEED;
}

str
str_2dec_int( int *res, str *val, int *d, int *sc )
{
	char *s = strip_extra_zeros(*val);
	char *dot = strchr(s, '.');
	int digits = _strlen(s) - 1;
	int scale = digits - (int) (dot-s);
	lng value = 0;

	if (!dot) {
		if (GDK_STRNIL(*val)) {
			*res = int_nil;
			return MAL_SUCCEED;
		} else {
			throw(SQL, "int", 
				"\"%s\" is no decimal value (doesn't contain a '.')", *val);
		}
	}
		
	value = decimal_from_str(s);
	if (*s == '+' || *s == '-')
		digits --;
	if (scale < *sc){
		/* the current scale is too small, increase it by adding 0's */
		int d = *sc - scale;	/* CANNOT be 0! */

		value *= scales[d];
		scale += d;
		digits += d;
	} else if (scale > *sc){
		/* the current scale is too big, decrease it by correctly rounding */
		int d = scale - *sc;	/* CANNOT be 0 */
		lng rnd = scales[d]>>1;

		value += rnd;
		value /= scales[d];
		scale -= d;
		digits -= d;
	}
	if (digits > *d) {
		throw(SQL,"int", 
		     "decimal (%s) doesn't have format (%d.%d)", *val, *d, *sc);
	}
	*res = (int) value;
	return MAL_SUCCEED;
}

str
str_2num_int( int *res, str *v, int *len )
{
	int zero = 0;
	return str_2dec_int( res, v, len, &zero );
}

str
batstr_2dec_int( int *res, int *bid, int *d, int *sc )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		int r;
		msg = str_2dec_int( &r, &v, d, sc );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str
batstr_2num_int( int *res, int *bid, int *len )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2num_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		int r;
		msg = str_2num_int( &r, &v, len );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str 
wrd_dec_round_wrap( wrd *res, wrd *v, wrd *r )
{
	wrd val = *v;
	wrd add = *r;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = wrd_nil;
		return NULL;
	}

	add >>=1;
	if (val < 0) 
		add = -add;
	val += add; 
	val /= *r;
	*res = val;
	return NULL;
}

str 
wrd_round_wrap( wrd *res, wrd *v, int *d, int *s, bte *r )
{
	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = wrd_nil;
	} else if (-*r > *d) {
		*res = 0;
	} else if (*r > 0 && *r < *s) {
		int dff = *s - *r;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_wrd_min < lres && lres <= (lng) GDK_wrd_max);
		*res = (wrd) lres;
	} else if (*r <= 0 && -*r + *s > 0) {
		int dff = -*r + *s;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_wrd_min < lres && lres <= (lng) GDK_wrd_max);
		*res = (wrd) lres;
	} else {
		*res = *v;
	}
	return MAL_SUCCEED;
}

str
str_2dec_wrd( wrd *res, str *val, int *d, int *sc )
{
	char *s = strip_extra_zeros(*val);
	char *dot = strchr(s, '.');
	int digits = _strlen(s) - 1;
	int scale = digits - (int) (dot-s);
	lng value = 0;

	if (!dot) {
		if (GDK_STRNIL(*val)) {
			*res = wrd_nil;
			return MAL_SUCCEED;
		} else {
			throw(SQL, "wrd", 
				"\"%s\" is no decimal value (doesn't contain a '.')", *val);
		}
	}
		
	value = decimal_from_str(s);
	if (*s == '+' || *s == '-')
		digits --;
	if (scale < *sc){
		/* the current scale is too small, increase it by adding 0's */
		int d = *sc - scale;	/* CANNOT be 0! */

		value *= scales[d];
		scale += d;
		digits += d;
	} else if (scale > *sc){
		/* the current scale is too big, decrease it by correctly rounding */
		int d = scale - *sc;	/* CANNOT be 0 */
		lng rnd = scales[d]>>1;

		value += rnd;
		value /= scales[d];
		scale -= d;
		digits -= d;
	}
	if (digits > *d) {
		throw(SQL,"wrd", 
		     "decimal (%s) doesn't have format (%d.%d)", *val, *d, *sc);
	}
	*res = (wrd) value;
	return MAL_SUCCEED;
}

str
str_2num_wrd( wrd *res, str *v, int *len )
{
	int zero = 0;
	return str_2dec_wrd( res, v, len, &zero );
}

str
batstr_2dec_wrd( int *res, int *bid, int *d, int *sc )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		wrd r;
		msg = str_2dec_wrd( &r, &v, d, sc );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str
batstr_2num_wrd( int *res, int *bid, int *len )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2num_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		wrd r;
		msg = str_2num_wrd( &r, &v, len );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str 
lng_dec_round_wrap( lng *res, lng *v, lng *r )
{
	lng val = *v;
	lng add = *r;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = lng_nil;
		return NULL;
	}

	add >>=1;
	if (val < 0) 
		add = -add;
	val += add; 
	val /= *r;
	*res = val;
	return NULL;
}

str 
lng_round_wrap( lng *res, lng *v, int *d, int *s, bte *r )
{
	/* shortcut nil */
	if (*v == lng_nil) {
		*res = lng_nil;
	} else if (-*r > *d) {
		*res = 0;
	} else if (*r > 0 && *r < *s) {
		int dff = *s - *r;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_lng_min < lres && lres <= (lng) GDK_lng_max);
		*res = (lng) lres;
	} else if (*r <= 0 && -*r + *s > 0) {
		int dff = -*r + *s;
		lng rnd = scales[dff]>>1;
		lng lres;
		if (*v > 0)
			lres = (((*v + rnd)/scales[dff])*scales[dff]);
		else
			lres = (((*v - rnd)/scales[dff])*scales[dff]);
		assert((lng) GDK_lng_min < lres && lres <= (lng) GDK_lng_max);
		*res = (lng) lres;
	} else {
		*res = *v;
	}
	return MAL_SUCCEED;
}

str
str_2dec_lng( lng *res, str *val, int *d, int *sc )
{
	char *s = strip_extra_zeros(*val);
	char *dot = strchr(s, '.');
	int digits = _strlen(s) - 1;
	int scale = digits - (int) (dot-s);
	lng value = 0;

	if (!dot) {
		if (GDK_STRNIL(*val)) {
			*res = lng_nil;
			return MAL_SUCCEED;
		} else {
			throw(SQL, "lng", 
				"\"%s\" is no decimal value (doesn't contain a '.')", *val);
		}
	}
		
	value = decimal_from_str(s);
	if (*s == '+' || *s == '-')
		digits --;
	if (scale < *sc){
		/* the current scale is too small, increase it by adding 0's */
		int d = *sc - scale;	/* CANNOT be 0! */

		value *= scales[d];
		scale += d;
		digits += d;
	} else if (scale > *sc){
		/* the current scale is too big, decrease it by correctly rounding */
		int d = scale - *sc;	/* CANNOT be 0 */
		lng rnd = scales[d]>>1;

		value += rnd;
		value /= scales[d];
		scale -= d;
		digits -= d;
	}
	if (digits > *d) {
		throw(SQL,"lng", 
		     "decimal (%s) doesn't have format (%d.%d)", *val, *d, *sc);
	}
	*res = (lng) value;
	return MAL_SUCCEED;
}

str
str_2num_lng( lng *res, str *v, int *len )
{
	int zero = 0;
	return str_2dec_lng( res, v, len, &zero );
}

str
batstr_2dec_lng( int *res, int *bid, int *d, int *sc )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		lng r;
		msg = str_2dec_lng( &r, &v, d, sc );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str
batstr_2num_lng( int *res, int *bid, int *len )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2num_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		lng r;
		msg = str_2num_lng( &r, &v, len );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
flt_dec_round_wrap( flt *res, flt *v, flt *r )
{
	flt val = *v;

	/* not nil */
	if (ATOMcmp(TYPE_flt, v, ATOMnilptr(TYPE_flt)) != 0) {
		val /= *r;
	}
	*res = val;
	return MAL_SUCCEED;
}

str 
flt_round_wrap( flt *res, flt *v, bte *r )
{
	/* shortcut nil */
	if (*v == flt_nil) {
		*res = flt_nil;
	} else if (*r < 0) {
		int d = -*r;
		flt rnd = (flt) (scales[d]>>1); 

		if (*v > 0)
			*res = (flt) (floor(((*v + rnd)/((flt)(scales[d]))))*scales[d]);
		else
			*res = (flt) (floor(((*v + rnd)/((flt)(scales[d]))))*scales[d]);
	} else if (*r > 0) {
		int d = *r;

		if (*v > 0)
			*res = (flt) (floor(*v*(flt)scales[d]+.5)/scales[d]);
		else
			*res = (flt) (floor(*v*(flt)scales[d]+.5)/scales[d]);
	} else {
		*res = *v;
	}
	return MAL_SUCCEED;
}





str 
dbl_dec_round_wrap( dbl *res, dbl *v, dbl *r )
{
	dbl val = *v;

	/* not nil */
	if (ATOMcmp(TYPE_dbl, v, ATOMnilptr(TYPE_dbl)) != 0) {
		val /= *r;
	}
	*res = val;
	return MAL_SUCCEED;
}

str 
dbl_round_wrap( dbl *res, dbl *v, bte *r )
{
	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = dbl_nil;
	} else if (*r < 0) {
		int d = -*r;
		dbl rnd = (dbl) (scales[d]>>1); 

		if (*v > 0)
			*res = (dbl) (floor(((*v + rnd)/((dbl)(scales[d]))))*scales[d]);
		else
			*res = (dbl) (floor(((*v + rnd)/((dbl)(scales[d]))))*scales[d]);
	} else if (*r > 0) {
		int d = *r;

		if (*v > 0)
			*res = (dbl) (floor(*v*(dbl)scales[d]+.5)/scales[d]);
		else
			*res = (dbl) (floor(*v*(dbl)scales[d]+.5)/scales[d]);
	} else {
		*res = *v;
	}
	return MAL_SUCCEED;
}





#if SIZEOF_WRD == SIZEOF_INT
#define wrdToStr(sptr, lptr, p) intToStr(sptr, lptr, (int*)p)
#else
#define wrdToStr(sptr, lptr, p) lngToStr(sptr, lptr, (lng*)p)
#endif




str
str_2_bit( bit *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_bit, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_bit, p, ATOMnilptr(TYPE_bit)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "bit", buf);
	}
	 *res = *(bit*)p;
	if (!ATOMextern(TYPE_bit)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLbit_2_str( str *res, bit *val )
{
	char *p = NULL;
	int len = 0;
	bitToStr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_bit( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_bit", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bit, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		bit r;
		msg = str_2_bit( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_oid( oid *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_oid, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_oid, p, ATOMnilptr(TYPE_oid)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "oid", buf);
	}
	 *res = *(oid*)p;
	if (!ATOMextern(TYPE_oid)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLoid_2_str( str *res, oid *val )
{
	char *p = NULL;
	int len = 0;
	OIDtoStr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_oid( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_oid", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_oid, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		oid r;
		msg = str_2_oid( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_bte( bte *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_bte, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_bte, p, ATOMnilptr(TYPE_bte)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "bte", buf);
	}
	 *res = *(bte*)p;
	if (!ATOMextern(TYPE_bte)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLbte_2_str( str *res, bte *val )
{
	char *p = NULL;
	int len = 0;
	bteToStr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_bte( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		bte r;
		msg = str_2_bte( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_sht( sht *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_sht, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_sht, p, ATOMnilptr(TYPE_sht)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "sht", buf);
	}
	 *res = *(sht*)p;
	if (!ATOMextern(TYPE_sht)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLsht_2_str( str *res, sht *val )
{
	char *p = NULL;
	int len = 0;
	shtToStr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_sht( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		sht r;
		msg = str_2_sht( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_int( int *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_int, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_int, p, ATOMnilptr(TYPE_int)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "int", buf);
	}
	 *res = *(int*)p;
	if (!ATOMextern(TYPE_int)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLint_2_str( str *res, int *val )
{
	char *p = NULL;
	int len = 0;
	intToStr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_int( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		int r;
		msg = str_2_int( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_lng( lng *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_lng, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_lng, p, ATOMnilptr(TYPE_lng)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "lng", buf);
	}
	 *res = *(lng*)p;
	if (!ATOMextern(TYPE_lng)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLlng_2_str( str *res, lng *val )
{
	char *p = NULL;
	int len = 0;
	lngToStr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_lng( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		lng r;
		msg = str_2_lng( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_wrd( wrd *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_wrd, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_wrd, p, ATOMnilptr(TYPE_wrd)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "wrd", buf);
	}
	 *res = *(wrd*)p;
	if (!ATOMextern(TYPE_wrd)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLwrd_2_str( str *res, wrd *val )
{
	char *p = NULL;
	int len = 0;
	wrdToStr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_wrd( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		wrd r;
		msg = str_2_wrd( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_flt( flt *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_flt, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_flt, p, ATOMnilptr(TYPE_flt)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "flt", buf);
	}
	 *res = *(flt*)p;
	if (!ATOMextern(TYPE_flt)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLflt_2_str( str *res, flt *val )
{
	char *p = NULL;
	int len = 0;
	fltToStr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_flt( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		flt r;
		msg = str_2_flt( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_dbl( dbl *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_dbl, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_dbl, p, ATOMnilptr(TYPE_dbl)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "dbl", buf);
	}
	 *res = *(dbl*)p;
	if (!ATOMextern(TYPE_dbl)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLdbl_2_str( str *res, dbl *val )
{
	char *p = NULL;
	int len = 0;
	dblToStr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_dbl( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		dbl r;
		msg = str_2_dbl( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_timestamp( timestamp *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_timestamp, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_timestamp, p, ATOMnilptr(TYPE_timestamp)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "timestamp", buf);
	}
	 *res = *(timestamp*)p;
	if (!ATOMextern(TYPE_timestamp)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLtimestamp_2_str( str *res, timestamp *val )
{
	char *p = NULL;
	int len = 0;
	timestamp_tostr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_timestamp( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_timestamp", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_timestamp, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		timestamp r;
		msg = str_2_timestamp( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_daytime( daytime *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_daytime, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_daytime, p, ATOMnilptr(TYPE_daytime)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "daytime", buf);
	}
	 *res = *(daytime*)p;
	if (!ATOMextern(TYPE_daytime)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLdaytime_2_str( str *res, daytime *val )
{
	char *p = NULL;
	int len = 0;
	daytime_tostr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_daytime( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_daytime", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_daytime, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		daytime r;
		msg = str_2_daytime( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_date( date *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_date, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_date, p, ATOMnilptr(TYPE_date)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "date", buf);
	}
	 *res = *(date*)p;
	if (!ATOMextern(TYPE_date)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLdate_2_str( str *res, date *val )
{
	char *p = NULL;
	int len = 0;
	date_tostr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_date( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_date", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_date, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		date r;
		msg = str_2_date( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
str_2_sqlblob( sqlblob* *res, str *val )
{
	ptr p = NULL;
	int len = 0;
	int e; 
	
	e = ATOMfromstr(TYPE_sqlblob, &p, &len, *val);
	if (e < 0 || !p || 
	   (ATOMcmp(TYPE_sqlblob, p, ATOMnilptr(TYPE_sqlblob)) == 0 &&
	    ATOMcmp(TYPE_str, *val, ATOMnilptr(TYPE_str)) != 0))
	{
		char buf[BUFSIZ];
		if (p)
			GDKfree(p);
                snprintf(buf, BUFSIZ,"conversion of string '%s' failed",*val);
		throw(SQL, "sqlblob", buf);
	}
	 *res = (sqlblob*)p;
	if (!ATOMextern(TYPE_sqlblob)) {
		if (p)
			GDKfree(p);
	}
	return MAL_SUCCEED;
}

str
SQLsqlblob_2_str( str *res, sqlblob *val )
{
	char *p = NULL;
	int len = 0;
	sqlblob_tostr( &p, &len, val);
	*res = p;
	return MAL_SUCCEED;
}

str
batstr_2_sqlblob( int *res, int *bid )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2_sqlblob", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sqlblob, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		str v = (str)BUNtail(bi,p);
		sqlblob* r;
		msg = str_2_sqlblob( &r, &v );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}




str
SQLstr_2_str( str *res, str *val )
{
	if( *val == str_nil) 
		*res = (str)str_nil;
	else
		*res =  GDKstrdup(*val);
	return MAL_SUCCEED;
}

str
SQLstr_cast_(str *res, mvc *m, int eclass, int d, int s, int has_tz, ptr p, int tpe, int len)
{
	char *r = NULL;
	int sz = MAX(2,len + 1);	/* nil should fit */

	if (tpe != TYPE_str) {
		r = GDKmalloc(sz);
		sz = convert2str(m, eclass, d, s, has_tz, p, tpe, &r, sz);
	} else {
		str v = (str)p; 
		strLength(&sz, v);
		if (len == 0 || sz <= len) 
			r = GDKstrdup(v);
	}
	if (len > 0 && sz > len) {
		if (r) 
			GDKfree(r);
		if (ATOMcmp(TYPE_str, ATOMnilptr(TYPE_str), p) != 0) {
			throw(SQL, "str_cast", "value too long for type (var)char(%d)", len);
		} else {
			r = GDKstrdup(str_nil);
		}
	}
	*res = r;
	return MAL_SUCCEED;
}

str
SQLstr_cast(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	str *res  = (str *) getArgReference(stk, pci, 0);
	int eclass = *(int*) getArgReference(stk, pci, 1);
	int d = *(int*) getArgReference(stk, pci, 2);
	int s = *(int*) getArgReference(stk, pci, 3);
	int has_tz = *(int*) getArgReference(stk, pci, 4);
	ptr p   = (ptr) getArgReference(stk, pci, 5);
	int tpe = getArgType(mb, pci, 5);
	int len   = *(int *) getArgReference(stk, pci, 6);
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);

	(void) cntxt;
	if (msg)
		return msg;
	if (ATOMextern(tpe)) 
		p = *(ptr*)p;
	return SQLstr_cast_(res, m, eclass, d, s, has_tz, p, tpe, len );
}

/* str SQLbatstr_cast(int *res, int *eclass, int *d1, int *s1, int *has_tz, int *bid, int *digits ); */
str 
SQLbatstr_cast(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	char *r = NULL;
	int *res  = (int *) getArgReference(stk, pci, 0);
	int *eclass = (int*) getArgReference(stk, pci, 1);
	int *d1 = (int*) getArgReference(stk, pci, 2);
	int *s1 = (int*) getArgReference(stk, pci, 3);
	int *has_tz = (int*) getArgReference(stk, pci, 4);
	int *bid   = (int*) getArgReference(stk, pci, 5);
	int *digits   = (int *) getArgReference(stk, pci, 6);

	(void) cntxt;
	if (msg)
		return msg;
	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.str_2dec_@1", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_str, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATaccessBegin(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BATloop(b,p,q) {
		ptr v = (ptr)BUNtail(bi,p);
		msg = SQLstr_cast_( &r, m, *eclass, *d1, *s1, *has_tz, v, b->ttype, *digits);
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), r, FALSE);
		GDKfree(r);
	}
	BATaccessEnd(b, USE_HEAD|USE_TAIL, MMAP_SEQUENTIAL);
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}








str 
bte_2_bte( bte *res, bte *v )
{
	/* shortcut nil */
	if (*v == bte_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the bte type is bigger than or equal to the bte type, it will
	   always fit */
	*res = (bte)*v;
	return(MAL_SUCCEED);
}

str batbte_2_bte( int *res, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	bte *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (bte)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == bte_nil) {
			*o = bte_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (bte)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str bte_dec2_bte( bte *res, int *s1, bte *v )
{
	int scale = *s1;
	bte r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}

	/* since the bte type is bigger than or equal to the bte type, it will
	   always fit */
	r = (bte)*v;
	if (scale) 
		r = (bte) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_dec2dec_bte( bte *res, int *S1, bte *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	bte cpyval = *v;
	int s1 = *S1, s2 = *S2;
	bte r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the bte type is bigger than or equal to the bte type, it will
	   always fit */
	r = (bte)*v;
	if (s2 > s1) 
		r *= (bte) scales[s2 - s1];
	else if (s2 != s1) 
		r = (bte) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_num2dec_bte( bte *res, bte *v, int *d2, int *s2 )
{
	int zero = 0;
	return bte_dec2dec_bte( res, &zero, v, d2, s2 );
}

str batbte_dec2_bte( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	char *msg = NULL;
	int scale = *s1;
	bte *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (bte) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (bte) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == bte_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (bte) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (bte) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batbte_dec2dec_bte( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		bte r;
		msg = bte_dec2dec_bte( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batbte_num2dec_bte( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_num2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		bte r;
		msg = bte_num2dec_bte( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
bte_2_sht( sht *res, bte *v )
{
	/* shortcut nil */
	if (*v == bte_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the sht type is bigger than or equal to the bte type, it will
	   always fit */
	*res = (sht)*v;
	return(MAL_SUCCEED);
}

str batbte_2_sht( int *res, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	sht *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (sht)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == bte_nil) {
			*o = sht_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (sht)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str bte_dec2_sht( sht *res, int *s1, bte *v )
{
	int scale = *s1;
	sht r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}

	/* since the sht type is bigger than or equal to the bte type, it will
	   always fit */
	r = (sht)*v;
	if (scale) 
		r = (sht) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_dec2dec_sht( sht *res, int *S1, bte *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	bte cpyval = *v;
	int s1 = *S1, s2 = *S2;
	sht r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the sht type is bigger than or equal to the bte type, it will
	   always fit */
	r = (sht)*v;
	if (s2 > s1) 
		r *= (sht) scales[s2 - s1];
	else if (s2 != s1) 
		r = (sht) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_num2dec_sht( sht *res, bte *v, int *d2, int *s2 )
{
	int zero = 0;
	return bte_dec2dec_sht( res, &zero, v, d2, s2 );
}

str batbte_dec2_sht( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	char *msg = NULL;
	int scale = *s1;
	sht *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (sht) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (sht) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == bte_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (sht) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (sht) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batbte_dec2dec_sht( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		sht r;
		msg = bte_dec2dec_sht( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batbte_num2dec_sht( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_num2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		sht r;
		msg = bte_num2dec_sht( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
sht_2_sht( sht *res, sht *v )
{
	/* shortcut nil */
	if (*v == sht_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the sht type is bigger than or equal to the sht type, it will
	   always fit */
	*res = (sht)*v;
	return(MAL_SUCCEED);
}

str batsht_2_sht( int *res, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	sht *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (sht)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == sht_nil) {
			*o = sht_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (sht)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str sht_dec2_sht( sht *res, int *s1, sht *v )
{
	int scale = *s1;
	sht r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}

	/* since the sht type is bigger than or equal to the sht type, it will
	   always fit */
	r = (sht)*v;
	if (scale) 
		r = (sht) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
sht_dec2dec_sht( sht *res, int *S1, sht *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	sht cpyval = *v;
	int s1 = *S1, s2 = *S2;
	sht r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the sht type is bigger than or equal to the sht type, it will
	   always fit */
	r = (sht)*v;
	if (s2 > s1) 
		r *= (sht) scales[s2 - s1];
	else if (s2 != s1) 
		r = (sht) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
sht_num2dec_sht( sht *res, sht *v, int *d2, int *s2 )
{
	int zero = 0;
	return sht_dec2dec_sht( res, &zero, v, d2, s2 );
}

str batsht_dec2_sht( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	char *msg = NULL;
	int scale = *s1;
	sht *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (sht) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (sht) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == sht_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (sht) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (sht) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batsht_dec2dec_sht( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		sht r;
		msg = sht_dec2dec_sht( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batsht_num2dec_sht( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_num2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		sht r;
		msg = sht_num2dec_sht( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
bte_2_int( int *res, bte *v )
{
	/* shortcut nil */
	if (*v == bte_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the int type is bigger than or equal to the bte type, it will
	   always fit */
	*res = (int)*v;
	return(MAL_SUCCEED);
}

str batbte_2_int( int *res, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	int *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (int)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == bte_nil) {
			*o = int_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (int)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str bte_dec2_int( int *res, int *s1, bte *v )
{
	int scale = *s1;
	int r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}

	/* since the int type is bigger than or equal to the bte type, it will
	   always fit */
	r = (int)*v;
	if (scale) 
		r = (int) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_dec2dec_int( int *res, int *S1, bte *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	bte cpyval = *v;
	int s1 = *S1, s2 = *S2;
	int r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the int type is bigger than or equal to the bte type, it will
	   always fit */
	r = (int)*v;
	if (s2 > s1) 
		r *= (int) scales[s2 - s1];
	else if (s2 != s1) 
		r = (int) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_num2dec_int( int *res, bte *v, int *d2, int *s2 )
{
	int zero = 0;
	return bte_dec2dec_int( res, &zero, v, d2, s2 );
}

str batbte_dec2_int( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	char *msg = NULL;
	int scale = *s1;
	int *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (int) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (int) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == bte_nil) {
				*o = int_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (int) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (int) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batbte_dec2dec_int( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		int r;
		msg = bte_dec2dec_int( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batbte_num2dec_int( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_num2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		int r;
		msg = bte_num2dec_int( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
sht_2_int( int *res, sht *v )
{
	/* shortcut nil */
	if (*v == sht_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the int type is bigger than or equal to the sht type, it will
	   always fit */
	*res = (int)*v;
	return(MAL_SUCCEED);
}

str batsht_2_int( int *res, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	int *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (int)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == sht_nil) {
			*o = int_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (int)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str sht_dec2_int( int *res, int *s1, sht *v )
{
	int scale = *s1;
	int r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}

	/* since the int type is bigger than or equal to the sht type, it will
	   always fit */
	r = (int)*v;
	if (scale) 
		r = (int) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
sht_dec2dec_int( int *res, int *S1, sht *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	sht cpyval = *v;
	int s1 = *S1, s2 = *S2;
	int r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the int type is bigger than or equal to the sht type, it will
	   always fit */
	r = (int)*v;
	if (s2 > s1) 
		r *= (int) scales[s2 - s1];
	else if (s2 != s1) 
		r = (int) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
sht_num2dec_int( int *res, sht *v, int *d2, int *s2 )
{
	int zero = 0;
	return sht_dec2dec_int( res, &zero, v, d2, s2 );
}

str batsht_dec2_int( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	char *msg = NULL;
	int scale = *s1;
	int *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (int) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (int) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == sht_nil) {
				*o = int_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (int) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (int) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batsht_dec2dec_int( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		int r;
		msg = sht_dec2dec_int( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batsht_num2dec_int( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_num2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		int r;
		msg = sht_num2dec_int( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
int_2_int( int *res, int *v )
{
	/* shortcut nil */
	if (*v == int_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the int type is bigger than or equal to the int type, it will
	   always fit */
	*res = (int)*v;
	return(MAL_SUCCEED);
}

str batint_2_int( int *res, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	int *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (int)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == int_nil) {
			*o = int_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (int)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str int_dec2_int( int *res, int *s1, int *v )
{
	int scale = *s1;
	int r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == int_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}

	/* since the int type is bigger than or equal to the int type, it will
	   always fit */
	r = (int)*v;
	if (scale) 
		r = (int) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
int_dec2dec_int( int *res, int *S1, int *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	int cpyval = *v;
	int s1 = *S1, s2 = *S2;
	int r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == int_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the int type is bigger than or equal to the int type, it will
	   always fit */
	r = (int)*v;
	if (s2 > s1) 
		r *= (int) scales[s2 - s1];
	else if (s2 != s1) 
		r = (int) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
int_num2dec_int( int *res, int *v, int *d2, int *s2 )
{
	int zero = 0;
	return int_dec2dec_int( res, &zero, v, d2, s2 );
}

str batint_dec2_int( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	char *msg = NULL;
	int scale = *s1;
	int *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (int) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (int) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == int_nil) {
				*o = int_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (int) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (int) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batint_dec2dec_int( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		int r;
		msg = int_dec2dec_int( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batint_num2dec_int( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_num2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		int r;
		msg = int_num2dec_int( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
bte_2_wrd( wrd *res, bte *v )
{
	/* shortcut nil */
	if (*v == bte_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the wrd type is bigger than or equal to the bte type, it will
	   always fit */
	*res = (wrd)*v;
	return(MAL_SUCCEED);
}

str batbte_2_wrd( int *res, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	wrd *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (wrd)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == bte_nil) {
			*o = wrd_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (wrd)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str bte_dec2_wrd( wrd *res, int *s1, bte *v )
{
	int scale = *s1;
	wrd r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}

	/* since the wrd type is bigger than or equal to the bte type, it will
	   always fit */
	r = (wrd)*v;
	if (scale) 
		r = (wrd) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_dec2dec_wrd( wrd *res, int *S1, bte *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	bte cpyval = *v;
	int s1 = *S1, s2 = *S2;
	wrd r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the wrd type is bigger than or equal to the bte type, it will
	   always fit */
	r = (wrd)*v;
	if (s2 > s1) 
		r *= (wrd) scales[s2 - s1];
	else if (s2 != s1) 
		r = (wrd) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_num2dec_wrd( wrd *res, bte *v, int *d2, int *s2 )
{
	int zero = 0;
	return bte_dec2dec_wrd( res, &zero, v, d2, s2 );
}

str batbte_dec2_wrd( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	char *msg = NULL;
	int scale = *s1;
	wrd *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (wrd) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (wrd) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == bte_nil) {
				*o = wrd_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (wrd) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (wrd) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batbte_dec2dec_wrd( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		wrd r;
		msg = bte_dec2dec_wrd( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batbte_num2dec_wrd( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_num2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		wrd r;
		msg = bte_num2dec_wrd( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
sht_2_wrd( wrd *res, sht *v )
{
	/* shortcut nil */
	if (*v == sht_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the wrd type is bigger than or equal to the sht type, it will
	   always fit */
	*res = (wrd)*v;
	return(MAL_SUCCEED);
}

str batsht_2_wrd( int *res, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	wrd *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (wrd)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == sht_nil) {
			*o = wrd_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (wrd)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str sht_dec2_wrd( wrd *res, int *s1, sht *v )
{
	int scale = *s1;
	wrd r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}

	/* since the wrd type is bigger than or equal to the sht type, it will
	   always fit */
	r = (wrd)*v;
	if (scale) 
		r = (wrd) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
sht_dec2dec_wrd( wrd *res, int *S1, sht *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	sht cpyval = *v;
	int s1 = *S1, s2 = *S2;
	wrd r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the wrd type is bigger than or equal to the sht type, it will
	   always fit */
	r = (wrd)*v;
	if (s2 > s1) 
		r *= (wrd) scales[s2 - s1];
	else if (s2 != s1) 
		r = (wrd) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
sht_num2dec_wrd( wrd *res, sht *v, int *d2, int *s2 )
{
	int zero = 0;
	return sht_dec2dec_wrd( res, &zero, v, d2, s2 );
}

str batsht_dec2_wrd( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	char *msg = NULL;
	int scale = *s1;
	wrd *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (wrd) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (wrd) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == sht_nil) {
				*o = wrd_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (wrd) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (wrd) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batsht_dec2dec_wrd( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		wrd r;
		msg = sht_dec2dec_wrd( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batsht_num2dec_wrd( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_num2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		wrd r;
		msg = sht_num2dec_wrd( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
int_2_wrd( wrd *res, int *v )
{
	/* shortcut nil */
	if (*v == int_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the wrd type is bigger than or equal to the int type, it will
	   always fit */
	*res = (wrd)*v;
	return(MAL_SUCCEED);
}

str batint_2_wrd( int *res, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	wrd *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (wrd)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == int_nil) {
			*o = wrd_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (wrd)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str int_dec2_wrd( wrd *res, int *s1, int *v )
{
	int scale = *s1;
	wrd r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == int_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}

	/* since the wrd type is bigger than or equal to the int type, it will
	   always fit */
	r = (wrd)*v;
	if (scale) 
		r = (wrd) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
int_dec2dec_wrd( wrd *res, int *S1, int *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	int cpyval = *v;
	int s1 = *S1, s2 = *S2;
	wrd r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == int_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the wrd type is bigger than or equal to the int type, it will
	   always fit */
	r = (wrd)*v;
	if (s2 > s1) 
		r *= (wrd) scales[s2 - s1];
	else if (s2 != s1) 
		r = (wrd) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
int_num2dec_wrd( wrd *res, int *v, int *d2, int *s2 )
{
	int zero = 0;
	return int_dec2dec_wrd( res, &zero, v, d2, s2 );
}

str batint_dec2_wrd( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	char *msg = NULL;
	int scale = *s1;
	wrd *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (wrd) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (wrd) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == int_nil) {
				*o = wrd_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (wrd) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (wrd) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batint_dec2dec_wrd( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		wrd r;
		msg = int_dec2dec_wrd( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batint_num2dec_wrd( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_num2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		wrd r;
		msg = int_num2dec_wrd( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
wrd_2_wrd( wrd *res, wrd *v )
{
	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the wrd type is bigger than or equal to the wrd type, it will
	   always fit */
	*res = (wrd)*v;
	return(MAL_SUCCEED);
}

str batwrd_2_wrd( int *res, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	wrd *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (wrd)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == wrd_nil) {
			*o = wrd_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (wrd)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str wrd_dec2_wrd( wrd *res, int *s1, wrd *v )
{
	int scale = *s1;
	wrd r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}

	/* since the wrd type is bigger than or equal to the wrd type, it will
	   always fit */
	r = (wrd)*v;
	if (scale) 
		r = (wrd) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
wrd_dec2dec_wrd( wrd *res, int *S1, wrd *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	wrd cpyval = *v;
	int s1 = *S1, s2 = *S2;
	wrd r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the wrd type is bigger than or equal to the wrd type, it will
	   always fit */
	r = (wrd)*v;
	if (s2 > s1) 
		r *= (wrd) scales[s2 - s1];
	else if (s2 != s1) 
		r = (wrd) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
wrd_num2dec_wrd( wrd *res, wrd *v, int *d2, int *s2 )
{
	int zero = 0;
	return wrd_dec2dec_wrd( res, &zero, v, d2, s2 );
}

str batwrd_dec2_wrd( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	int scale = *s1;
	wrd *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (wrd) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (wrd) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = wrd_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (wrd) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (wrd) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batwrd_dec2dec_wrd( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		wrd r;
		msg = wrd_dec2dec_wrd( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batwrd_num2dec_wrd( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_num2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		wrd r;
		msg = wrd_num2dec_wrd( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
bte_2_lng( lng *res, bte *v )
{
	/* shortcut nil */
	if (*v == bte_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the lng type is bigger than or equal to the bte type, it will
	   always fit */
	*res = (lng)*v;
	return(MAL_SUCCEED);
}

str batbte_2_lng( int *res, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (lng)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == bte_nil) {
			*o = lng_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (lng)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str bte_dec2_lng( lng *res, int *s1, bte *v )
{
	int scale = *s1;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}

	/* since the lng type is bigger than or equal to the bte type, it will
	   always fit */
	r = (lng)*v;
	if (scale) 
		r = (lng) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_dec2dec_lng( lng *res, int *S1, bte *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	bte cpyval = *v;
	int s1 = *S1, s2 = *S2;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the lng type is bigger than or equal to the bte type, it will
	   always fit */
	r = (lng)*v;
	if (s2 > s1) 
		r *= (lng) scales[s2 - s1];
	else if (s2 != s1) 
		r = (lng) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
bte_num2dec_lng( lng *res, bte *v, int *d2, int *s2 )
{
	int zero = 0;
	return bte_dec2dec_lng( res, &zero, v, d2, s2 );
}

str batbte_dec2_lng( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	char *msg = NULL;
	int scale = *s1;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (lng) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == bte_nil) {
				*o = lng_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (lng) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batbte_dec2dec_lng( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		lng r;
		msg = bte_dec2dec_lng( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batbte_num2dec_lng( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_num2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		lng r;
		msg = bte_num2dec_lng( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
sht_2_lng( lng *res, sht *v )
{
	/* shortcut nil */
	if (*v == sht_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the lng type is bigger than or equal to the sht type, it will
	   always fit */
	*res = (lng)*v;
	return(MAL_SUCCEED);
}

str batsht_2_lng( int *res, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (lng)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == sht_nil) {
			*o = lng_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (lng)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str sht_dec2_lng( lng *res, int *s1, sht *v )
{
	int scale = *s1;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}

	/* since the lng type is bigger than or equal to the sht type, it will
	   always fit */
	r = (lng)*v;
	if (scale) 
		r = (lng) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
sht_dec2dec_lng( lng *res, int *S1, sht *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	sht cpyval = *v;
	int s1 = *S1, s2 = *S2;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the lng type is bigger than or equal to the sht type, it will
	   always fit */
	r = (lng)*v;
	if (s2 > s1) 
		r *= (lng) scales[s2 - s1];
	else if (s2 != s1) 
		r = (lng) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
sht_num2dec_lng( lng *res, sht *v, int *d2, int *s2 )
{
	int zero = 0;
	return sht_dec2dec_lng( res, &zero, v, d2, s2 );
}

str batsht_dec2_lng( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	char *msg = NULL;
	int scale = *s1;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (lng) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == sht_nil) {
				*o = lng_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (lng) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batsht_dec2dec_lng( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		lng r;
		msg = sht_dec2dec_lng( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batsht_num2dec_lng( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_num2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		lng r;
		msg = sht_num2dec_lng( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
int_2_lng( lng *res, int *v )
{
	/* shortcut nil */
	if (*v == int_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the lng type is bigger than or equal to the int type, it will
	   always fit */
	*res = (lng)*v;
	return(MAL_SUCCEED);
}

str batint_2_lng( int *res, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (lng)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == int_nil) {
			*o = lng_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (lng)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str int_dec2_lng( lng *res, int *s1, int *v )
{
	int scale = *s1;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == int_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}

	/* since the lng type is bigger than or equal to the int type, it will
	   always fit */
	r = (lng)*v;
	if (scale) 
		r = (lng) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
int_dec2dec_lng( lng *res, int *S1, int *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	int cpyval = *v;
	int s1 = *S1, s2 = *S2;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == int_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the lng type is bigger than or equal to the int type, it will
	   always fit */
	r = (lng)*v;
	if (s2 > s1) 
		r *= (lng) scales[s2 - s1];
	else if (s2 != s1) 
		r = (lng) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
int_num2dec_lng( lng *res, int *v, int *d2, int *s2 )
{
	int zero = 0;
	return int_dec2dec_lng( res, &zero, v, d2, s2 );
}

str batint_dec2_lng( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	char *msg = NULL;
	int scale = *s1;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (lng) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == int_nil) {
				*o = lng_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (lng) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batint_dec2dec_lng( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		lng r;
		msg = int_dec2dec_lng( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batint_num2dec_lng( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_num2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		lng r;
		msg = int_num2dec_lng( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
wrd_2_lng( lng *res, wrd *v )
{
	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the lng type is bigger than or equal to the wrd type, it will
	   always fit */
	*res = (lng)*v;
	return(MAL_SUCCEED);
}

str batwrd_2_lng( int *res, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (lng)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == wrd_nil) {
			*o = lng_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (lng)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str wrd_dec2_lng( lng *res, int *s1, wrd *v )
{
	int scale = *s1;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}

	/* since the lng type is bigger than or equal to the wrd type, it will
	   always fit */
	r = (lng)*v;
	if (scale) 
		r = (lng) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
wrd_dec2dec_lng( lng *res, int *S1, wrd *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	wrd cpyval = *v;
	int s1 = *S1, s2 = *S2;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the lng type is bigger than or equal to the wrd type, it will
	   always fit */
	r = (lng)*v;
	if (s2 > s1) 
		r *= (lng) scales[s2 - s1];
	else if (s2 != s1) 
		r = (lng) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
wrd_num2dec_lng( lng *res, wrd *v, int *d2, int *s2 )
{
	int zero = 0;
	return wrd_dec2dec_lng( res, &zero, v, d2, s2 );
}

str batwrd_dec2_lng( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	int scale = *s1;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (lng) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = lng_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (lng) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batwrd_dec2dec_lng( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		lng r;
		msg = wrd_dec2dec_lng( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batwrd_num2dec_lng( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_num2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		lng r;
		msg = wrd_num2dec_lng( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
lng_2_lng( lng *res, lng *v )
{
	/* shortcut nil */
	if (*v == lng_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the lng type is bigger than or equal to the lng type, it will
	   always fit */
	*res = (lng)*v;
	return(MAL_SUCCEED);
}

str batlng_2_lng( int *res, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (lng)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == lng_nil) {
			*o = lng_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (lng)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str lng_dec2_lng( lng *res, int *s1, lng *v )
{
	int scale = *s1;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}

	/* since the lng type is bigger than or equal to the lng type, it will
	   always fit */
	r = (lng)*v;
	if (scale) 
		r = (lng) ((r + h) / scales[scale]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
lng_dec2dec_lng( lng *res, int *S1, lng *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng cpyval = *v;
	int s1 = *S1, s2 = *S2;
	lng r, h = (*v<0)?-5:5; 

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;

	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the lng type is bigger than or equal to the lng type, it will
	   always fit */
	r = (lng)*v;
	if (s2 > s1) 
		r *= (lng) scales[s2 - s1];
	else if (s2 != s1) 
		r = (lng) ((r + h) / scales[s1 - s2]);
	*res = r;
	return(MAL_SUCCEED);
}

str 
lng_num2dec_lng( lng *res, lng *v, int *d2, int *s2 )
{
	int zero = 0;
	return lng_dec2dec_lng( res, &zero, v, d2, s2 );
}

str batlng_dec2_lng( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	int scale = *s1;
	lng *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		if (scale)
			for (; p<q; p++, o++)
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
		else
			for (; p<q; p++, o++)
				*o = (lng) (*p);
	} else {
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = lng_nil;
				bn->T->nonil= FALSE;
			} else if (scale) {
				*o = (lng) ((*p +  (*p<0)?-5:5) / scales[scale]);
			} else {
				*o = (lng) (*p);
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batlng_dec2dec_lng( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		lng r;
		msg = lng_dec2dec_lng( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}
str batlng_num2dec_lng( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_num2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		lng r;
		msg = lng_num2dec_lng( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}









str flt_2_bte( bte *res, flt *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(bte)val > (dbl) GDK_bte_min && 
	    val > (dbl) GDK_bte_min && val <= (dbl) GDK_bte_max)	{
		*res = (bte)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type bte", val);
	}
}

str batflt_2_bte( int *res, int *bid )
{
	BAT *b, *bn;
	flt *p,*q;
	char *msg = NULL;
	bte *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (flt*) Tloc(b, BUNfirst(b));
	q = (flt*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(bte)val > (dbl) GDK_bte_min && val > (dbl) GDK_bte_min && val <= (dbl) GDK_bte_max)	{
				*o = (bte)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type bte", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == flt_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(bte)val > (dbl) GDK_bte_min && val > (dbl) GDK_bte_min && val <= (dbl) GDK_bte_max)	{
					*o = (bte)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type bte", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
flt_num2dec_bte( bte *res, flt *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	flt r;
	lng cpyval;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the bte type is bigger than or equal to the flt type, it will
	   always fit */
	r = (flt) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (bte) r;
	return MAL_SUCCEED;
}
str 
batflt_num2dec_bte( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_num2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		flt *v = (flt*)BUNtail(bi,p);
		bte r;
		msg = flt_num2dec_bte( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str flt_2_sht( sht *res, flt *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(sht)val > (dbl) GDK_sht_min && 
	    val > (dbl) GDK_sht_min && val <= (dbl) GDK_sht_max)	{
		*res = (sht)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type sht", val);
	}
}

str batflt_2_sht( int *res, int *bid )
{
	BAT *b, *bn;
	flt *p,*q;
	char *msg = NULL;
	sht *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (flt*) Tloc(b, BUNfirst(b));
	q = (flt*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(sht)val > (dbl) GDK_sht_min && val > (dbl) GDK_sht_min && val <= (dbl) GDK_sht_max)	{
				*o = (sht)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type sht", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == flt_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(sht)val > (dbl) GDK_sht_min && val > (dbl) GDK_sht_min && val <= (dbl) GDK_sht_max)	{
					*o = (sht)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type sht", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
flt_num2dec_sht( sht *res, flt *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	flt r;
	lng cpyval;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the sht type is bigger than or equal to the flt type, it will
	   always fit */
	r = (flt) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (sht) r;
	return MAL_SUCCEED;
}
str 
batflt_num2dec_sht( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_num2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		flt *v = (flt*)BUNtail(bi,p);
		sht r;
		msg = flt_num2dec_sht( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str flt_2_int( int *res, flt *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(int)val > (dbl) GDK_int_min && 
	    val > (dbl) GDK_int_min && val <= (dbl) GDK_int_max)	{
		*res = (int)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type int", val);
	}
}

str batflt_2_int( int *res, int *bid )
{
	BAT *b, *bn;
	flt *p,*q;
	char *msg = NULL;
	int *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (flt*) Tloc(b, BUNfirst(b));
	q = (flt*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(int)val > (dbl) GDK_int_min && val > (dbl) GDK_int_min && val <= (dbl) GDK_int_max)	{
				*o = (int)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type int", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == flt_nil) {
				*o = int_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(int)val > (dbl) GDK_int_min && val > (dbl) GDK_int_min && val <= (dbl) GDK_int_max)	{
					*o = (int)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type int", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
flt_num2dec_int( int *res, flt *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	flt r;
	lng cpyval;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the int type is bigger than or equal to the flt type, it will
	   always fit */
	r = (flt) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (int) r;
	return MAL_SUCCEED;
}
str 
batflt_num2dec_int( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_num2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		flt *v = (flt*)BUNtail(bi,p);
		int r;
		msg = flt_num2dec_int( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str flt_2_wrd( wrd *res, flt *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(wrd)val > (dbl) GDK_wrd_min && 
	    val > (dbl) GDK_wrd_min && val <= (dbl) GDK_wrd_max)	{
		*res = (wrd)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type wrd", val);
	}
}

str batflt_2_wrd( int *res, int *bid )
{
	BAT *b, *bn;
	flt *p,*q;
	char *msg = NULL;
	wrd *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (flt*) Tloc(b, BUNfirst(b));
	q = (flt*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(wrd)val > (dbl) GDK_wrd_min && val > (dbl) GDK_wrd_min && val <= (dbl) GDK_wrd_max)	{
				*o = (wrd)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type wrd", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == flt_nil) {
				*o = wrd_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(wrd)val > (dbl) GDK_wrd_min && val > (dbl) GDK_wrd_min && val <= (dbl) GDK_wrd_max)	{
					*o = (wrd)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type wrd", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
flt_num2dec_wrd( wrd *res, flt *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	flt r;
	lng cpyval;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the wrd type is bigger than or equal to the flt type, it will
	   always fit */
	r = (flt) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (wrd) r;
	return MAL_SUCCEED;
}
str 
batflt_num2dec_wrd( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_num2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		flt *v = (flt*)BUNtail(bi,p);
		wrd r;
		msg = flt_num2dec_wrd( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str flt_2_lng( lng *res, flt *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(lng)val > (dbl) GDK_lng_min && 
	    val > (dbl) GDK_lng_min && val <= (dbl) GDK_lng_max)	{
		*res = (lng)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type lng", val);
	}
}

str batflt_2_lng( int *res, int *bid )
{
	BAT *b, *bn;
	flt *p,*q;
	char *msg = NULL;
	lng *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (flt*) Tloc(b, BUNfirst(b));
	q = (flt*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(lng)val > (dbl) GDK_lng_min && val > (dbl) GDK_lng_min && val <= (dbl) GDK_lng_max)	{
				*o = (lng)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type lng", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == flt_nil) {
				*o = lng_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(lng)val > (dbl) GDK_lng_min && val > (dbl) GDK_lng_min && val <= (dbl) GDK_lng_max)	{
					*o = (lng)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type lng", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
flt_num2dec_lng( lng *res, flt *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	flt r;
	lng cpyval;

	/* shortcut nil */
	if (*v == flt_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the lng type is bigger than or equal to the flt type, it will
	   always fit */
	r = (flt) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (lng) r;
	return MAL_SUCCEED;
}
str 
batflt_num2dec_lng( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.flt_num2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		flt *v = (flt*)BUNtail(bi,p);
		lng r;
		msg = flt_num2dec_lng( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str dbl_2_bte( bte *res, dbl *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(bte)val > (dbl) GDK_bte_min && 
	    val > (dbl) GDK_bte_min && val <= (dbl) GDK_bte_max)	{
		*res = (bte)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type bte", val);
	}
}

str batdbl_2_bte( int *res, int *bid )
{
	BAT *b, *bn;
	dbl *p,*q;
	char *msg = NULL;
	bte *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (dbl*) Tloc(b, BUNfirst(b));
	q = (dbl*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(bte)val > (dbl) GDK_bte_min && val > (dbl) GDK_bte_min && val <= (dbl) GDK_bte_max)	{
				*o = (bte)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type bte", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == dbl_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(bte)val > (dbl) GDK_bte_min && val > (dbl) GDK_bte_min && val <= (dbl) GDK_bte_max)	{
					*o = (bte)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type bte", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
dbl_num2dec_bte( bte *res, dbl *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	dbl r;
	lng cpyval;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the bte type is bigger than or equal to the dbl type, it will
	   always fit */
	r = (dbl) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (bte) r;
	return MAL_SUCCEED;
}
str 
batdbl_num2dec_bte( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_num2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		dbl *v = (dbl*)BUNtail(bi,p);
		bte r;
		msg = dbl_num2dec_bte( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str dbl_2_sht( sht *res, dbl *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(sht)val > (dbl) GDK_sht_min && 
	    val > (dbl) GDK_sht_min && val <= (dbl) GDK_sht_max)	{
		*res = (sht)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type sht", val);
	}
}

str batdbl_2_sht( int *res, int *bid )
{
	BAT *b, *bn;
	dbl *p,*q;
	char *msg = NULL;
	sht *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (dbl*) Tloc(b, BUNfirst(b));
	q = (dbl*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(sht)val > (dbl) GDK_sht_min && val > (dbl) GDK_sht_min && val <= (dbl) GDK_sht_max)	{
				*o = (sht)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type sht", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == dbl_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(sht)val > (dbl) GDK_sht_min && val > (dbl) GDK_sht_min && val <= (dbl) GDK_sht_max)	{
					*o = (sht)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type sht", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
dbl_num2dec_sht( sht *res, dbl *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	dbl r;
	lng cpyval;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the sht type is bigger than or equal to the dbl type, it will
	   always fit */
	r = (dbl) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (sht) r;
	return MAL_SUCCEED;
}
str 
batdbl_num2dec_sht( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_num2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		dbl *v = (dbl*)BUNtail(bi,p);
		sht r;
		msg = dbl_num2dec_sht( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str dbl_2_int( int *res, dbl *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(int)val > (dbl) GDK_int_min && 
	    val > (dbl) GDK_int_min && val <= (dbl) GDK_int_max)	{
		*res = (int)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type int", val);
	}
}

str batdbl_2_int( int *res, int *bid )
{
	BAT *b, *bn;
	dbl *p,*q;
	char *msg = NULL;
	int *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (dbl*) Tloc(b, BUNfirst(b));
	q = (dbl*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(int)val > (dbl) GDK_int_min && val > (dbl) GDK_int_min && val <= (dbl) GDK_int_max)	{
				*o = (int)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type int", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == dbl_nil) {
				*o = int_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(int)val > (dbl) GDK_int_min && val > (dbl) GDK_int_min && val <= (dbl) GDK_int_max)	{
					*o = (int)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type int", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
dbl_num2dec_int( int *res, dbl *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	dbl r;
	lng cpyval;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the int type is bigger than or equal to the dbl type, it will
	   always fit */
	r = (dbl) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (int) r;
	return MAL_SUCCEED;
}
str 
batdbl_num2dec_int( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_num2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		dbl *v = (dbl*)BUNtail(bi,p);
		int r;
		msg = dbl_num2dec_int( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str dbl_2_wrd( wrd *res, dbl *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(wrd)val > (dbl) GDK_wrd_min && 
	    val > (dbl) GDK_wrd_min && val <= (dbl) GDK_wrd_max)	{
		*res = (wrd)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type wrd", val);
	}
}

str batdbl_2_wrd( int *res, int *bid )
{
	BAT *b, *bn;
	dbl *p,*q;
	char *msg = NULL;
	wrd *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (dbl*) Tloc(b, BUNfirst(b));
	q = (dbl*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(wrd)val > (dbl) GDK_wrd_min && val > (dbl) GDK_wrd_min && val <= (dbl) GDK_wrd_max)	{
				*o = (wrd)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type wrd", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == dbl_nil) {
				*o = wrd_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(wrd)val > (dbl) GDK_wrd_min && val > (dbl) GDK_wrd_min && val <= (dbl) GDK_wrd_max)	{
					*o = (wrd)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type wrd", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
dbl_num2dec_wrd( wrd *res, dbl *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	dbl r;
	lng cpyval;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the wrd type is bigger than or equal to the dbl type, it will
	   always fit */
	r = (dbl) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (wrd) r;
	return MAL_SUCCEED;
}
str 
batdbl_num2dec_wrd( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_num2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		dbl *v = (dbl*)BUNtail(bi,p);
		wrd r;
		msg = dbl_num2dec_wrd( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str dbl_2_lng( lng *res, dbl *v )
{
	dbl val = *v;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((dbl)(lng)val > (dbl) GDK_lng_min && 
	    val > (dbl) GDK_lng_min && val <= (dbl) GDK_lng_max)	{
		*res = (lng)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" "%f" ") exceeds limits of type lng", val);
	}
}

str batdbl_2_lng( int *res, int *bid )
{
	BAT *b, *bn;
	dbl *p,*q;
	char *msg = NULL;
	lng *o;
	dbl val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_2_lng", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_lng, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (lng*) Tloc(bn,BUNfirst(bn));
	p = (dbl*) Tloc(b, BUNfirst(b));
	q = (dbl*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((dbl)(lng)val > (dbl) GDK_lng_min && val > (dbl) GDK_lng_min && val <= (dbl) GDK_lng_max)	{
				*o = (lng)val;
			} else {
				msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type lng", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == dbl_nil) {
				*o = lng_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((dbl)(lng)val > (dbl) GDK_lng_min && val > (dbl) GDK_lng_min && val <= (dbl) GDK_lng_max)	{
					*o = (lng)val;
				} else {
					msg= createException(SQL, "convert", "value (" "%f" ") exceeds limits of type lng", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




/* when casting a floating point to an decimal we like to preserve the 
 * precision.  This means we first scale the float before converting.
*/
str 
dbl_num2dec_lng( lng *res, dbl *v, int *d2, int *s2 )
{
	int p = *d2, inlen = 1, scale = *s2;
	dbl r;
	lng cpyval;

	/* shortcut nil */
	if (*v == dbl_nil) {
		*res = lng_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the lng type is bigger than or equal to the dbl type, it will
	   always fit */
	r = (dbl) *v;
	if (scale) 
		r *= scales[scale];
	cpyval = (lng) r;

	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}
	*res = (lng) r;
	return MAL_SUCCEED;
}
str 
batdbl_num2dec_lng( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.dbl_num2dec_lng", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_lng, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		dbl *v = (dbl*)BUNtail(bi,p);
		lng r;
		msg = dbl_num2dec_lng( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}









str 
bte_2_flt( flt *res, bte *v )
{
	/* shortcut nil */
	if (*v == bte_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the bte type, it will
	   always fit */
	*res = (flt)*v;
	return(MAL_SUCCEED);
}

str batbte_2_flt( int *res, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (flt)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == bte_nil) {
			*o = flt_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (flt)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
bte_dec2_flt( flt *res, int *s1, bte *v )
{
	int scale = *s1;
	flt r;

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the bte type, it will
	   always fit */
	r = (flt)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
bte_dec2dec_flt( flt *res, int *S1, bte *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	bte cpyval = *v;
	int s1 = *S1, s2 = *S2;
	flt r;

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the flt type is bigger than or equal to the bte type, it will
	   always fit */
	r = (flt)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
bte_num2dec_flt( flt *res, bte *v, int *d2, int *s2 )
{
	int zero = 0;
	return bte_dec2dec_flt( res, &zero, v, d2, s2 );
}

str batbte_dec2_flt( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	char *msg = NULL;
	int scale = *s1;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((flt)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == bte_nil) {
				*o = flt_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((flt)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batbte_dec2dec_flt( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		flt r;
		msg = bte_dec2dec_flt( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batbte_num2dec_flt( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_num2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		flt r;
		msg = bte_num2dec_flt( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
sht_2_flt( flt *res, sht *v )
{
	/* shortcut nil */
	if (*v == sht_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the sht type, it will
	   always fit */
	*res = (flt)*v;
	return(MAL_SUCCEED);
}

str batsht_2_flt( int *res, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (flt)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == sht_nil) {
			*o = flt_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (flt)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
sht_dec2_flt( flt *res, int *s1, sht *v )
{
	int scale = *s1;
	flt r;

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the sht type, it will
	   always fit */
	r = (flt)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
sht_dec2dec_flt( flt *res, int *S1, sht *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	sht cpyval = *v;
	int s1 = *S1, s2 = *S2;
	flt r;

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the flt type is bigger than or equal to the sht type, it will
	   always fit */
	r = (flt)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
sht_num2dec_flt( flt *res, sht *v, int *d2, int *s2 )
{
	int zero = 0;
	return sht_dec2dec_flt( res, &zero, v, d2, s2 );
}

str batsht_dec2_flt( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	char *msg = NULL;
	int scale = *s1;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((flt)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == sht_nil) {
				*o = flt_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((flt)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batsht_dec2dec_flt( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		flt r;
		msg = sht_dec2dec_flt( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batsht_num2dec_flt( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_num2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		flt r;
		msg = sht_num2dec_flt( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
int_2_flt( flt *res, int *v )
{
	/* shortcut nil */
	if (*v == int_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the int type, it will
	   always fit */
	*res = (flt)*v;
	return(MAL_SUCCEED);
}

str batint_2_flt( int *res, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (flt)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == int_nil) {
			*o = flt_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (flt)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
int_dec2_flt( flt *res, int *s1, int *v )
{
	int scale = *s1;
	flt r;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the int type, it will
	   always fit */
	r = (flt)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
int_dec2dec_flt( flt *res, int *S1, int *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	int cpyval = *v;
	int s1 = *S1, s2 = *S2;
	flt r;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the flt type is bigger than or equal to the int type, it will
	   always fit */
	r = (flt)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
int_num2dec_flt( flt *res, int *v, int *d2, int *s2 )
{
	int zero = 0;
	return int_dec2dec_flt( res, &zero, v, d2, s2 );
}

str batint_dec2_flt( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	char *msg = NULL;
	int scale = *s1;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((flt)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == int_nil) {
				*o = flt_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((flt)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batint_dec2dec_flt( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		flt r;
		msg = int_dec2dec_flt( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batint_num2dec_flt( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_num2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		flt r;
		msg = int_num2dec_flt( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
wrd_2_flt( flt *res, wrd *v )
{
	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the wrd type, it will
	   always fit */
	*res = (flt)*v;
	return(MAL_SUCCEED);
}

str batwrd_2_flt( int *res, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (flt)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == wrd_nil) {
			*o = flt_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (flt)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
wrd_dec2_flt( flt *res, int *s1, wrd *v )
{
	int scale = *s1;
	flt r;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the wrd type, it will
	   always fit */
	r = (flt)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
wrd_dec2dec_flt( flt *res, int *S1, wrd *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	wrd cpyval = *v;
	int s1 = *S1, s2 = *S2;
	flt r;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the flt type is bigger than or equal to the wrd type, it will
	   always fit */
	r = (flt)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
wrd_num2dec_flt( flt *res, wrd *v, int *d2, int *s2 )
{
	int zero = 0;
	return wrd_dec2dec_flt( res, &zero, v, d2, s2 );
}

str batwrd_dec2_flt( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	int scale = *s1;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((flt)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = flt_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((flt)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batwrd_dec2dec_flt( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		flt r;
		msg = wrd_dec2dec_flt( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batwrd_num2dec_flt( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_num2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		flt r;
		msg = wrd_num2dec_flt( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
lng_2_flt( flt *res, lng *v )
{
	/* shortcut nil */
	if (*v == lng_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the lng type, it will
	   always fit */
	*res = (flt)*v;
	return(MAL_SUCCEED);
}

str batlng_2_flt( int *res, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (flt)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == lng_nil) {
			*o = flt_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (flt)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
lng_dec2_flt( flt *res, int *s1, lng *v )
{
	int scale = *s1;
	flt r;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the flt type is bigger than or equal to the lng type, it will
	   always fit */
	r = (flt)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
lng_dec2dec_flt( flt *res, int *S1, lng *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng cpyval = *v;
	int s1 = *S1, s2 = *S2;
	flt r;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = flt_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the flt type is bigger than or equal to the lng type, it will
	   always fit */
	r = (flt)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
lng_num2dec_flt( flt *res, lng *v, int *d2, int *s2 )
{
	int zero = 0;
	return lng_dec2dec_flt( res, &zero, v, d2, s2 );
}

str batlng_dec2_flt( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	int scale = *s1;
	flt *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2_flt", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_flt, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (flt*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((flt)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = flt_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((flt)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batlng_dec2dec_flt( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		flt r;
		msg = lng_dec2dec_flt( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batlng_num2dec_flt( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_num2dec_flt", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_flt, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		flt r;
		msg = lng_num2dec_flt( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
bte_2_dbl( dbl *res, bte *v )
{
	/* shortcut nil */
	if (*v == bte_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the bte type, it will
	   always fit */
	*res = (dbl)*v;
	return(MAL_SUCCEED);
}

str batbte_2_dbl( int *res, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (dbl)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == bte_nil) {
			*o = dbl_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (dbl)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
bte_dec2_dbl( dbl *res, int *s1, bte *v )
{
	int scale = *s1;
	dbl r;

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the bte type, it will
	   always fit */
	r = (dbl)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
bte_dec2dec_dbl( dbl *res, int *S1, bte *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	bte cpyval = *v;
	int s1 = *S1, s2 = *S2;
	dbl r;

	/* shortcut nil */
	if (*v == bte_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the dbl type is bigger than or equal to the bte type, it will
	   always fit */
	r = (dbl)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
bte_num2dec_dbl( dbl *res, bte *v, int *d2, int *s2 )
{
	int zero = 0;
	return bte_dec2dec_dbl( res, &zero, v, d2, s2 );
}

str batbte_dec2_dbl( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	bte *p,*q;
	char *msg = NULL;
	int scale = *s1;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (bte*) Tloc(b, BUNfirst(b));
	q = (bte*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((dbl)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == bte_nil) {
				*o = dbl_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((dbl)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batbte_dec2dec_dbl( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_dec2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		dbl r;
		msg = bte_dec2dec_dbl( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batbte_num2dec_dbl( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.bte_num2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		bte *v = (bte*)BUNtail(bi,p);
		dbl r;
		msg = bte_num2dec_dbl( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
sht_2_dbl( dbl *res, sht *v )
{
	/* shortcut nil */
	if (*v == sht_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the sht type, it will
	   always fit */
	*res = (dbl)*v;
	return(MAL_SUCCEED);
}

str batsht_2_dbl( int *res, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (dbl)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == sht_nil) {
			*o = dbl_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (dbl)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
sht_dec2_dbl( dbl *res, int *s1, sht *v )
{
	int scale = *s1;
	dbl r;

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the sht type, it will
	   always fit */
	r = (dbl)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
sht_dec2dec_dbl( dbl *res, int *S1, sht *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	sht cpyval = *v;
	int s1 = *S1, s2 = *S2;
	dbl r;

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the dbl type is bigger than or equal to the sht type, it will
	   always fit */
	r = (dbl)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
sht_num2dec_dbl( dbl *res, sht *v, int *d2, int *s2 )
{
	int zero = 0;
	return sht_dec2dec_dbl( res, &zero, v, d2, s2 );
}

str batsht_dec2_dbl( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	char *msg = NULL;
	int scale = *s1;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((dbl)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == sht_nil) {
				*o = dbl_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((dbl)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batsht_dec2dec_dbl( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		dbl r;
		msg = sht_dec2dec_dbl( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batsht_num2dec_dbl( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_num2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		dbl r;
		msg = sht_num2dec_dbl( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
int_2_dbl( dbl *res, int *v )
{
	/* shortcut nil */
	if (*v == int_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the int type, it will
	   always fit */
	*res = (dbl)*v;
	return(MAL_SUCCEED);
}

str batint_2_dbl( int *res, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (dbl)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == int_nil) {
			*o = dbl_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (dbl)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
int_dec2_dbl( dbl *res, int *s1, int *v )
{
	int scale = *s1;
	dbl r;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the int type, it will
	   always fit */
	r = (dbl)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
int_dec2dec_dbl( dbl *res, int *S1, int *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	int cpyval = *v;
	int s1 = *S1, s2 = *S2;
	dbl r;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the dbl type is bigger than or equal to the int type, it will
	   always fit */
	r = (dbl)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
int_num2dec_dbl( dbl *res, int *v, int *d2, int *s2 )
{
	int zero = 0;
	return int_dec2dec_dbl( res, &zero, v, d2, s2 );
}

str batint_dec2_dbl( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	char *msg = NULL;
	int scale = *s1;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((dbl)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == int_nil) {
				*o = dbl_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((dbl)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batint_dec2dec_dbl( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		dbl r;
		msg = int_dec2dec_dbl( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batint_num2dec_dbl( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_num2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		dbl r;
		msg = int_num2dec_dbl( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
wrd_2_dbl( dbl *res, wrd *v )
{
	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the wrd type, it will
	   always fit */
	*res = (dbl)*v;
	return(MAL_SUCCEED);
}

str batwrd_2_dbl( int *res, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (dbl)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == wrd_nil) {
			*o = dbl_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (dbl)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
wrd_dec2_dbl( dbl *res, int *s1, wrd *v )
{
	int scale = *s1;
	dbl r;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the wrd type, it will
	   always fit */
	r = (dbl)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
wrd_dec2dec_dbl( dbl *res, int *S1, wrd *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	wrd cpyval = *v;
	int s1 = *S1, s2 = *S2;
	dbl r;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the dbl type is bigger than or equal to the wrd type, it will
	   always fit */
	r = (dbl)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
wrd_num2dec_dbl( dbl *res, wrd *v, int *d2, int *s2 )
{
	int zero = 0;
	return wrd_dec2dec_dbl( res, &zero, v, d2, s2 );
}

str batwrd_dec2_dbl( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	int scale = *s1;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((dbl)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = dbl_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((dbl)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batwrd_dec2dec_dbl( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		dbl r;
		msg = wrd_dec2dec_dbl( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batwrd_num2dec_dbl( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_num2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		dbl r;
		msg = wrd_num2dec_dbl( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str 
lng_2_dbl( dbl *res, lng *v )
{
	/* shortcut nil */
	if (*v == lng_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the lng type, it will
	   always fit */
	*res = (dbl)*v;
	return(MAL_SUCCEED);
}

str batlng_2_dbl( int *res, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++)
			*o = (dbl)*p;
	} else{
		for (; p<q; p++, o++)
		if (*p == lng_nil) {
			*o = dbl_nil;
			bn->T->nonil= FALSE;
		} else
			*o = (dbl)*p;
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return MAL_SUCCEED;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return MAL_SUCCEED;
}




str 
lng_dec2_dbl( dbl *res, int *s1, lng *v )
{
	int scale = *s1;
	dbl r;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* since the dbl type is bigger than or equal to the lng type, it will
	   always fit */
	r = (dbl)*v;
	if (scale) 
		r /= scales[scale];
	*res = r;
	return MAL_SUCCEED;
}

str 
lng_dec2dec_dbl( dbl *res, int *S1, lng *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng cpyval = *v;
	int s1 = *S1, s2 = *S2;
	dbl r;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = dbl_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "convert",
			"too many digits (%d > %d)", inlen, p);
	}

	/* since the dbl type is bigger than or equal to the lng type, it will
	   always fit */
	r = (dbl)*v;
	if(s2 > s1) 
		r *= scales[s2-s1];
	else if (s2 != s1) 
		r /= scales[s1-s2];
	*res = r;
	return MAL_SUCCEED;
}

str 
lng_num2dec_dbl( dbl *res, lng *v, int *d2, int *s2 )
{
	int zero = 0;
	return lng_dec2dec_dbl( res, &zero, v, d2, s2 );
}

str batlng_dec2_dbl( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	int scale = *s1;
	dbl *o;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2_dbl", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_dbl, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (dbl*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) 
			*o = (((dbl)*p)/scales[scale]);
	} else {
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = dbl_nil;
				bn->T->nonil= FALSE;
			} else 
				*o = (((dbl)*p)/scales[scale]);
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batlng_dec2dec_dbl( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		dbl r;
		msg = lng_dec2dec_dbl( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batlng_num2dec_dbl( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_num2dec_dbl", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_dbl, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		dbl r;
		msg = lng_num2dec_dbl( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}












str sht_2_bte( bte *res, sht *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(bte)val > (lng) GDK_bte_min && 
	    val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
		*res = (bte)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str batsht_2_bte( int *res, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	char *msg = NULL;
	bte *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(bte)val > (lng) GDK_bte_min && val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
				*o = (bte)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == sht_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(bte)val > (lng) GDK_bte_min && val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
					*o = (bte)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
sht_dec2_bte( bte *res, int *s1, sht *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_bte_min && val <= GDK_bte_max)	{
		*res = (bte)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str 
sht_dec2dec_bte( bte *res, int *S1, sht *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == sht_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "sht_2_bte",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_bte_min && val <= GDK_bte_max)	{
		*res = (bte)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str 
sht_num2dec_bte( bte *res, sht *v, int *d2, int *s2 )
{
	int zero = 0;
	return sht_dec2dec_bte( res, &zero, v, d2, s2 );
}

str batsht_dec2_bte( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	sht *p,*q;
	char *msg = NULL;
	int scale = *s1;
	bte *o;
	sht val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (sht*) Tloc(b, BUNfirst(b));
	q = (sht*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (sht) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (sht) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_bte_min && val <= GDK_bte_max)
				*o = (bte)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == sht_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (sht) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (sht) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_bte_min && val <= GDK_bte_max)
					*o = (bte)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batsht_dec2dec_bte( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_dec2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		bte r;
		msg = sht_dec2dec_bte( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batsht_num2dec_bte( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.sht_num2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		sht *v = (sht*)BUNtail(bi,p);
		bte r;
		msg = sht_num2dec_bte( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str int_2_bte( bte *res, int *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(bte)val > (lng) GDK_bte_min && 
	    val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
		*res = (bte)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str batint_2_bte( int *res, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	char *msg = NULL;
	bte *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(bte)val > (lng) GDK_bte_min && val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
				*o = (bte)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == int_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(bte)val > (lng) GDK_bte_min && val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
					*o = (bte)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
int_dec2_bte( bte *res, int *s1, int *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_bte_min && val <= GDK_bte_max)	{
		*res = (bte)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str 
int_dec2dec_bte( bte *res, int *S1, int *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "int_2_bte",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_bte_min && val <= GDK_bte_max)	{
		*res = (bte)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str 
int_num2dec_bte( bte *res, int *v, int *d2, int *s2 )
{
	int zero = 0;
	return int_dec2dec_bte( res, &zero, v, d2, s2 );
}

str batint_dec2_bte( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	char *msg = NULL;
	int scale = *s1;
	bte *o;
	int val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (int) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (int) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_bte_min && val <= GDK_bte_max)
				*o = (bte)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == int_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (int) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (int) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_bte_min && val <= GDK_bte_max)
					*o = (bte)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batint_dec2dec_bte( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		bte r;
		msg = int_dec2dec_bte( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batint_num2dec_bte( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_num2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		bte r;
		msg = int_num2dec_bte( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str wrd_2_bte( bte *res, wrd *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(bte)val > (lng) GDK_bte_min && 
	    val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
		*res = (bte)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str batwrd_2_bte( int *res, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	bte *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(bte)val > (lng) GDK_bte_min && val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
				*o = (bte)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(bte)val > (lng) GDK_bte_min && val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
					*o = (bte)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
wrd_dec2_bte( bte *res, int *s1, wrd *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_bte_min && val <= GDK_bte_max)	{
		*res = (bte)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str 
wrd_dec2dec_bte( bte *res, int *S1, wrd *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "wrd_2_bte",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_bte_min && val <= GDK_bte_max)	{
		*res = (bte)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str 
wrd_num2dec_bte( bte *res, wrd *v, int *d2, int *s2 )
{
	int zero = 0;
	return wrd_dec2dec_bte( res, &zero, v, d2, s2 );
}

str batwrd_dec2_bte( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	int scale = *s1;
	bte *o;
	wrd val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (wrd) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (wrd) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_bte_min && val <= GDK_bte_max)
				*o = (bte)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (wrd) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (wrd) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_bte_min && val <= GDK_bte_max)
					*o = (bte)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batwrd_dec2dec_bte( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		bte r;
		msg = wrd_dec2dec_bte( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batwrd_num2dec_bte( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_num2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		bte r;
		msg = wrd_num2dec_bte( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str lng_2_bte( bte *res, lng *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(bte)val > (lng) GDK_bte_min && 
	    val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
		*res = (bte)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str batlng_2_bte( int *res, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	bte *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(bte)val > (lng) GDK_bte_min && val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
				*o = (bte)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(bte)val > (lng) GDK_bte_min && val > (lng) GDK_bte_min && val <= (lng) GDK_bte_max)	{
					*o = (bte)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
lng_dec2_bte( bte *res, int *s1, lng *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_bte_min && val <= GDK_bte_max)	{
		*res = (bte)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str 
lng_dec2dec_bte( bte *res, int *S1, lng *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = bte_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "lng_2_bte",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_bte_min && val <= GDK_bte_max)	{
		*res = (bte)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type bte", val);
	}
}

str 
lng_num2dec_bte( bte *res, lng *v, int *d2, int *s2 )
{
	int zero = 0;
	return lng_dec2dec_bte( res, &zero, v, d2, s2 );
}

str batlng_dec2_bte( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	int scale = *s1;
	bte *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2_bte", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_bte, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (bte*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (lng) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (lng) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_bte_min && val <= GDK_bte_max)
				*o = (bte)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = bte_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (lng) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (lng) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_bte_min && val <= GDK_bte_max)
					*o = (bte)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type bte", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batlng_dec2dec_bte( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		bte r;
		msg = lng_dec2dec_bte( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batlng_num2dec_bte( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_num2dec_bte", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_bte, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		bte r;
		msg = lng_num2dec_bte( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str int_2_sht( sht *res, int *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(sht)val > (lng) GDK_sht_min && 
	    val > (lng) GDK_sht_min && val <= (lng) GDK_sht_max)	{
		*res = (sht)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type sht", val);
	}
}

str batint_2_sht( int *res, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	char *msg = NULL;
	sht *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(sht)val > (lng) GDK_sht_min && val > (lng) GDK_sht_min && val <= (lng) GDK_sht_max)	{
				*o = (sht)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == int_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(sht)val > (lng) GDK_sht_min && val > (lng) GDK_sht_min && val <= (lng) GDK_sht_max)	{
					*o = (sht)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
int_dec2_sht( sht *res, int *s1, int *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_sht_min && val <= GDK_sht_max)	{
		*res = (sht)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type sht", val);
	}
}

str 
int_dec2dec_sht( sht *res, int *S1, int *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == int_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "int_2_sht",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_sht_min && val <= GDK_sht_max)	{
		*res = (sht)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type sht", val);
	}
}

str 
int_num2dec_sht( sht *res, int *v, int *d2, int *s2 )
{
	int zero = 0;
	return int_dec2dec_sht( res, &zero, v, d2, s2 );
}

str batint_dec2_sht( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	int *p,*q;
	char *msg = NULL;
	int scale = *s1;
	sht *o;
	int val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (int*) Tloc(b, BUNfirst(b));
	q = (int*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (int) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (int) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_sht_min && val <= GDK_sht_max)
				*o = (sht)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == int_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (int) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (int) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_sht_min && val <= GDK_sht_max)
					*o = (sht)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batint_dec2dec_sht( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_dec2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		sht r;
		msg = int_dec2dec_sht( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batint_num2dec_sht( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.int_num2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		int *v = (int*)BUNtail(bi,p);
		sht r;
		msg = int_num2dec_sht( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str wrd_2_sht( sht *res, wrd *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(sht)val > (lng) GDK_sht_min && 
	    val > (lng) GDK_sht_min && val <= (lng) GDK_sht_max)	{
		*res = (sht)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type sht", val);
	}
}

str batwrd_2_sht( int *res, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	sht *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(sht)val > (lng) GDK_sht_min && val > (lng) GDK_sht_min && val <= (lng) GDK_sht_max)	{
				*o = (sht)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(sht)val > (lng) GDK_sht_min && val > (lng) GDK_sht_min && val <= (lng) GDK_sht_max)	{
					*o = (sht)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
wrd_dec2_sht( sht *res, int *s1, wrd *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_sht_min && val <= GDK_sht_max)	{
		*res = (sht)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type sht", val);
	}
}

str 
wrd_dec2dec_sht( sht *res, int *S1, wrd *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "wrd_2_sht",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_sht_min && val <= GDK_sht_max)	{
		*res = (sht)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type sht", val);
	}
}

str 
wrd_num2dec_sht( sht *res, wrd *v, int *d2, int *s2 )
{
	int zero = 0;
	return wrd_dec2dec_sht( res, &zero, v, d2, s2 );
}

str batwrd_dec2_sht( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	int scale = *s1;
	sht *o;
	wrd val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (wrd) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (wrd) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_sht_min && val <= GDK_sht_max)
				*o = (sht)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (wrd) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (wrd) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_sht_min && val <= GDK_sht_max)
					*o = (sht)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batwrd_dec2dec_sht( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		sht r;
		msg = wrd_dec2dec_sht( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batwrd_num2dec_sht( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_num2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		sht r;
		msg = wrd_num2dec_sht( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str lng_2_sht( sht *res, lng *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(sht)val > (lng) GDK_sht_min && 
	    val > (lng) GDK_sht_min && val <= (lng) GDK_sht_max)	{
		*res = (sht)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type sht", val);
	}
}

str batlng_2_sht( int *res, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	sht *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(sht)val > (lng) GDK_sht_min && val > (lng) GDK_sht_min && val <= (lng) GDK_sht_max)	{
				*o = (sht)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(sht)val > (lng) GDK_sht_min && val > (lng) GDK_sht_min && val <= (lng) GDK_sht_max)	{
					*o = (sht)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
lng_dec2_sht( sht *res, int *s1, lng *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_sht_min && val <= GDK_sht_max)	{
		*res = (sht)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type sht", val);
	}
}

str 
lng_dec2dec_sht( sht *res, int *S1, lng *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = sht_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "lng_2_sht",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_sht_min && val <= GDK_sht_max)	{
		*res = (sht)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type sht", val);
	}
}

str 
lng_num2dec_sht( sht *res, lng *v, int *d2, int *s2 )
{
	int zero = 0;
	return lng_dec2dec_sht( res, &zero, v, d2, s2 );
}

str batlng_dec2_sht( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	int scale = *s1;
	sht *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2_sht", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_sht, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (sht*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (lng) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (lng) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_sht_min && val <= GDK_sht_max)
				*o = (sht)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = sht_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (lng) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (lng) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_sht_min && val <= GDK_sht_max)
					*o = (sht)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type sht", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batlng_dec2dec_sht( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		sht r;
		msg = lng_dec2dec_sht( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batlng_num2dec_sht( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_num2dec_sht", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_sht, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		sht r;
		msg = lng_num2dec_sht( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str wrd_2_int( int *res, wrd *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(int)val > (lng) GDK_int_min && 
	    val > (lng) GDK_int_min && val <= (lng) GDK_int_max)	{
		*res = (int)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type int", val);
	}
}

str batwrd_2_int( int *res, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	int *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(int)val > (lng) GDK_int_min && val > (lng) GDK_int_min && val <= (lng) GDK_int_max)	{
				*o = (int)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type int", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = int_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(int)val > (lng) GDK_int_min && val > (lng) GDK_int_min && val <= (lng) GDK_int_max)	{
					*o = (int)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type int", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
wrd_dec2_int( int *res, int *s1, wrd *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_int_min && val <= GDK_int_max)	{
		*res = (int)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type int", val);
	}
}

str 
wrd_dec2dec_int( int *res, int *S1, wrd *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == wrd_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "wrd_2_int",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_int_min && val <= GDK_int_max)	{
		*res = (int)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type int", val);
	}
}

str 
wrd_num2dec_int( int *res, wrd *v, int *d2, int *s2 )
{
	int zero = 0;
	return wrd_dec2dec_int( res, &zero, v, d2, s2 );
}

str batwrd_dec2_int( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	wrd *p,*q;
	char *msg = NULL;
	int scale = *s1;
	int *o;
	wrd val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (wrd*) Tloc(b, BUNfirst(b));
	q = (wrd*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (wrd) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (wrd) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_int_min && val <= GDK_int_max)
				*o = (int)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type int", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == wrd_nil) {
				*o = int_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (wrd) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (wrd) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_int_min && val <= GDK_int_max)
					*o = (int)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type int", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batwrd_dec2dec_int( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_dec2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		int r;
		msg = wrd_dec2dec_int( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batwrd_num2dec_int( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.wrd_num2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		wrd *v = (wrd*)BUNtail(bi,p);
		int r;
		msg = wrd_num2dec_int( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str lng_2_int( int *res, lng *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(int)val > (lng) GDK_int_min && 
	    val > (lng) GDK_int_min && val <= (lng) GDK_int_max)	{
		*res = (int)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type int", val);
	}
}

str batlng_2_int( int *res, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	int *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(int)val > (lng) GDK_int_min && val > (lng) GDK_int_min && val <= (lng) GDK_int_max)	{
				*o = (int)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type int", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = int_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(int)val > (lng) GDK_int_min && val > (lng) GDK_int_min && val <= (lng) GDK_int_max)	{
					*o = (int)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type int", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
lng_dec2_int( int *res, int *s1, lng *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_int_min && val <= GDK_int_max)	{
		*res = (int)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type int", val);
	}
}

str 
lng_dec2dec_int( int *res, int *S1, lng *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = int_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "lng_2_int",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_int_min && val <= GDK_int_max)	{
		*res = (int)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type int", val);
	}
}

str 
lng_num2dec_int( int *res, lng *v, int *d2, int *s2 )
{
	int zero = 0;
	return lng_dec2dec_int( res, &zero, v, d2, s2 );
}

str batlng_dec2_int( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	int scale = *s1;
	int *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2_int", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_int, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (int*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (lng) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (lng) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_int_min && val <= GDK_int_max)
				*o = (int)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type int", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = int_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (lng) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (lng) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_int_min && val <= GDK_int_max)
					*o = (int)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type int", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batlng_dec2dec_int( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		int r;
		msg = lng_dec2dec_int( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batlng_num2dec_int( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_num2dec_int", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_int, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		int r;
		msg = lng_num2dec_int( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}







str lng_2_wrd( wrd *res, lng *v )
{
	lng val = *v;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* see if the number fits in the data type */
	if ((lng)(wrd)val > (lng) GDK_wrd_min && 
	    val > (lng) GDK_wrd_min && val <= (lng) GDK_wrd_max)	{
		*res = (wrd)val;
		return(MAL_SUCCEED);
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type wrd", val);
	}
}

str batlng_2_wrd( int *res, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	wrd *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	BATseqbase(bn, b->hseqbase);
	bn->H->nonil = 1;
	bn->T->nonil = 1;
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	if ( b->T->nonil){
		for (; p<q; p++, o++){
			val = *p;
			/* see if the number fits in the data type */
			if ((lng)(wrd)val > (lng) GDK_wrd_min && val > (lng) GDK_wrd_min && val <= (lng) GDK_wrd_max)	{
				*o = (wrd)val;
			} else {
				msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type wrd", val);
				break;
			}
		}
	} else {
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = wrd_nil;
				bn->T->nonil= FALSE;
			} else {
				val = *p;
				/* see if the number fits in the data type */
				if ((lng)(wrd)val > (lng) GDK_wrd_min && val > (lng) GDK_wrd_min && val <= (lng) GDK_wrd_max)	{
					*o = (wrd)val;
				} else {
					msg= createException(SQL, "convert", "value (" LLFMT ") exceeds limits of type wrd", val);
					break;
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}




str 
lng_dec2_wrd( wrd *res, int *s1, lng *v )
{
	int scale = *s1;
	lng val = *v, h = (val<0)?-5:5;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	if (scale) 
		val = (val+h)/scales[scale];
	/* see if the number fits in the data type */
	if (val > GDK_wrd_min && val <= GDK_wrd_max)	{
		*res = (wrd)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type wrd", val);
	}
}

str 
lng_dec2dec_wrd( wrd *res, int *S1, lng *v, int *d2, int *S2 )
{
	int p = *d2, inlen = 1;
	lng val = *v, cpyval = val, h = (val<0)?-5:5;
	int s1 = *S1, s2 = *S2;

	/* shortcut nil */
	if (*v == lng_nil) {
		*res = wrd_nil;
		return(MAL_SUCCEED);
	}
	
	/* count the number of digits in the input */
	while (cpyval /= 10)
		inlen++;
	/* rounding is allowed */
	inlen += (s2-s1);
	if (p && inlen > p) {
		throw(SQL, "lng_2_wrd",
			"too many digits (%d > %d)", inlen, p);
	}

	if(s2 > s1) 
		val *= scales[s2-s1];
	else if (s2 != s1) 
		val = (val+h)/scales[s1-s2];

	/* see if the number fits in the data type */
	if (val > GDK_wrd_min && val <= GDK_wrd_max)	{
		*res = (wrd)val;
		return MAL_SUCCEED;
	} else {
		throw(SQL, "convert",
			"value (" LLFMT ") exceeds limits of type wrd", val);
	}
}

str 
lng_num2dec_wrd( wrd *res, lng *v, int *d2, int *s2 )
{
	int zero = 0;
	return lng_dec2dec_wrd( res, &zero, v, d2, s2 );
}

str batlng_dec2_wrd( int *res, int *s1, int *bid )
{
	BAT *b, *bn;
	lng *p,*q;
	char *msg = NULL;
	int scale = *s1;
	wrd *o;
	lng val;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2_wrd", "Cannot access descriptor");
	}
	bn = BATnew(TYPE_void, TYPE_wrd, BATcount(b));
	bn->hsorted = b->hsorted;
	BATseqbase(bn, b->hseqbase);
	o = (wrd*) Tloc(bn,BUNfirst(bn));
	p = (lng*) Tloc(b, BUNfirst(b));
	q = (lng*) Tloc(b, BUNlast(b));
	bn->T->nonil = 1;
	if ( b->T->nonil){
		for (; p<q; p++, o++) {
			if (scale) 
				val = (lng) ((*p+ (*p<0)?-5:5)/scales[scale]);
			else
				val = (lng) (*p);
			/* see if the number fits in the data type */
			if (val > GDK_wrd_min && val <= GDK_wrd_max)
				*o = (wrd)val;
			else {
				BBPreleaseref(b->batCacheid);
				BBPreleaseref(bn->batCacheid);
				throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type wrd", val);
			}
		}
	} else{
		for (; p<q; p++, o++) {
			if (*p == lng_nil) {
				*o = wrd_nil;
				bn->T->nonil= FALSE;
			} else {
				if (scale)
					val = (lng) (( *p + ( *p<0)?-5:5)/scales[scale]);
				else
					val = (lng) (*p);
				/* see if the number fits in the data type */
				if (val > GDK_wrd_min && val <= GDK_wrd_max)
					*o = (wrd)val;
				else {
					BBPreleaseref(b->batCacheid);
					BBPreleaseref(bn->batCacheid);
					throw(SQL, "convert", "value (" LLFMT ") exceeds limits of type wrd", val);
				}
			}
		}
	}
	BATsetcount(bn, BATcount(b));
	bn->tsorted = 0;
	BATkey(BATmirror(bn),FALSE);

	if (!(bn->batDirty&2)) bn = BATsetaccess(bn, BAT_READ);

	if (b->htype != bn->htype) {
		BAT *r = VIEWcreate(b,bn);

		BBPkeepref(*res = r->batCacheid);
		BBPreleaseref(bn->batCacheid);
		BBPreleaseref(b->batCacheid);
		return msg;
	}
	BBPkeepref(*res = bn->batCacheid);
	BBPreleaseref(b->batCacheid);
	return msg;
}

str batlng_dec2dec_wrd( int *res, int *S1, int *bid, int *d2, int *S2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_dec2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		wrd r;
		msg = lng_dec2dec_wrd( &r, S1, v, d2, S2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}

str batlng_num2dec_wrd( int *res, int *bid, int *d2, int *s2 )
{
	BAT *b, *dst;
	BATiter bi;
	BUN p,q;
	char *msg = NULL;

	if( (b = BATdescriptor(*bid)) == NULL ){
		throw(SQL, "batcalc.lng_num2dec_wrd", "Cannot access descriptor");
	}
	bi = bat_iterator(b);
	dst = BATnew(b->htype, TYPE_wrd, BATcount(b));
	BATseqbase(dst, b->hseqbase);
	BATloop(b,p,q) {
		lng *v = (lng*)BUNtail(bi,p);
		wrd r;
		msg = lng_num2dec_wrd( &r, v, d2, s2 );
		if (msg)
			break;
		BUNins(dst, BUNhead(bi,p), &r, FALSE);
	}
	BBPkeepref( *res = dst->batCacheid);
	BBPunfix(b->batCacheid);
	return msg;
}






str 
month_interval_str( int *ret, str *s, int *d, int *sk )
{
	lng res;

	(void)sk;
	if (interval_from_str( *s, *d, &res ) < 0)
		throw(SQL, "calc.month_interval",
			"wrong format (%s)", *s);
	assert((lng) GDK_int_min <= res && res <= (lng) GDK_int_max);
	*ret = (int) res;
	return MAL_SUCCEED;
}

str 
second_interval_str( lng *res, str *s, int *d, int *sk )
{
	(void)sk;
	if (interval_from_str( *s, *d, res ) < 0)
		throw(SQL, "calc.second_interval",
			"wrong format (%s)", *s);
	return MAL_SUCCEED;
}




str
month_interval_bte( int *ret, bte *s, int *d, int *sk ){
	int k = digits2sk(*d);
	int r = (int)*s;

	(void)sk;
	switch(k) {
	case iyear: 	
		r *= 12;
	case imonth:
		break;
	default:
		throw(ILLARG, "calc.month_interval","illegal argument");
	}
	*ret = r;
	return MAL_SUCCEED;
}
str
second_interval_bte( lng *res, bte *s, int *d, int *sk ){
	int k = digits2sk(*d);
	lng r = *s;

	(void)sk;
	switch(k) {
	case iday: 	
		r *= 24;
	case ihour:
		r *= 60;
	case imin:
		r *= 60;
	case isec:
		break;
	default:
		throw(ILLARG, "calc.second_interval","illegal argument");
	}
	*res = r;
	return MAL_SUCCEED;
}




str
month_interval_sht( int *ret, sht *s, int *d, int *sk ){
	int k = digits2sk(*d);
	int r = (int)*s;

	(void)sk;
	switch(k) {
	case iyear: 	
		r *= 12;
	case imonth:
		break;
	default:
		throw(ILLARG, "calc.month_interval","illegal argument");
	}
	*ret = r;
	return MAL_SUCCEED;
}
str
second_interval_sht( lng *res, sht *s, int *d, int *sk ){
	int k = digits2sk(*d);
	lng r = *s;

	(void)sk;
	switch(k) {
	case iday: 	
		r *= 24;
	case ihour:
		r *= 60;
	case imin:
		r *= 60;
	case isec:
		break;
	default:
		throw(ILLARG, "calc.second_interval","illegal argument");
	}
	*res = r;
	return MAL_SUCCEED;
}




str
month_interval_int( int *ret, int *s, int *d, int *sk ){
	int k = digits2sk(*d);
	int r = (int)*s;

	(void)sk;
	switch(k) {
	case iyear: 	
		r *= 12;
	case imonth:
		break;
	default:
		throw(ILLARG, "calc.month_interval","illegal argument");
	}
	*ret = r;
	return MAL_SUCCEED;
}
str
second_interval_int( lng *res, int *s, int *d, int *sk ){
	int k = digits2sk(*d);
	lng r = *s;

	(void)sk;
	switch(k) {
	case iday: 	
		r *= 24;
	case ihour:
		r *= 60;
	case imin:
		r *= 60;
	case isec:
		break;
	default:
		throw(ILLARG, "calc.second_interval","illegal argument");
	}
	*res = r;
	return MAL_SUCCEED;
}




str
month_interval_wrd( int *ret, wrd *s, int *d, int *sk ){
	int k = digits2sk(*d);
	int r = (int)*s;

	(void)sk;
	switch(k) {
	case iyear: 	
		r *= 12;
	case imonth:
		break;
	default:
		throw(ILLARG, "calc.month_interval","illegal argument");
	}
	*ret = r;
	return MAL_SUCCEED;
}
str
second_interval_wrd( lng *res, wrd *s, int *d, int *sk ){
	int k = digits2sk(*d);
	lng r = *s;

	(void)sk;
	switch(k) {
	case iday: 	
		r *= 24;
	case ihour:
		r *= 60;
	case imin:
		r *= 60;
	case isec:
		break;
	default:
		throw(ILLARG, "calc.second_interval","illegal argument");
	}
	*res = r;
	return MAL_SUCCEED;
}




str
month_interval_lng( int *ret, lng *s, int *d, int *sk ){
	int k = digits2sk(*d);
	int r = (int)*s;

	(void)sk;
	switch(k) {
	case iyear: 	
		r *= 12;
	case imonth:
		break;
	default:
		throw(ILLARG, "calc.month_interval","illegal argument");
	}
	*ret = r;
	return MAL_SUCCEED;
}
str
second_interval_lng( lng *res, lng *s, int *d, int *sk ){
	int k = digits2sk(*d);
	lng r = *s;

	(void)sk;
	switch(k) {
	case iday: 	
		r *= 24;
	case ihour:
		r *= 60;
	case imin:
		r *= 60;
	case isec:
		break;
	default:
		throw(ILLARG, "calc.second_interval","illegal argument");
	}
	*res = r;
	return MAL_SUCCEED;
}




str
second_interval_daytime( lng *res, daytime *s, int *d, int *sk )
{
	int k = digits2sk(*d);
	lng r = *(int*)s;

	(void)sk;
	switch(k) {
	case isec:
		r /= 1000;
		break;
	case imin:
		r /= 60000;
		break;
	case ihour:
		r /= 3600000;
		break;
	case iday: 	
		r /= (24*3600000);
		break;
	default:
		throw(ILLARG, "calc.second_interval","illegal argument");
	}
	*res = r;
	return MAL_SUCCEED;
}

/* str dump_cache(int *r); */
str
dump_cache(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	int cnt; 
	BAT *t, *query, *count;
	cq *q = NULL;
	int *r = (int *) getArgReference(stk, pci, 0);

	(void) cntxt;
	if (msg)
		return msg;
	cnt = m->qc->id; 
 	t = BATnew(TYPE_str, TYPE_bat, 1);
	query = BATnew(TYPE_void, TYPE_str, cnt);
	BATseqbase(query, 0);
	count = BATnew(TYPE_void, TYPE_int, cnt);
	BATseqbase(count, 0);
	BUNins(t, "query", &query->batCacheid, FALSE);
	BUNins(t, "count", &count->batCacheid, FALSE);

	for(q = m->qc->q; q; q = q->next) {
		BUNappend(query, q->codestring, FALSE);
		BUNappend(count, &q->count, FALSE);
	}

	BBPunfix(query->batCacheid);
	BBPunfix(count->batCacheid);
	*r = t->batCacheid;
	BBPkeepref(*r);
	return MAL_SUCCEED;
}

/* str dump_opt_stats(int *r); */
str
dump_opt_stats(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	int cnt; 
	BAT *t, *rewrite, *count;
	int *r = (int *) getArgReference(stk, pci, 0);

	(void) cntxt;
	if (msg)
		return msg;
	cnt = m->qc->id; 
 	t = BATnew(TYPE_str, TYPE_bat, 1);
	rewrite = BATnew(TYPE_void, TYPE_str, cnt);
	BATseqbase(rewrite, 0);
	count = BATnew(TYPE_void, TYPE_int, cnt);
	BATseqbase(count, 0);
	BUNins(t, "rewrite", &rewrite->batCacheid, FALSE);
	BUNins(t, "count", &count->batCacheid, FALSE);

	BUNappend(rewrite, "joinidx", FALSE);
	BUNappend(count, &m->opt_stats[0], FALSE);
	/* TODO add other rewrites */

	BBPunfix(rewrite->batCacheid);
	BBPunfix(count->batCacheid);
	*r = t->batCacheid;
	BBPkeepref(*r);
	return MAL_SUCCEED;
}

/* str dump_opt_stats(int *r); */
str
dump_trace(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	BAT *t;
	int *r = (int *) getArgReference(stk, pci, 0);

	if (msg)
		return msg;
	t= TRACEtable();
	*r = t->batCacheid;
	BBPkeepref(*r);
	return MAL_SUCCEED;
}

/* str slave(int *r); */
str
dump_slave(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	BAT *t;
	int *r = (int *) getArgReference(stk, pci, 0);

	if (msg)
		return msg;
	t= SLAVEtable();
	if ( t == NULL || BATcount(t) == 0)
		throw(ILLARG, "sql.slavelog","tables not initialized");

	*r = t->batCacheid;
	BBPkeepref(*r);
	return MAL_SUCCEED;
}

/* str sql_rowid(oid *rid, ptr v, str *sname, str *tname); */
str 
sql_rowid(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	BAT *b;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_schema *s = NULL;
	sql_table *t = NULL;
	sql_column *c = NULL;
	sql_delta *d;
	oid *rid = (oid *) getArgReference(stk, pci, 0);
	str *sname = (str *) getArgReference(stk, pci, 2);
	str *tname = (str *) getArgReference(stk, pci, 3);

	(void) cntxt;
	if (msg)
		return msg;
	s = mvc_bind_schema(m, *sname);
	if ( s == NULL)
		throw(MAL,"sql.rowid","Schema missing");
	t = mvc_bind_table(m, s, *tname);
	if ( s == NULL)
		throw(MAL,"sql.rowid","Table missing");
	if (!s || !t || !t->columns.set->h)
                throw(SQL, "calc.rowid", "Cannot find column");
	c = t->columns.set->h->data;
	/* HACK, get insert bat */
	b = store_funcs.bind_col(m->session->tr, c, RD_INS);
	/* UGH (move into storage backends!!) */
	d = c->data;
	*rid = d->ibase + BATcount(b);
	BBPunfix(b->batCacheid);
	return MAL_SUCCEED;
}




str 
sql_rank_grp(bat *rid, bat *bid, bat *gid, bat *gpe)
{
	BAT *r, *b, *g;
	BUN p, q;
	BATiter bi, gi;
	int (*ocmp)(ptr, ptr), (*gcmp)(ptr, ptr);
	ptr oc, gc;
	int rank = 1, nrank = 1;

	if( (b = BATdescriptor(*bid)) == NULL )
		throw(SQL, "sql.rank_grp", "Cannot access descriptor");
	if( (g = BATdescriptor(*gid)) == NULL )
		throw(SQL, "sql.rank_grp", "Cannot access descriptor");
	(void)gpe;
	bi = bat_iterator(b);
	gi = bat_iterator(g);
	ocmp = BATatoms[b->ttype].atomCmp;
	gcmp = BATatoms[g->ttype].atomCmp;
	oc = BUNtail(bi, BUNfirst(b));
	gc = BUNtail(gi, BUNfirst(g));
	if (!ALIGNsynced(b, g))
		throw(SQL, "sql.rank_grp", "bats not aligned");
	if (BATtordered(b) != GDK_SORTED)
		throw(SQL, "sql.rank_grp", "bat not sorted");
	r = BATnew(TYPE_oid, TYPE_int, BATcount(b));
	if (r == NULL)
		throw(SQL, "sql.rank_grp", "cannot allocate result bat");
	BATloop(b, p, q) {
		ptr on = BUNtail(bi, p);
		ptr gn = BUNtail(gi, p);
		int c = 0;

		if ((c = ocmp(on, oc)) != 0) 
			rank = nrank;
		if (gcmp(gn, gc) != 0) 
			c = rank = nrank = 1;  
		oc = on;
		gc = gn;
		BUNins(r, BUNhead(bi,p), &rank, FALSE);
		(void)c;nrank++;
	}
	BBPunfix(b->batCacheid);
	BBPunfix(g->batCacheid);
	BBPkeepref(*rid = r->batCacheid);
	return MAL_SUCCEED;
}

str
sql_rank(bat *rid, bat *bid)
{ 
	BAT *r, *b;
	BATiter bi;
	int (*cmp)(ptr, ptr);
	ptr cur;
	BUN p, q;
	int rank = 1, nrank = 1;

	if( (b = BATdescriptor(*bid)) == NULL )
		throw(SQL, "sql.rank", "Cannot access descriptor");
	if (!(BATtordered(b) & GDK_SORTED) && !(BATtordered(b) & GDK_SORTED_REV))
		throw(SQL, "sql.rank", "bat not sorted");

	bi = bat_iterator(b);
	cmp = BATatoms[b->ttype].atomCmp;
	cur = BUNtail(bi, BUNfirst(b));
	r = BATnew(TYPE_oid, TYPE_int, BATcount(b));
	if (r == NULL)
		throw(SQL, "sql.rank", "cannot allocate result bat");
	if (BATtdense(b)) {
		BATloop(b, p, q) {
			BUNins(r, BUNhead(bi,p), &rank, FALSE);
			rank++;
		}
	} else {
		BATloop(b, p, q) {
			ptr n = BUNtail(bi, p);
			int c = 0;
	
			if ((c = cmp(n, cur)) != 0) 
				rank = nrank; 
			cur = n;
			BUNins(r, BUNhead(bi,p), &rank, FALSE);
			(void)c;nrank++;
		}
	}
	BBPunfix(b->batCacheid);
	BBPkeepref(*rid = r->batCacheid);
	return MAL_SUCCEED;
}




str 
sql_dense_rank_grp(bat *rid, bat *bid, bat *gid, bat *gpe)
{
	BAT *r, *b, *g;
	BUN p, q;
	BATiter bi, gi;
	int (*ocmp)(ptr, ptr), (*gcmp)(ptr, ptr);
	ptr oc, gc;
	int rank = 1, nrank = 2;

	if( (b = BATdescriptor(*bid)) == NULL )
		throw(SQL, "sql.dense_rank_grp", "Cannot access descriptor");
	if( (g = BATdescriptor(*gid)) == NULL )
		throw(SQL, "sql.dense_rank_grp", "Cannot access descriptor");
	(void)gpe;
	bi = bat_iterator(b);
	gi = bat_iterator(g);
	ocmp = BATatoms[b->ttype].atomCmp;
	gcmp = BATatoms[g->ttype].atomCmp;
	oc = BUNtail(bi, BUNfirst(b));
	gc = BUNtail(gi, BUNfirst(g));
	if (!ALIGNsynced(b, g))
		throw(SQL, "sql.dense_rank_grp", "bats not aligned");
	if (BATtordered(b) != GDK_SORTED)
		throw(SQL, "sql.dense_rank_grp", "bat not sorted");
	r = BATnew(TYPE_oid, TYPE_int, BATcount(b));
	if (r == NULL)
		throw(SQL, "sql.dense_rank_grp", "cannot allocate result bat");
	BATloop(b, p, q) {
		ptr on = BUNtail(bi, p);
		ptr gn = BUNtail(gi, p);
		int c = 0;

		if ((c = ocmp(on, oc)) != 0) 
			rank = nrank;
		if (gcmp(gn, gc) != 0) 
			c = rank = nrank = 1;  
		oc = on;
		gc = gn;
		BUNins(r, BUNhead(bi,p), &rank, FALSE);
		nrank+=(c?1:0);
	}
	BBPunfix(b->batCacheid);
	BBPunfix(g->batCacheid);
	BBPkeepref(*rid = r->batCacheid);
	return MAL_SUCCEED;
}

str
sql_dense_rank(bat *rid, bat *bid)
{ 
	BAT *r, *b;
	BATiter bi;
	int (*cmp)(ptr, ptr);
	ptr cur;
	BUN p, q;
	int rank = 1, nrank = 2;

	if( (b = BATdescriptor(*bid)) == NULL )
		throw(SQL, "sql.dense_rank", "Cannot access descriptor");
	if (!(BATtordered(b) & GDK_SORTED) && !(BATtordered(b) & GDK_SORTED_REV))
		throw(SQL, "sql.dense_rank", "bat not sorted");

	bi = bat_iterator(b);
	cmp = BATatoms[b->ttype].atomCmp;
	cur = BUNtail(bi, BUNfirst(b));
	r = BATnew(TYPE_oid, TYPE_int, BATcount(b));
	if (r == NULL)
		throw(SQL, "sql.dense_rank", "cannot allocate result bat");
	if (BATtdense(b)) {
		BATloop(b, p, q) {
			BUNins(r, BUNhead(bi,p), &rank, FALSE);
			rank++;
		}
	} else {
		BATloop(b, p, q) {
			ptr n = BUNtail(bi, p);
			int c = 0;
	
			if ((c = cmp(n, cur)) != 0) 
				rank = nrank; 
			cur = n;
			BUNins(r, BUNhead(bi,p), &rank, FALSE);
			nrank+=(c?1:0);
		}
	}
	BBPunfix(b->batCacheid);
	BBPkeepref(*rid = r->batCacheid);
	return MAL_SUCCEED;
}




str SQLargRecord(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci){
	str *ret;

	(void) cntxt;
	ret= (str *) getArgReference(stk,pci,0);
	*ret = instruction2str(mb,stk,getInstrPtr(mb,0),LIST_MAL_VALUE);
	return MAL_SUCCEED;
}




str SQLcluster(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci){
	str *sch = (str *) getArgReference(stk,pci,1);
	str *tbl = (str *) getArgReference(stk,pci,2);
	sql_trans	*tr;
	sql_schema	*s;
	sql_table 	*t;
	sql_column	*c;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	int first=1;
	bat mid, hid, bid;
	BAT *map= NULL, *b;
	node *o;

	if (msg)
		return msg;
   	s = mvc_bind_schema(m, *sch);
	if ( s == NULL)
		throw(MAL,"sql.cluster","Schema missing");
	t = mvc_bind_table(m, s, *tbl);
	if ( t == NULL)
		throw(MAL,"sql.cluster","Table missing");
	tr = m->session->tr;
	t->base.wtime = s->base.wtime = tr->wtime = tr->stime;
	t->base.rtime = s->base.rtime = tr->rtime = tr->stime;

	/* actually build the hash on the multi-column primary key */

	for (o = t->columns.set->h; o; o = o->next) {
		sql_delta *d;
		c = o->data;
		if (first) {
			first = 0;
			b = store_funcs.bind_col(tr, c, 0);
			msg = CLUSTER_key(&hid, &b->batCacheid);
			BBPreleaseref(b->batCacheid);
			if (msg)
				return msg;
			msg = CLUSTER_map(&mid,&hid);
			BBPdecref(hid, TRUE);
			if (msg)
				return msg;
			map = BATdescriptor(mid);
			if (map== NULL)
				throw(MAL,"sql.cluster","Can not access descriptor");
		}

		b = store_funcs.bind_col(tr, c, 0);
		if (b == NULL)
			throw(MAL,"sql.cluster","Can not access descriptor");
		msg = CLUSTER_apply(&bid, b, map);
		BBPreleaseref(b->batCacheid);
		if (msg){
			BBPreleaseref(map->batCacheid);
			return msg;
		}
		BBPdecref(b->batCacheid, TRUE);
		d = c->data;
		d->bid = 0;
		d->ibase = 0;
		d->ibid = bid; /* use the insert bat */
		c->base.wtime = c->base.rtime = tr->stime;
	}
	/* bat was cleared */
	t->cleared = 1;
	if (map) {
		BBPreleaseref(map->batCacheid);
		BBPdecref(mid, TRUE);
	}
	return MAL_SUCCEED;
}

str 
SQLcluster2(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
	str *sch = (str *) getArgReference(stk,pci,1);
	str *tbl = (str *) getArgReference(stk,pci,2);
	sql_trans	*tr;
	sql_schema	*s;
	sql_table 	*t;
	sql_column	*c;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	int first=1;
	bat mid, hid, bid;
	BAT *b;
	node *o;

	if (msg)
		return msg;
   	s = mvc_bind_schema(m, *sch);
	if ( s == NULL)
		throw(MAL,"sql.cluster","Schema missing");
	t = mvc_bind_table(m, s, *tbl);
	if ( t == NULL)
		throw(MAL,"sql.cluster","Table missing");
	tr = m->session->tr;

        t->base.wtime = s->base.wtime = tr->wtime = tr->stime;
        t->base.rtime = s->base.rtime = tr->rtime = tr->stime;
	for (o = t->columns.set->h; o; o = o->next) {
		sql_delta *d;
		c = o->data;
		if (first) {
			bat psum;
			unsigned int bits = 10, off = 0;
			first = 0;
			b = store_funcs.bind_col(tr, c, 0);
			msg = MKEYbathash(&hid, &b->batCacheid);
			BBPreleaseref(b->batCacheid);
			if (msg)
				return msg;
			msg = CLS_create_wrd(&psum, &mid, &hid, &bits, &off);
			BBPdecref(hid, TRUE);
			BBPdecref(psum, TRUE);
			if (msg)
				return msg;
		}

		b = store_funcs.bind_col(tr, c, 0);
		if (b == NULL)
			throw(MAL,"sql.cluster","Can not access descriptor");
		msg = CLS_map(&bid,&mid,&b->batCacheid);
		BBPreleaseref(b->batCacheid);
		if (msg){
			BBPreleaseref(bid);
			return msg;
		}
		BBPdecref(b->batCacheid, TRUE);

		d = c->data;
		/* its probably an empty bat which we shouldn't free 
		BBPdecref(d->ibid, TRUE);*/
		d->bid = 0;	
		d->ibase = 0;
		d->ibid = bid; /* use the insert bat */

		c->base.wtime = c->base.rtime = tr->stime;
	}
	/* bat was cleared */
	t->cleared = 1;
	return MAL_SUCCEED;
}



str SQLnewDictionary(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci){
	
	str *sch = (str *) getArgReference(stk,pci,1);
	str *tbl = (str *) getArgReference(stk,pci,2);
	sql_schema	*s;
	sql_table 	*t;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_trans *tr = m->session->tr;
	node *o;
	char buf[BUFSIZ], *nme= buf;
	int ret;

	if (msg)
		return msg;
   	s = mvc_bind_schema(m, *sch);
	if ( s == NULL)
		throw(MAL,"sql.cluster","Schema missing");
	t = mvc_bind_table(m, s, *tbl);
	if ( t == NULL)
		throw(MAL,"sql.cluster","Table missing");

	/* actually build the hash on the multi-column primary key */

	for (o = t->columns.set->h; o; o = o->next) {
		BAT *b, *e;
		sql_delta *d;
		sql_column	*c = o->data;

		b = store_funcs.bind_col(tr, c, 0);
		if (b == NULL)
			throw(MAL,"sql.compress","Can not access descriptor");
		e = BATnew(b->htype, b->ttype, 0);
		BATsetaccess(e, BAT_READ);
		d = c->data;
		d->bid = 0;
		d->ibase = 0;
		d->ibid = e->batCacheid; /* use the insert bat */
		c->base.wtime = c->base.rtime = tr->stime;
		snprintf(buf,BUFSIZ,"%s/%s/%s/0", *sch, *tbl, c->base.name);
		msg = DICTcompress(&ret, &nme, &b->batCacheid);
		BBPkeepref(e->batCacheid);
		BBPreleaseref(b->batCacheid);
		BBPdecref(b->batCacheid, TRUE);
	}
	/* bat was cleared */
	t->cleared = 1;
	t->base.wtime = s->base.wtime = tr->wtime = tr->stime;
	t->base.rtime = s->base.rtime = tr->rtime = tr->stime;
	return MAL_SUCCEED;



}
str SQLdropDictionary(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci){
	
	str *sch = (str *) getArgReference(stk,pci,1);
	str *tbl = (str *) getArgReference(stk,pci,2);
	sql_schema	*s;
	sql_table 	*t;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_trans *tr = m->session->tr;
	node *o;
	char buf[BUFSIZ], *nme= buf;
	int ret;

	if (msg)
		return msg;
   	s = mvc_bind_schema(m, *sch);
	if ( s == NULL)
		throw(MAL,"sql.cluster","Schema missing");
	t = mvc_bind_table(m, s, *tbl);
	if ( t == NULL)
		throw(MAL,"sql.cluster","Table missing");

	/* actually build the hash on the multi-column primary key */

	for (o = t->columns.set->h; o; o = o->next) {
		BAT *b, *e;
		sql_delta *d;
		sql_column	*c = o->data;

		b = store_funcs.bind_col(tr, c, 0);
		if (b == NULL)
			throw(MAL,"sql.compress","Can not access descriptor");
		e = BATnew(b->htype, b->ttype, 0);
		BATsetaccess(e, BAT_READ);
		d = c->data;
		d->bid = 0;
		d->ibase = 0;
		d->ibid = e->batCacheid; /* use the insert bat */
		c->base.wtime = c->base.rtime = tr->stime;
		snprintf(buf,BUFSIZ,"%s/%s/%s/0", *sch, *tbl, c->base.name);
		msg = DICTdecompress(&ret, &nme);
		BBPkeepref(e->batCacheid);
		BBPreleaseref(b->batCacheid);
		BBPdecref(b->batCacheid, TRUE);
	}
	/* bat was cleared */
	t->cleared = 1;
	t->base.wtime = s->base.wtime = tr->wtime = tr->stime;
	t->base.rtime = s->base.rtime = tr->rtime = tr->stime;
	return MAL_SUCCEED;



}


str SQLgzcompress(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci){
	
	str *sch = (str *) getArgReference(stk,pci,1);
	str *tbl = (str *) getArgReference(stk,pci,2);
	sql_schema	*s;
	sql_table 	*t;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_trans *tr = m->session->tr;
	node *o;
	int ret,i;
	char buf[PATHLENGTH], *sbuf= buf;

	if (msg)
		return msg;
   	s = mvc_bind_schema(m, *sch);
	if ( s == NULL)
		throw(MAL,"sql.gzcompress","Schema missing");
	t = mvc_bind_table(m, s, *tbl);
	if ( t == NULL)
		throw(MAL,"sql.gzcompress","Table missing");

	/* actually build the hash on the multi-column primary key */

	for (o = t->columns.set->h; o; o = o->next) {
		BAT *b;
		sql_column	*c = o->data;

		for ( i= 0; i<3; i++) {
			b = store_funcs.bind_col(tr, c, i);
			if (b == NULL)
				throw(MAL,"sql.gzcompress","Can not access descriptor");
			snprintf(buf,PATHLENGTH,"%s_%s_%s_%d", *sch, *tbl, c->base.name,i);
			msg = CMDbbpcompress(&ret, &b->batCacheid,&sbuf);
			BBPreleaseref(b->batCacheid);
		}
	}
	return MAL_SUCCEED;



}
str SQLgzdecompress(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci){
	
	str *sch = (str *) getArgReference(stk,pci,1);
	str *tbl = (str *) getArgReference(stk,pci,2);
	sql_schema	*s;
	sql_table 	*t;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_trans *tr = m->session->tr;
	node *o;
	int ret,i;
	char buf[PATHLENGTH], *sbuf= buf;

	if (msg)
		return msg;
   	s = mvc_bind_schema(m, *sch);
	if ( s == NULL)
		throw(MAL,"sql.gzdecompress","Schema missing");
	t = mvc_bind_table(m, s, *tbl);
	if ( t == NULL)
		throw(MAL,"sql.gzdecompress","Table missing");

	/* actually build the hash on the multi-column primary key */

	for (o = t->columns.set->h; o; o = o->next) {
		BAT *b;
		sql_column	*c = o->data;

		for ( i= 0; i<3; i++) {
			b = store_funcs.bind_col(tr, c, i);
			if (b == NULL)
				throw(MAL,"sql.gzdecompress","Can not access descriptor");
			snprintf(buf,PATHLENGTH,"%s_%s_%s_%d", *sch, *tbl, c->base.name,i);
			msg = CMDbbpdecompress(&ret, &b->batCacheid,&sbuf);
			BBPreleaseref(b->batCacheid);
		}
	}
	return MAL_SUCCEED;



}
str SQLtruncate(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci){
	
	str *sch = (str *) getArgReference(stk,pci,1);
	str *tbl = (str *) getArgReference(stk,pci,2);
	sql_schema	*s;
	sql_table 	*t;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_trans *tr = m->session->tr;
	node *o;
	int ret,i;
	char buf[PATHLENGTH], *sbuf= buf;

	if (msg)
		return msg;
   	s = mvc_bind_schema(m, *sch);
	if ( s == NULL)
		throw(MAL,"sql.truncate","Schema missing");
	t = mvc_bind_table(m, s, *tbl);
	if ( t == NULL)
		throw(MAL,"sql.truncate","Table missing");

	/* actually build the hash on the multi-column primary key */

	for (o = t->columns.set->h; o; o = o->next) {
		BAT *b;
		sql_column	*c = o->data;

		for ( i= 0; i<3; i++) {
			b = store_funcs.bind_col(tr, c, i);
			if (b == NULL)
				throw(MAL,"sql.truncate","Can not access descriptor");
			snprintf(buf,PATHLENGTH,"%s_%s_%s_%d", *sch, *tbl, c->base.name,i);
			msg = CMDbbptruncate(&ret, &b->batCacheid,&sbuf);
			BBPreleaseref(b->batCacheid);
		}
	}
	return MAL_SUCCEED;



}
str SQLexpand(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci){
	
	str *sch = (str *) getArgReference(stk,pci,1);
	str *tbl = (str *) getArgReference(stk,pci,2);
	sql_schema	*s;
	sql_table 	*t;
	mvc *m = NULL;
	str msg = getContext(cntxt,mb, &m, NULL);
	sql_trans *tr = m->session->tr;
	node *o;
	int ret,i;
	char buf[PATHLENGTH], *sbuf= buf;

	if (msg)
		return msg;
   	s = mvc_bind_schema(m, *sch);
	if ( s == NULL)
		throw(MAL,"sql.expand","Schema missing");
	t = mvc_bind_table(m, s, *tbl);
	if ( t == NULL)
		throw(MAL,"sql.expand","Table missing");

	/* actually build the hash on the multi-column primary key */

	for (o = t->columns.set->h; o; o = o->next) {
		BAT *b;
		sql_column	*c = o->data;

		for ( i= 0; i<3; i++) {
			b = store_funcs.bind_col(tr, c, i);
			if (b == NULL)
				throw(MAL,"sql.expand","Can not access descriptor");
			snprintf(buf,PATHLENGTH,"%s_%s_%s_%d", *sch, *tbl, c->base.name,i);
			msg = CMDbbpexpand(&ret, &b->batCacheid,&sbuf);
			BBPreleaseref(b->batCacheid);
		}
	}
	return MAL_SUCCEED;



}


str
SQLrdfShred(Client cntxt, MalBlkPtr mb, MalStkPtr stk, InstrPtr pci)
{
#ifdef HAVE_MONETDB5_RDF
	bat docbats;
	BAT *graph = NULL;
	BAT *p, *s, *o;
	bat *g;
	sql_schema *sch;
	sql_table *g_tbl;
	sql_column *gname, *gid;
#if STORE == TRIPLE_STORE
	sql_table *spo_tbl, *sop_tbl, *pso_tbl, *pos_tbl, *osp_tbl, *ops_tbl;
#elif STORE == MLA_STORE
	sql_table *spo_tbl;
#endif /* STORE */
	sql_table *map_tbl;
	sql_subtype tpe;
	str *location = (str *) getArgReference(stk,pci,1);
	str *name = (str *) getArgReference(stk,pci,2);
	str *schema = (str *) getArgReference(stk,pci,3);
	char buff[24];
	mvc *m = NULL;
	int id = 0;
	oid rid = oid_nil;

	str msg = getContext(cntxt, mb, &m, NULL);
	if (msg)
		return msg;

	sch = mvc_bind_schema(m, *schema);
	if ( sch == NULL)
		throw(MAL,"sql.rdfSchred","Schema missing");
	g_tbl = mvc_bind_table(m, sch, "graph");
	gname = mvc_bind_column(m, g_tbl, "gname");
	gid = mvc_bind_column(m, g_tbl, "gid");

	rid = table_funcs.column_find_row(m->session->tr, gname, *name, NULL);
	if (rid != oid_nil) {
		throw(SQL,"sql.rdfShred","graph name allready exists in rdf.graph");
	}
	id = (int) store_funcs.count_col(gname);
	store_funcs.append_col(m->session->tr, gname, *name, TYPE_str);
	store_funcs.append_col(m->session->tr, gid, &id, TYPE_int);

	msg = RDFParser(&docbats, location, name, schema);
	if (msg)
		return msg;

	if (sizeof(oid) == 8) {
		sql_find_subtype(&tpe, "oid", 31, 0);
		/* todo for niels: if use int/bigint the @0 is serialized */
		/* sql_find_subtype(&tpe, "bigint", 64, 0); */
	} else {
		sql_find_subtype(&tpe, "oid", 31, 0);
		/* sql_find_subtype(&tpe, "int", 32, 0); */
	}
#if STORE == TRIPLE_STORE
	
sprintf(buff, "spo%d", id);
spo_tbl = mvc_create_table(m, sch, buff, tt_table, 0, SQL_PERSIST, 0, 3);
mvc_create_column(m, spo_tbl, "subject", &tpe);
mvc_create_column(m, spo_tbl, "property", &tpe);
mvc_create_column(m, spo_tbl, "object", &tpe);




	
sprintf(buff, "sop%d", id);
sop_tbl = mvc_create_table(m, sch, buff, tt_table, 0, SQL_PERSIST, 0, 3);
mvc_create_column(m, sop_tbl, "subject", &tpe);
mvc_create_column(m, sop_tbl, "object", &tpe);
mvc_create_column(m, sop_tbl, "property", &tpe);




	
sprintf(buff, "pso%d", id);
pso_tbl = mvc_create_table(m, sch, buff, tt_table, 0, SQL_PERSIST, 0, 3);
mvc_create_column(m, pso_tbl, "property", &tpe);
mvc_create_column(m, pso_tbl, "subject", &tpe);
mvc_create_column(m, pso_tbl, "object", &tpe);




	
sprintf(buff, "pos%d", id);
pos_tbl = mvc_create_table(m, sch, buff, tt_table, 0, SQL_PERSIST, 0, 3);
mvc_create_column(m, pos_tbl, "property", &tpe);
mvc_create_column(m, pos_tbl, "object", &tpe);
mvc_create_column(m, pos_tbl, "subject", &tpe);




	
sprintf(buff, "osp%d", id);
osp_tbl = mvc_create_table(m, sch, buff, tt_table, 0, SQL_PERSIST, 0, 3);
mvc_create_column(m, osp_tbl, "object", &tpe);
mvc_create_column(m, osp_tbl, "subject", &tpe);
mvc_create_column(m, osp_tbl, "property", &tpe);




	
sprintf(buff, "ops%d", id);
ops_tbl = mvc_create_table(m, sch, buff, tt_table, 0, SQL_PERSIST, 0, 3);
mvc_create_column(m, ops_tbl, "object", &tpe);
mvc_create_column(m, ops_tbl, "property", &tpe);
mvc_create_column(m, ops_tbl, "subject", &tpe);




#elif STORE == MLA_STORE
	
sprintf(buff, "spo%d", id);
spo_tbl = mvc_create_table(m, sch, buff, tt_table, 0, SQL_PERSIST, 0, 3);
mvc_create_column(m, spo_tbl, "subject", &tpe);
mvc_create_column(m, spo_tbl, "property", &tpe);
mvc_create_column(m, spo_tbl, "object", &tpe);




#endif /* STORE */

	sprintf(buff, "map%d", id);
	map_tbl = mvc_create_table(m, sch, buff, tt_table, 0, SQL_PERSIST, 0, 2);
	mvc_create_column(m, map_tbl, "sid", &tpe);
	sql_find_subtype(&tpe, "varchar", 1024, 0);
	mvc_create_column(m, map_tbl, "lexical", &tpe);

	if ((graph = BATdescriptor(docbats)) == NULL) {
		throw(SQL, "sql.rdfShred", "docbat is missing");
	}
	g = (bat *) Tloc(graph, BUNfirst(graph));

	s =  BATdescriptor(g[MAP_LEX]);
	if ( s == NULL){
		BBPreleaseref(graph->batCacheid);
		throw(SQL, "sql.rdfShred", "map_lex is missing");
	}

	store_funcs.append_col(m->session->tr,
			mvc_bind_column(m, map_tbl, "lexical"),
			BATmirror(BATmark(BATmirror(s),0)), TYPE_bat);
	store_funcs.append_col(m->session->tr,
			mvc_bind_column(m, map_tbl, "sid"),
			BATmirror(BATmark(s, 0)),
			TYPE_bat);
	BBPunfix(s->batCacheid);

#if STORE == TRIPLE_STORE
	
s = BATdescriptor(g[S_sort]);
p = BATdescriptor(g[P_PO]);
o = BATdescriptor(g[O_PO]);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, spo_tbl, "subject"),
		s, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, spo_tbl, "property"),
		p, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, spo_tbl, "object"),
		o, TYPE_bat);
BBPunfix(s->batCacheid);
BBPunfix(p->batCacheid);
BBPunfix(o->batCacheid);





	
s = BATdescriptor(g[S_sort]);
p = BATdescriptor(g[P_OP]);
o = BATdescriptor(g[O_OP]);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, sop_tbl, "subject"),
		s, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, sop_tbl, "property"),
		p, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, sop_tbl, "object"),
		o, TYPE_bat);
BBPunfix(s->batCacheid);
BBPunfix(p->batCacheid);
BBPunfix(o->batCacheid);





	
s = BATdescriptor(g[S_SO]);
p = BATdescriptor(g[P_sort]);
o = BATdescriptor(g[O_SO]);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, pso_tbl, "subject"),
		s, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, pso_tbl, "property"),
		p, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, pso_tbl, "object"),
		o, TYPE_bat);
BBPunfix(s->batCacheid);
BBPunfix(p->batCacheid);
BBPunfix(o->batCacheid);





	
s = BATdescriptor(g[S_OS]);
p = BATdescriptor(g[P_sort]);
o = BATdescriptor(g[O_OS]);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, pos_tbl, "subject"),
		s, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, pos_tbl, "property"),
		p, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, pos_tbl, "object"),
		o, TYPE_bat);
BBPunfix(s->batCacheid);
BBPunfix(p->batCacheid);
BBPunfix(o->batCacheid);





	
s = BATdescriptor(g[S_SP]);
p = BATdescriptor(g[P_SP]);
o = BATdescriptor(g[O_sort]);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, osp_tbl, "subject"),
		s, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, osp_tbl, "property"),
		p, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, osp_tbl, "object"),
		o, TYPE_bat);
BBPunfix(s->batCacheid);
BBPunfix(p->batCacheid);
BBPunfix(o->batCacheid);





	
s = BATdescriptor(g[S_PS]);
p = BATdescriptor(g[P_PS]);
o = BATdescriptor(g[O_sort]);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, ops_tbl, "subject"),
		s, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, ops_tbl, "property"),
		p, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, ops_tbl, "object"),
		o, TYPE_bat);
BBPunfix(s->batCacheid);
BBPunfix(p->batCacheid);
BBPunfix(o->batCacheid);





#elif STORE == MLA_STORE
	
s = BATdescriptor(g[S_sort]);
p = BATdescriptor(g[P_sort]);
o = BATdescriptor(g[O_sort]);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, spo_tbl, "subject"),
		s, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, spo_tbl, "property"),
		p, TYPE_bat);
store_funcs.append_col(m->session->tr, mvc_bind_column(m, spo_tbl, "object"),
		o, TYPE_bat);
BBPunfix(s->batCacheid);
BBPunfix(p->batCacheid);
BBPunfix(o->batCacheid);





#endif /* STORE */




	/* unfix graph */
	BBPunfix(graph->batCacheid);

	return MAL_SUCCEED;


#else
	(void) cntxt; (void) mb; (void) stk; (void) pci;
	throw(SQL, "sql.rdfShred", "RDF support is missing from Monet5");
#endif /* RDF */
}

